/*-------------------------------------------------------------------------

project: genFamMem - maintaining families of requirements

author: Jan Bredereke
copyright: Jan Bredereke, 2000
$Id: calc.c,v 1.34 2001/02/20 08:01:42 brederek Exp $

file: calc.c - perform various calculations on the relations, 
	       do semantic checking

-------------------------------------------------------------------------*/

#include <assert.h>
#include "global.h"
#include "relation.h"
#include "identifier.h"
#include "genidentifier.h"
#include "calc.h"

/*-------------------------------------------------------------------------
The local functions for internal use
-------------------------------------------------------------------------*/

/* get the first or last line number of a feature or family member */
/* returns: the line number */
/* The feature/family member must exist, and it must have exactly one such 
   line number. */
static int
lFirstOrLastLine(int name, relRelationH relFeatOrFamMem2Line)
{   int retCode; /* diverse return codes to check */
    relSetH tmpSet = NULL; /* an auxiliary set */
    relSetIt setIt; /* an iterator over a set */
    int line; /* the desired line number */

    assert(relFeatOrFamMem2Line);

    /* initialize the sets */

    tmpSet = relNewSet();

    retCode = relApply(tmpSet, relFeatOrFamMem2Line, name);
    assert(retCode == RELOK);
    assert(relCardSet(tmpSet) == 1);
    relInitSetIt(&setIt, tmpSet);
    assert(relMoreSetIt(&setIt));
    line = relCurrSetIt(&setIt);

    /* free allocated dynamic memory */

    relFreeSet(tmpSet);

    return(line);
}

/* get the first line number of a "parents" area, if it exists */
/* returns: the line number, if it exists */
/*          -1 otherwise */
/* The area must have at most one such line number. */
static int
lFirstLineOpt(int name, relRelationH relArea2Line)
{   int retCode; /* diverse return codes to check */
    relSetH tmpSet = NULL; /* an auxiliary set */
    relSetIt setIt; /* an iterator over a set */
    int line; /* the desired line number */

    assert(relArea2Line);

    /* initialize the sets */

    tmpSet = relNewSet();

    retCode = relApply(tmpSet, relArea2Line, name);
    if(retCode == RELCANTDO)
    {   line = -1;
    }
    else
    {
	assert(relCardSet(tmpSet) == 1);
	relInitSetIt(&setIt, tmpSet);
	assert(relMoreSetIt(&setIt));
	line = relCurrSetIt(&setIt);
    }

    /* free allocated dynamic memory */

    relFreeSet(tmpSet);

    return(line);
}

/* get the first line number of a feature */
/* returns: the line number */
/* The feature must exist, and it must have exactly one first line number. */
static int
lFirstFeatureLine(int name)
{   return(lFirstOrLastLine(name, idRels.feature2FirstLine));
}

/* get the last line number of a feature */
/* returns: the line number */
/* The feature must exist, and it must have exactly one last line number. */
static int
lLastFeatureLine(int name)
{   return(lFirstOrLastLine(name, idRels.feature2LastLine));
}

/* get the first line number of a family member */
/* returns: the line number */
/* The family member must exist, and it must have exactly one first line 
   number. */
static int
lFirstFamilyMemberLine(int name)
{   return(lFirstOrLastLine(name, idRels.famMem2FirstLine));
}

/* get the last line number of a family member */
/* returns: the line number */
/* The family member must exist, and it must have exactly one last line 
   number. */
static int
lLastFamilyMemberLine(int name)
{   return(lFirstOrLastLine(name, idRels.famMem2LastLine));
}

/* get the first line number of a section */
/* returns: the line number */
/* The section must exist, and it must have exactly one first line number. */
static int
lFirstSectionLine(int name)
{   return(lFirstOrLastLine(name, idRels.section2FirstLine));
}

/* get the last line number of a section */
/* returns: the line number */
/* The section must exist, and it must have exactly one last line number. */
static int
lLastSectionLine(int name)
{   return(lFirstOrLastLine(name, idRels.section2LastLine));
}

/* get the first line number of the "parents" area of a section */
/* returns: the line number, if it exists */
/*          -1 otherwise */
/* The section must have at most one first line number. */
static int
lFirstSectionParentsLine(int name)
{   return(lFirstLineOpt(name, idRels.sectionParents2FirstLine));
}

/* get the last line number of the "parents" area of a section */
/* returns: the line number */
/* The section must exist, and it must have exactly one last line number. */
static int
lLastSectionParentsLine(int name)
{   return(lFirstOrLastLine(name, idRels.sectionParents2LastLine));
}

/* get the feature of a section */
/* returns: the feature's name, if it exists */
/*          -1 otherwise */
static int
lSectionFeature(int section)
{   int retCode; /* a return code to check */
    relSetH tmpSet = NULL; /* an auxiliary set */
    relSetIt setIt; /* iterator over a set */
    int feature = 0; /* the result */

    /* initialize the set */

    tmpSet = relNewSet();

    /* look up the feature, and perform several assertions */

    retCode = relApplyInverse(tmpSet, idRels.feature2Section, section);
    if(retCode != RELOK)
    {    feature = -1;
    }
    else
    {   assert(relCardSet(tmpSet) == 1);
	relInitSetIt(&setIt, tmpSet);
	assert(relMoreSetIt(&setIt));
	feature = relCurrSetIt(&setIt);
    }

    /* free allocated dynamic memory */

    relFreeSet(tmpSet);

    return(feature);
}

/* set the uses-hierarchy level for the given set and recursively
   calculate the uses-hierarchy levels for their children */
/* Returns: the max uses-hierarchy level */
/* the section with no parents have the highest levels, and the sections
   which collect all the others, but are parents to none, are at level 0 */
static int
lUsesHierLevels()
{   int retCode; /* diverse return codes to check */
    relSetIt setIt; /* an iterator over a set */
    relSetH unprocessedNodes = NULL; /* the set of still unprocessed nodes */
    relSetH allNodes = NULL; /* the set of all nodes involved */
    relSetH tmpSet, tmpSet2 = NULL; /* auxiliary sets */
    relRelationH tmpRel = NULL; /* an auxiliary relation */
    int currLevel; /* the current level */
    int somethingDone; /* a flag whether there was progress */

    /* initialize the relations and sets */

    tmpSet = relNewSet();
    tmpSet2 = relNewSet();
    unprocessedNodes = relNewSet();
    allNodes = relNewSet();
    tmpRel = relNewRelation();

    /* calculate the set of all nodes involved */
    if(progOpts.onlyFeature)
    {   /* if requested, abstract the sections to features */
        relInverse(tmpRel, idRels.feature2Section);
	relApplyToSet(allNodes, tmpRel, idRels.selectedSections);
    }
    else
	relCopySet(allNodes, idRels.selectedSections);

    /* while still nodes unprocessed */
    for(/* start with level 0 */
        currLevel = 0;
	/* while still nodes unprocessed */
        ( relDomain(tmpSet, idRels.usesHierarchyLevel),
	  relCopySet(unprocessedNodes, allNodes),
	  relSetDiff(unprocessedNodes, tmpSet),
	  /* this is the end-loop condition: */
          relCardSet(unprocessedNodes)
	);
	/* when looping, go to next node level */
	currLevel++
       )
    {   /* loop over all elements of the set */
	somethingDone = 0;
	for(relInitSetIt(&setIt, unprocessedNodes);
	    relMoreSetIt(&setIt);
	    relNextSetIt(&setIt)
	    )
	{   /* if no unprocessed predecessor nodes */
	    relApplyInverse(tmpSet, idRels.usesHierarchy, relCurrSetIt(&setIt));
	    retCode = relSetIntersect(tmpSet2, tmpSet, unprocessedNodes);
	    if(retCode == RELCANTDO)
	    {   /* assign the current level to this node */
		retCode = relAddRelship(idRels.usesHierarchyLevel,
					relCurrSetIt(&setIt),
					currLevel);
		somethingDone = 1;
	    }
	}
	assert(somethingDone);
    }

    /* free allocated dynamic memory */

    relFreeSet(tmpSet);
    relFreeSet(tmpSet2);
    relFreeSet(unprocessedNodes);
    relFreeSet(allNodes);
    relFreeRelation(tmpRel);

    return(--currLevel);
}

/*-------------------------------------------------------------------------
The functions for external use
-------------------------------------------------------------------------*/

/* add the implicitly generated sections and classes for pure
   family members */
/* returns: the number of semantic errors encountered */
/* It also prints error messages on stdout. */
int
calcAddGeneratedSections()
{   int semanticErrorNum = 0; /* the numbers of semantic errors encountered */
    int retCode; /* diverse return codes to check */
    relSetIt setIt, setIt2; /* iterators over a set */
    relSetH currFamMemFeatures = NULL; /* features of the currently processed 
    				          family member */
    int anonFeature = IDERR; /* name of the generated anonymous feature */
    int tmpId, tmpId2 = IDERR; /* some temporary identifiers */

    /* initialize the sets and relations */

    currFamMemFeatures = relNewSet();

    /* for all family members */
    for(relInitSetIt(&setIt, idRels.familyMembers);
        relMoreSetIt(&setIt);
	relNextSetIt(&setIt))
    { /* if this family member is not pure, then we have to do nothing */
      if(relIsMember(idRels.nonPureFamilyMembers, 
                     relCurrSetIt(&setIt)) == RELOK)
	  continue;

      /* otherwise, generate a feature chapter body */

      /* generate an anonymous feature for the family member */
      anonFeature = gidGenUniqueIdentifier(
      			idId2Identifier(relCurrSetIt(&setIt)));
      relAddMember(idRels.features, anonFeature);
      relAddMember(idRels.anonymousFeatures, anonFeature);

      /* add the anonymous feature to the family member */
      relAddRelship(idRels.famMem2FeatureList,
      		    relCurrSetIt(&setIt),
		    anonFeature);

      /* generate the collect section */
      retCode = relAddMember(idRels.sections, relCurrSetIt(&setIt));
      if(retCode != RELOK)
      {   fprintf(stderr, "%s: semantic error: implicit section %s conflicts with explicit section of same name.\n",
		  PROGNAME, 
		  idId2Identifier(relCurrSetIt(&setIt)));
	  semanticErrorNum++;
      }
      relAddMember(idRels.collectSections, relCurrSetIt(&setIt));

      /* add the section to the anonymous feature */
      relAddRelship(idRels.feature2Section, anonFeature, relCurrSetIt(&setIt));

      /* generate the collecting class for the system requirements */
      relAddRelship(idRels.section2Class,
		    relCurrSetIt(&setIt),
		    idConcatIds2Id(relCurrSetIt(&setIt),
				   idIdentifier2Id("Sys")));

      /* generate the collecting class for the environment requirements */
      relAddRelship(idRels.section2Class,
		    relCurrSetIt(&setIt),
		    idConcatIds2Id(relCurrSetIt(&setIt),
				   idIdentifier2Id("Env")));

      /* generate the section's list of parents,
         and the classes' inherited-from list (from the family member's list
	 of features) */
      relApply(currFamMemFeatures,
      	       idRels.famMem2FeatureList,
	       relCurrSetIt(&setIt));
      for(relInitSetIt(&setIt2, currFamMemFeatures);
	  relMoreSetIt(&setIt2);
	  relNextSetIt(&setIt2))
      {
	  /* skip any anonymous features */
	  if(relIsMember(idRels.anonymousFeatures, relCurrSetIt(&setIt2))
	     == RELOK)
	  {   continue;
	  }
	  /* do the rest */
	  tmpId = idConcatIds2Id(relCurrSetIt(&setIt2),
				 idIdentifier2Id("Feature"));
	  relAddRelship(idRels.sectionParents, relCurrSetIt(&setIt), tmpId);
	  if(relIsMember(idRels.sections, tmpId) != RELOK)
	  {   fprintf(stderr, "%s: semantic error: section %s is missing,\n\tbut required for family member %s.\n",
		      PROGNAME, 
		      idId2Identifier(tmpId),
		      idId2Identifier(relCurrSetIt(&setIt)));
	      semanticErrorNum++;
	  }
	  tmpId = idConcatIds2Id(relCurrSetIt(&setIt), idIdentifier2Id("Sys")),
	  tmpId2 = idConcatIds2Id(relCurrSetIt(&setIt2),
				  idIdentifier2Id("FeatureSys"));
	  relAddRelship3(idRels.sectionClass2Inherits,
	  		relCurrSetIt(&setIt),
			tmpId,
			tmpId2);
	  tmpId = idConcatIds2Id(relCurrSetIt(&setIt), idIdentifier2Id("Env")),
	  tmpId2 = idConcatIds2Id(relCurrSetIt(&setIt2),
				  idIdentifier2Id("FeatureEnv"));
	  relAddRelship3(idRels.sectionClass2Inherits,
	  		relCurrSetIt(&setIt),
			tmpId,
			tmpId2);
      }

      /* generate the class for the composition of system and environment */
      relAddRelship(idRels.section2Class,
		    relCurrSetIt(&setIt),
		    relCurrSetIt(&setIt));
      /* ... and its interface-inheritance relationship to the previous
         environment collection class */
      relAddRelship3(idRels.sectionClass2Inherits,
		    relCurrSetIt(&setIt),
		    relCurrSetIt(&setIt),
		    idConcatIds2Id(relCurrSetIt(&setIt),
				   idIdentifier2Id("Env")));
    }

    /* free allocated dynamic memory */

    relFreeSet(currFamMemFeatures);

    /* return the number of semantic errors encountered */
    return(semanticErrorNum);
}

/* calculate the relations for the specific family member */
/* returns: the number of semantic errors encountered */
/* It also prints error messages on stdout. */
int
calcRelsSpecificMember()
{   int semanticErrorNum = 0; /* the numbers of semantic errors encountered */
    int retCode; /* diverse return codes to check */
    relRelationH tmpRel = NULL; /* an auxiliary relation */
    relRelationH tmpRel2 = NULL; /* an auxiliary relation */
    relRelation3H tmpRel3 = NULL; /* an auxiliary ternary relation */
    relSetH tmpSet, tmpSet2 = NULL; /* auxiliary sets */
    relSetH allSelectedSections = NULL; /* an auxiliary set with all selected
    				sections, disregarding any removes */
    relSetIt setIt, setIt2; /* iterators over a set */
    relRelationIt relIt; /* an iterator over a relation */
    relRelation3It rel3It; /* an iterator over a ternary relation */
    int sectionFeature; /* temporary buffer for the feature of a section */
    int tmpLine = -1; /* temporary buffer for a line number */
    int tmpClass = IDERR; /* temporary buffer for a line number */

    /* initialize the sets and relations */

    tmpSet = relNewSet();
    tmpSet2 = relNewSet();
    allSelectedSections = relNewSet();
    tmpRel = relNewRelation();
    tmpRel2 = relNewRelation();
    tmpRel3 = relNewRelation3();

    /* check for family-member-independent specification errors */

    /* check for proper parents */
    for(relInitSetIt(&setIt, idRels.sections);
        relMoreSetIt(&setIt);
	relNextSetIt(&setIt))
    {   relApply(tmpSet, idRels.sectionParents, relCurrSetIt(&setIt));
	for(relInitSetIt(&setIt2, tmpSet);
	    relMoreSetIt(&setIt2);
	    relNextSetIt(&setIt2))
	{   
	    /* check that all parents are sections, after all */
	    /* (This might not be the case, since they can be used before
	       defined, so we can't check at that point of time.) */
	    if(relIsMember(idRels.sections, relCurrSetIt(&setIt2))
	       == RELCANTDO)
	    {   fprintf(stderr, "%s: semantic error: section %s\n\thas %s as a parent,\n\tbut that is not a section.\n",
			PROGNAME, 
			idId2Identifier(relCurrSetIt(&setIt)),
			idId2Identifier(relCurrSetIt(&setIt2)));
		semanticErrorNum++;
	    }

	    /* check that normal or default sections do not have a collect 
	       section as a parent */
	    if(relIsMember(idRels.collectSections, relCurrSetIt(&setIt))
	       == RELCANTDO)
	    {   if(relIsMember(idRels.collectSections, relCurrSetIt(&setIt2))
		   == RELOK)
		{   fprintf(stderr, "%s: semantic error: the essential or changeable \n\tsection %s has the collect section %s as a parent.\n",
			    PROGNAME, 
			    idId2Identifier(relCurrSetIt(&setIt)),
			    idId2Identifier(relCurrSetIt(&setIt2)));
		    semanticErrorNum++;
		}
	    }

	    /* check that if the parent of a normal section is from
	       the same feature or family member, then it must not
	       be a default section */
	    if(   (relIsMember(idRels.defaultSections, relCurrSetIt(&setIt))
	          == RELCANTDO)
	       && (relIsMember(idRels.collectSections, relCurrSetIt(&setIt))
	          == RELCANTDO)
	       && (relIsMember(idRels.defaultSections, relCurrSetIt(&setIt2))
	          == RELOK)
	       && (  lSectionFeature(relCurrSetIt(&setIt)) != -1)
	       && (  lSectionFeature(relCurrSetIt(&setIt)) 
		  == lSectionFeature(relCurrSetIt(&setIt2))))
	    {   fprintf(stderr, "%s: semantic error: the essential section %s has \n\tthe changeable section %s from the same feature %s as a parent.\n",
			PROGNAME, 
			idId2Identifier(relCurrSetIt(&setIt)),
			idId2Identifier(relCurrSetIt(&setIt2)),
			idId2Identifier(
			    lSectionFeature(relCurrSetIt(&setIt))));
		semanticErrorNum++;
	    }

	    /*  check that the parents lists of collect sections
		contain sections only that are either from the same
		feature or that are collect sections. */
	    if(   (relIsMember(idRels.collectSections, relCurrSetIt(&setIt))
	          == RELOK)
	       && (  lSectionFeature(relCurrSetIt(&setIt)) != -1)
	       && (  lSectionFeature(relCurrSetIt(&setIt2)) != -1)
	       && (  lSectionFeature(relCurrSetIt(&setIt)) 
		  != lSectionFeature(relCurrSetIt(&setIt2)))
	       && (   relIsMember(idRels.collectSections, relCurrSetIt(&setIt2))
	           == RELCANTDO)
		  )
	    {   fprintf(stderr, "%s: semantic error: the collect section %s from the feature %s has \n\tthe non-collect section %s from the different \n\tfeature %s as a parent.\n",
			PROGNAME, 
			idId2Identifier(relCurrSetIt(&setIt)),
			idId2Identifier(
			    lSectionFeature(relCurrSetIt(&setIt))),
			idId2Identifier(relCurrSetIt(&setIt2)),
			idId2Identifier(
			    lSectionFeature(relCurrSetIt(&setIt2))));
		semanticErrorNum++;
	    }

	}
    }

    /* check that a remove clause names default sections only */
    relRange(tmpSet, idRels.remove);
    relSetDiff(tmpSet, idRels.defaultSections);
    for(relInitSetIt(&setIt, tmpSet);
        relMoreSetIt(&setIt);
	relNextSetIt(&setIt))
    {   /* get more information for the error messages */
        retCode = relApplyInverse(tmpSet2, idRels.remove, relCurrSetIt(&setIt));
	assert(relCardSet(tmpSet2) >= 1);
	assert(retCode == RELOK);
	for(relInitSetIt(&setIt2, tmpSet2);
	    relMoreSetIt(&setIt2);
	    relNextSetIt(&setIt2))
	{   fprintf(stderr, "%s: semantic error: a remove clause of feature %s \n\tremoves the non-default section %s.\n",
		    PROGNAME, 
		    idId2Identifier(relCurrSetIt(&setIt2)),
		    idId2Identifier(relCurrSetIt(&setIt)));
	    semanticErrorNum++;
	}
    }

    /* calculations for the selected family member */

    /* the features selected by the selected family member */
    retCode = relApply(idRels.selectedFeatures, 
    		       idRels.famMem2FeatureList,
		       progOpts.selectedFamilyMember);
    if(retCode == RELCANTDO)
    {   fprintf(stderr, "%s: warning: the selected family member '%s' has no features.\n",
		PROGNAME, idId2Identifier(progOpts.selectedFamilyMember));
    }

    /* the sections removed by the features of the selected family member */
    relProjectDomain(tmpRel, idRels.remove, idRels.selectedFeatures);
    relRange(idRels.selectedRemove, tmpRel);

    /* the sections selected by the selected family member */
    /* take all sections of all selected features ... */
    relProjectDomain(tmpRel, idRels.feature2Section, idRels.selectedFeatures);
    relRange(idRels.selectedSections, tmpRel);
    relCopySet(allSelectedSections, idRels.selectedSections);
    /* ... and delete the "remove"d ones */
    relSetDiff(idRels.selectedSections, idRels.selectedRemove);

    /* the selected parents of the selected sections */
    relProjectDomain(tmpRel, idRels.sectionParents, idRels.selectedSections);
    relCopySet(tmpSet, idRels.sections);
    relSetDiff(tmpSet, idRels.selectedRemove);
    relProjectRange(idRels.selectedSectionParents, tmpRel, tmpSet);
    /* the deleted parents of the selected sections */
    relProjectRange(idRels.selectedSectionDeletedParents,
		    tmpRel,
		    idRels.selectedRemove);

    /* for each class of each selected section, the classes deleted in its
       "inherit" area */
    relApplyToSet(tmpSet, idRels.section2Class, idRels.selectedSections);
    relProjectFirstElem3(tmpRel3,
    			 idRels.sectionClass2DefaultProperties,
			 idRels.selectedSections);
    for(relInitRelation3It(&rel3It, tmpRel3);
        relMoreRelation3It(&rel3It);
	relNextRelation3It(&rel3It))
    {   if(   relIsMember(tmpSet, relThirdElem3(relCurrRelation3It(&rel3It)))
	   == RELCANTDO)
        {   relAddRelship3(idRels.selectedSectionClassDeletedDefaultProperties,
	                   relFirstElem3(relCurrRelation3It(&rel3It)),
	                   relSecondElem3(relCurrRelation3It(&rel3It)),
	                   relThirdElem3(relCurrRelation3It(&rel3It)));
	}
    }

    /* check for specification errors */

    /* check whether the removed sections are part of the selected
       features, at all */
    relCopySet(tmpSet, idRels.selectedRemove);
    relSetDiff(tmpSet, allSelectedSections);
    for(relInitSetIt(&setIt, tmpSet);
        relMoreSetIt(&setIt);
	relNextSetIt(&setIt))
    {   fprintf(stderr, "%s: semantic error: the removed section %s\n\tis not part of the selected features.\n",
		PROGNAME, 
		idId2Identifier(relCurrSetIt(&setIt)));
	semanticErrorNum++;
    }

    /* check whether some section is removed more than once by the
       selected features */
    for(relInitSetIt(&setIt, idRels.selectedRemove);
        relMoreSetIt(&setIt);
	relNextSetIt(&setIt))
    {   retCode = relApplyInverse(tmpSet, idRels.remove, relCurrSetIt(&setIt));
	assert(retCode == RELOK);
	if(relCardSet(tmpSet) != 1)
	{   fprintf(stderr, "%s: warning: the section %s is removed more than once\n\tby the selected features.\n",
		    PROGNAME, idId2Identifier(relCurrSetIt(&setIt)));
	}
    }

    /* check whether the parents of the selected non-collect sections are
       selected sections, only */
    /* (Please note that we have reduced the parents of collect sections
       already. We therefore do not need to handle them separately anymore.
       Also, we have already checked that the reference parents from the
       same feature, only.) */
    relRange(tmpSet, idRels.selectedSectionParents);
    relSetDiff(tmpSet, idRels.selectedSections);
    for(relInitSetIt(&setIt, tmpSet);
        relMoreSetIt(&setIt);
	relNextSetIt(&setIt))
    {   /* get more information for the error messages */
        retCode = relApplyInverse(tmpSet2,
				  idRels.selectedSectionParents,
				  relCurrSetIt(&setIt));
        assert(retCode == RELOK);
	for(relInitSetIt(&setIt2, tmpSet2);
	    relMoreSetIt(&setIt2);
	    relNextSetIt(&setIt2))
	{   fprintf(stderr, "%s: semantic error: the parent section %s\n\tof section %s is not part of the selected features.\n",
		    PROGNAME, 
		    idId2Identifier(relCurrSetIt(&setIt)),
		    idId2Identifier(relCurrSetIt(&setIt2)));
	    semanticErrorNum++;
	}
    }

    /* the features deleted because not selected */
    relCopySet(idRels.deletedFeatures, idRels.features);
    relSetDiff(idRels.deletedFeatures, idRels.selectedFeatures);
    /* do not count the anonymous features */
    relSetDiff(idRels.deletedFeatures, idRels.anonymousFeatures);

    /* the family members deleted because not selected */
    relCopySet(idRels.deletedFamilyMembers, idRels.familyMembers);
    relDelMember(idRels.deletedFamilyMembers, progOpts.selectedFamilyMember);

    /* compute the set of line numbers to delete */

    /* take all deleted family members */
    for(relInitSetIt(&setIt, idRels.deletedFamilyMembers);
        relMoreSetIt(&setIt);
        relNextSetIt(&setIt)
	)
    {   retCode = linAddRange(idRels.deletedLineRanges,
			      lFirstFamilyMemberLine(relCurrSetIt(&setIt)),
			      lLastFamilyMemberLine(relCurrSetIt(&setIt)));
        assert(retCode == LINOK);
    }
    /* take all deleted features */
    for(relInitSetIt(&setIt, idRels.deletedFeatures);
        relMoreSetIt(&setIt);
        relNextSetIt(&setIt)
	)
    {   retCode = linAddRange(idRels.deletedLineRanges,
			      lFirstFeatureLine(relCurrSetIt(&setIt)),
			      lLastFeatureLine(relCurrSetIt(&setIt)));
        assert(retCode == LINOK);
    }
    /* take all deleted sections in the features selected */
    for(relInitSetIt(&setIt, idRels.selectedRemove);
        relMoreSetIt(&setIt);
        relNextSetIt(&setIt)
	)
    {   retCode = linAddRange(idRels.deletedLineRanges,
			      lFirstSectionLine(relCurrSetIt(&setIt)),
			      lLastSectionLine(relCurrSetIt(&setIt)));
        assert(retCode == LINOK);
    }

    /* compute the set of line numbers to edit as "parents" area */

    /* take all selected sections in the features selected */
    for(relInitSetIt(&setIt, idRels.selectedSections);
        relMoreSetIt(&setIt);
        relNextSetIt(&setIt)
	)
    {   tmpLine = lFirstSectionParentsLine(relCurrSetIt(&setIt));
	/* if there is a "parents" area, add it */
        if(tmpLine != -1)
	{   retCode = linAddRange(idRels.parentsLineRanges,
				 tmpLine,
				 lLastSectionParentsLine(relCurrSetIt(&setIt)));
	    assert(retCode == LINOK);
	}
    }

    /* compute the set of line numbers to edit as "inherit" area */

    /* take all selected sections in the features selected */
    for(relInitSetIt(&setIt, idRels.selectedSections);
        relMoreSetIt(&setIt);
        relNextSetIt(&setIt)
	)
    {   /* take all the (class, inherit-first-line) pairs for such a section */
        relApply3(tmpRel,
		  idRels.sectionClassInherit2FirstLine,
		  relCurrSetIt(&setIt));
        /* we will need also the (class, inherit-last-line) relation */
	relApply3(tmpRel2,
		  idRels.sectionClassInherit2LastLine,
		  relCurrSetIt(&setIt));
	/* iterate over all such pairs */
	for(relInitRelationIt(&relIt, tmpRel);
	    relMoreRelationIt(&relIt);
	    relNextRelationIt(&relIt)
	    )
	{   /* get the last line for the corresponding first line */
	    tmpClass = relFirstElem(relCurrRelationIt(&relIt));
	    retCode = relApply(tmpSet, tmpRel2, tmpClass);
	    assert(retCode == RELOK);
	    /* there must be only one "inherit" area for each class: check */
	    assert(relCardSet(tmpSet) == 1);
	    relInitSetIt(&setIt2, tmpSet);
	    tmpLine = relCurrSetIt(&setIt2);
	    /* finally, add to set of line ranges */
	    retCode = linAddRange(idRels.inheritLineRanges,
				 relSecondElem(relCurrRelationIt(&relIt)),
				 tmpLine);
	    assert(retCode == LINOK);
	}
    }

    /* free allocated dynamic memory */

    relFreeSet(tmpSet);
    relFreeSet(tmpSet2);
    relFreeSet(allSelectedSections);
    relFreeRelation(tmpRel);
    relFreeRelation(tmpRel2);
    relFreeRelation3(tmpRel3);

    /* return the number of semantic errors encountered */
    return(semanticErrorNum);
}

/* extend the line numbers towards the end of the document just before
   the next item begins, except for *_end. */
void
calcExtendLineNumbers()
{   int retCode; /* diverse return codes to check */
    relSetH tmpSet = NULL; /* an auxiliary set */
    relSetIt setIt, setIt2, setIt3; /* iterators over a set */
    int oldLastLine = 0; /* old last line of the current feature */
    int newLastLine = 0; /* new last line of the current feature */
    int tmpFirstLine = 0; /* the first line of some feature */
    int tmpLastLine = 0; /* the last line of some feature */
    int featureLastLine = 0; /* last line of the current feature of the
                                section */

    /* initialize the sets */

    tmpSet = relNewSet();

    /* extend the last-line numbers of the features to the next feature
       or to the next family member, except if it has got a feature_end
       construct, or if it is an anonymous feature with a familymember_end
       construct */
    for(relInitSetIt(&setIt, idRels.features);
        relMoreSetIt(&setIt);
        relNextSetIt(&setIt)
	)
    {   /* skip if we have a feature_end construct for the current feature */
        retCode = relIsMember(idRels.featureEnds, relCurrSetIt(&setIt));
        if(retCode == RELCANTDO)
	{   /* get old last line of feature */
	    oldLastLine = lLastFeatureLine(relCurrSetIt(&setIt));
	    assert(oldLastLine >= 0);
	    /* skip if anonymous feature with no last line */
	    if(oldLastLine)
	    {   /* look for smallest first line of any feature that is
	           larger than the old last line of the current feature */
		/* -1 for a last line means end-of-input */
		newLastLine = -1;
		for(relInitSetIt(&setIt2, idRels.features);
		    relMoreSetIt(&setIt2);
		    relNextSetIt(&setIt2)
		    )
		{   tmpFirstLine = lFirstFeatureLine(relCurrSetIt(&setIt2));
		    /* skip if anonymous feature with no first line */
		    if(tmpFirstLine)
		    {   assert(tmpFirstLine > 0);
		        tmpLastLine = tmpFirstLine - 1;
			if(   (tmpLastLine > oldLastLine)
			   && (   (newLastLine == -1)
			       || (tmpLastLine < newLastLine))
			  )
			{   newLastLine = tmpLastLine;
			}
		    }
		}
	        /* look for smallest first line of any family member that is
	           larger than the old last line of the current feature */
		for(relInitSetIt(&setIt2, idRels.familyMembers);
		    relMoreSetIt(&setIt2);
		    relNextSetIt(&setIt2)
		    )
		{   tmpFirstLine = 
		        lFirstFamilyMemberLine(relCurrSetIt(&setIt2));
		    assert(tmpFirstLine > 0);
		    tmpLastLine = tmpFirstLine - 1;
		    if(   (tmpLastLine > oldLastLine)
		       && (   (newLastLine == -1)
			   || (tmpLastLine < newLastLine))
		      )
		    {   newLastLine = tmpLastLine;
		    }
		}
		assert((newLastLine == -1) || (newLastLine >= oldLastLine));
		/* if something found */
		if((newLastLine == -1) || (newLastLine > oldLastLine))
		{   /* delete the old last line number ... */
		    retCode = relDelRelship(idRels.feature2LastLine,
		                            relCurrSetIt(&setIt),
					    oldLastLine);
		    assert(retCode == RELOK);
		    /* ... and add the new last line number */
		    retCode = relAddRelship(idRels.feature2LastLine,
		                            relCurrSetIt(&setIt),
					    newLastLine);
		    assert(retCode == RELOK);
		}
	    }
	}
    }

    /* extend the last-line numbers of the family members to the next feature
       or to the next family member, except if it has got a familymember_end
       construct */
    for(relInitSetIt(&setIt, idRels.familyMembers);
        relMoreSetIt(&setIt);
        relNextSetIt(&setIt)
	)
    {   /* skip if we have a familymember_end construct for the current family
           member */
        retCode = relIsMember(idRels.familyMemberEnds, relCurrSetIt(&setIt));
        if(retCode == RELCANTDO)
	{   /* get old last line of family member */
	    oldLastLine = lLastFamilyMemberLine(relCurrSetIt(&setIt));
	    assert(oldLastLine > 0);
	    /* look for smallest first line of any feature that is
	       larger than the old last line of the current family member */
	    /* -1 for a last line means end-of-input */
	    newLastLine = -1;
	    for(relInitSetIt(&setIt2, idRels.features);
		relMoreSetIt(&setIt2);
		relNextSetIt(&setIt2)
		)
	    {   tmpFirstLine = lFirstFeatureLine(relCurrSetIt(&setIt2));
		/* skip if anonymous feature with no first line */
		if(tmpFirstLine)
		{   assert(tmpFirstLine > 0);
		    tmpLastLine = tmpFirstLine - 1;
		    if(   (tmpLastLine > oldLastLine)
		       && (   (newLastLine == -1)
			   || (tmpLastLine < newLastLine))
		      )
		    {   newLastLine = tmpLastLine;
		    }
		}
	    }
	    /* look for smallest first line of any family member that is
	       larger than the old last line of the current family member */
	    for(relInitSetIt(&setIt2, idRels.familyMembers);
		relMoreSetIt(&setIt2);
		relNextSetIt(&setIt2)
		)
	    {   tmpFirstLine = 
		    lFirstFamilyMemberLine(relCurrSetIt(&setIt2));
		assert(tmpFirstLine > 0);
		tmpLastLine = tmpFirstLine - 1;
		if(   (tmpLastLine > oldLastLine)
		   && (   (newLastLine == -1)
		       || (tmpLastLine < newLastLine))
		  )
		{   newLastLine = tmpLastLine;
		}
	    }
	    assert((newLastLine == -1) || (newLastLine >= oldLastLine));
	    /* if something found */
	    if((newLastLine == -1) || (newLastLine > oldLastLine))
	    {   /* delete the old last line number ... */
		retCode = relDelRelship(idRels.famMem2LastLine,
					relCurrSetIt(&setIt),
					oldLastLine);
		assert(retCode == RELOK);
		/* ... and add the new last line number */
		retCode = relAddRelship(idRels.famMem2LastLine,
					relCurrSetIt(&setIt),
					newLastLine);
		assert(retCode == RELOK);
	    }
	}
    }

    /* extend the last-line numbers of the sections within a feature
     to the next section, except if it has got a section_end */

    /* iterate over all features */
    for(relInitSetIt(&setIt, idRels.features);
        relMoreSetIt(&setIt);
        relNextSetIt(&setIt)
	)
    {   /* get the last line of the current feature */
	featureLastLine = lLastFeatureLine(relCurrSetIt(&setIt));
        /* iterate over all sections within the current feature */
	relApply(tmpSet, idRels.feature2Section, relCurrSetIt(&setIt));
	for(relInitSetIt(&setIt2, tmpSet);
	    relMoreSetIt(&setIt2);
	    relNextSetIt(&setIt2)
	    )
        {   /* skip if we have a section_end construct for the current section*/
	    retCode = relIsMember(idRels.sectionEnds, relCurrSetIt(&setIt2));
	    if(retCode == RELCANTDO)
	    {   /* get old last line of section */
		oldLastLine = lLastSectionLine(relCurrSetIt(&setIt2));
		assert(oldLastLine > 0);
		assert(   (oldLastLine <= featureLastLine)
		       || (featureLastLine == -1));
	        /* look for smallest first line of any section of this feature
		   that is larger than the old last line of the current 
		   section */
		/* -1 for a last line means end-of-input */
		newLastLine = featureLastLine;
		for(relInitSetIt(&setIt3, tmpSet);
		    relMoreSetIt(&setIt3);
		    relNextSetIt(&setIt3)
		    )
		{   tmpFirstLine = lFirstSectionLine(relCurrSetIt(&setIt3));
		    assert(tmpFirstLine > 0);
		    tmpLastLine = tmpFirstLine - 1;
		    if(   (tmpLastLine > oldLastLine)
		       && (   (newLastLine == -1)
			   || (tmpLastLine < newLastLine))
		      )
		    {   newLastLine = tmpLastLine;
		    }
		}
		assert((newLastLine == -1) || (newLastLine >= oldLastLine));
		assert(   (newLastLine <= featureLastLine)
		       || (featureLastLine == -1));
		/* if something found */
		if((newLastLine == -1) || (newLastLine > oldLastLine))
		{   /* delete the old last line number ... */
		    retCode = relDelRelship(idRels.section2LastLine,
		                            relCurrSetIt(&setIt2),
					    oldLastLine);
		    assert(retCode == RELOK);
		    /* ... and add the new last line number */
		    retCode = relAddRelship(idRels.section2LastLine,
		                            relCurrSetIt(&setIt2),
					    newLastLine);
		    assert(retCode == RELOK);
		}
	    }
        }
    }


    /* free allocated dynamic memory */

    relFreeSet(tmpSet);
}

/* calculate the uses-hierarchy for the parameters specified on the 
   command line */
/* The root nodes get level 0, the others increasing values. */
void
calcUsesHierarchy()
{   int retCode; /* diverse return codes to check */
    relRelationH tmpRel, tmpRel2 = NULL; /* auxiliary relations */
    relSetH tmpSet = NULL; /* an auxiliary set */
    int semanticErrorNum = 0; /* the numbers of semantic errors encountered */

    /* initialize the relations and sets */

    tmpRel = relNewRelation();
    tmpRel2 = relNewRelation();
    tmpSet = relNewSet();

    /* if requested, restrict the uses-hierarchy to one family member */
    if(progOpts.usesHier == UHIERFAMMEM)
    {   /* calculate the relations for the specific family member */
	semanticErrorNum = calcRelsSpecificMember();
	/* don't continue if any semantic errors encountered */
	if(semanticErrorNum)
	{   fprintf(stderr, 
		    "\n%s: %i semantic error(s) encountered. Exiting ...\n\n",
		    PROGNAME,
		    semanticErrorNum); 
	    exit(EXITSYNERR);
	}
    }

    /* else if requested, restrict the uses-hierarchy to one feature  */
    else if(progOpts.usesHier == UHIERFEAT)
    {   retCode = relIsMember(idRels.features, progOpts.selectedFeature);
        if(retCode == RELCANTDO)
	{   fprintf(stderr, 
		   "\n%s: '%s' is not a feature of the family. Exiting ...\n\n",
		    PROGNAME,
		    idId2Identifier(progOpts.selectedFeature));
	    exit(EXITSYNERR);
	}
        relApply(idRels.selectedSections, 
		 idRels.feature2Section,
		 progOpts.selectedFeature);
    }

    /* else take all sections */
    else
    {   relCopySet(idRels.selectedSections, idRels.sections);
    }

    /* any section at all selected? */
    if(!relCardSet(idRels.selectedSections))
    {   fprintf(stderr, 
		"\n%s: no sections selected. Exiting ...\n\n",
		PROGNAME);
	exit(EXITSYNERR);
    }

    /* compute the uses-hierarchy relation */
    relProjectDomain(idRels.usesHierarchy,
		     idRels.sectionParents,
		     idRels.selectedSections);

    /* if a specific family member was requested, remove all relationships
       to sections outside of it */
    if(progOpts.usesHier == UHIERFAMMEM)
    {   relCopyRelation(tmpRel, idRels.usesHierarchy);
        relProjectRange(idRels.usesHierarchy, tmpRel, idRels.selectedSections);
    }

    /* find the set of root nodes */
    retCode = relCopySet(idRels.usesHierarchyRoots, idRels.selectedSections);
    assert(retCode == RELOK);
    retCode = relRange(tmpSet, idRels.usesHierarchy);
    assert(retCode == RELOK);
    relSetDiff(idRels.usesHierarchyRoots, tmpSet);
    if(relIsEmptySet(idRels.usesHierarchyRoots) == RELOK)
    {   fprintf(stderr, 
	       "\n%s: the uses-hierarchy is circular. Exiting ...\n\n",
		PROGNAME);
	exit(EXITSYNERR);
    }

    /* if requested, abstract the sections to features */
    if(progOpts.onlyFeature)
    {   /* abstract the uses-hierarchy */
        relInverse(tmpRel2, idRels.feature2Section);
	relApplyToRelation(tmpRel, tmpRel2, idRels.usesHierarchy);
	relCopyRelation(idRels.usesHierarchy, tmpRel);
	/* and abstract the set of its root nodes */
	relApplyToSet(tmpSet, tmpRel2, idRels.usesHierarchyRoots);
	relCopySet(idRels.usesHierarchyRoots, tmpSet);
    }

    /* remove superfluous arrows from transitive closure, 
       in the uses-hierarchy  */
    retCode = relDeTransitive(idRels.usesHierarchy);
    assert(retCode == RELOK);

    /* remove superfluous reflexive arrows, i.e., back-arrows to the same 
       node, in the uses-hierarchy  */
    retCode = relDeReflexive(idRels.usesHierarchy);
    assert(retCode == RELOK);

    /* calculate the uses-hierarchy levels */
    idRels.usesHierarchyDepth = lUsesHierLevels();

    /* free allocated dynamic memory */

    relFreeRelation(tmpRel);
    relFreeRelation(tmpRel2);
    relFreeSet(tmpSet);

}
