/*-------------------------------------------------------------------------

project: genFamMem - maintaining families of requirements

author: Jan Bredereke
copyright: Jan Bredereke, 2000
$Id: debug.c,v 1.26 2001/02/15 08:59:52 brederek Exp $

file: debug.c - functions to output debug information

-------------------------------------------------------------------------*/

#include <stdio.h>
#include <assert.h>
#include "global.h"
#include "identifier.h"
#include "debug.h"

/*-------------------------------------------------------------------------
The local functions for internal use
-------------------------------------------------------------------------*/

/* some print functions for passing to the dump routines */

/* print one member of a set */
static void
lDumpMember(FILE* stream, int name)
{   fprintf(stream, "%s", idId2Identifier(name));
}

/* print one domain member */
static void
lDumpDomainMember(FILE* stream, int name)
{   fprintf(stream, "%s: ", idId2Identifier(name));
}

/* print one domain member, ignoring the selector value */
static void
lDumpDomainMemberS(FILE* stream, int name, int selector)
{   fprintf(stream, "%s: ", idId2Identifier(name));
}

/* print one range member */
static void
lDumpRangeMember(FILE* stream, int name)
{    lDumpMember(stream, name);
}

/* print one range member, ignoring the domainMember value */
static void
lDumpRangeMemberS(FILE* stream, int name, int domainMember)
{    lDumpMember(stream, name);
}

/* print one number, ignoring the domainMember value */
static void
lDumpNumberS(FILE* stream, int number, int domainMember)
{   fprintf(stream, "%i", number);
}

/* print a member at the first position of a relationship */
static void
lDumpFirstPosMember(FILE* stream, int name)
{   fprintf(stream, "%s\n", idId2Identifier(name));
    fprintf(stream, "-------------------------------\n");
}

/* print a member at the second position of a relationship, 
   ignoring the selector value */
static void
lDumpSecondPosMemberS(FILE* stream, int name, int selector)
{   lDumpDomainMember(stream, name);
}

/* print a member at the third position of a relationship */
static void
lDumpThirdPosMember(FILE* stream, int name)
{   lDumpRangeMember(stream, name);
}

/* print a member at the third position of a relationship, , 
   ignoring the secondPosMember value */
static void
lDumpThirdPosMemberS(FILE* stream, int name, int secondPosMember)
{   lDumpRangeMember(stream, name);
}

/* print a "default_section" or "collect_section" flag */
static void
lDumpSectionFlag(FILE* stream, int dFlag, int cFlag)
{   if(dFlag)
    {   assert(!cFlag);
	fprintf(stream, "(changeable)");
    }
    else if(cFlag)
	fprintf(stream, "(collecting)");
    else
	fprintf(stream, "(essential)");
}

/* print a line number */
static void
lDumpLineNumber(FILE* stream, int lineNumber)
{   fprintf(stream, "%i", lineNumber);
}

/* print one member of a set, including its line numbers */
static void
lDumpMemberLineNumber(FILE* stream, int name, 
		      relRelationH toFirstLine, relRelationH toLastLine)
{   relSetH tmpSet;
    int retCode;

    tmpSet = relNewSet();
    lDumpMember(stream, name);
    fprintf(stream, "[");
    retCode = relApply(tmpSet, toFirstLine, name);
    if(retCode == RELOK)
	relDumpSet(stream, tmpSet, lDumpLineNumber, relDumpMemberSpace);
    fprintf(stream, "-");
    retCode = relApply(tmpSet, toLastLine, name);
    if(retCode == RELOK)
	relDumpSet(stream, tmpSet, lDumpLineNumber, relDumpMemberSpace);
    fprintf(stream, "]");
    relFreeSet(tmpSet);
}

/* print one member of a set, including its line numbers, for a specific
   selector value */
static void
lDumpSelectorMemberLineNumber(FILE* stream, int name, 
			      relRelation3H toFirstLine,
			      relRelation3H toLastLine,
			      int selector)
{   relSetH tmpSet;
    relRelationH tmpRel;
    int retCode;

    tmpSet = relNewSet();
    tmpRel = relNewRelation();
    lDumpMember(stream, name);
    fprintf(stream, "[");
    retCode = relApply3(tmpRel, toFirstLine, selector);
    if(retCode == RELOK)
    {   retCode = relApply(tmpSet, tmpRel, name);
	if(retCode == RELOK)
	    relDumpSet(stream, tmpSet, lDumpLineNumber, relDumpMemberSpace);
    }
    fprintf(stream, "-");
    retCode = relApply3(tmpRel, toLastLine, selector);
    if(retCode == RELOK)
    {   retCode = relApply(tmpSet, tmpRel, name);
	if(retCode == RELOK)
	    relDumpSet(stream, tmpSet, lDumpLineNumber, relDumpMemberSpace);
    }
    fprintf(stream, "]");
    relFreeRelation(tmpRel);
    relFreeSet(tmpSet);
}

/* print one feature, including its line numbers */
static void
lDumpFeature(FILE* stream, int name)
{   lDumpMemberLineNumber(stream, name, 
			  idRels.feature2FirstLine, idRels.feature2LastLine);
}

/* print one family member, including its line numbers */
static void
lDumpFamMem(FILE* stream, int name)
{   lDumpMemberLineNumber(stream, name, 
			  idRels.famMem2FirstLine, idRels.famMem2LastLine);
}

/* print one section, including its line numbers */
static void
lDumpSectionLineNumber(FILE* stream, int name)
{   lDumpMemberLineNumber(stream, name, 
			  idRels.section2FirstLine, idRels.section2LastLine);
}

/* print one class of a section, including its line numbers */
static void
lDumpSectionClassLineNumber(FILE* stream, int class, int section)
{   lDumpSelectorMemberLineNumber(stream, class, 
				  idRels.sectionClass2FirstLine,
				  idRels.sectionClass2LastLine,
				  section);
}

/* print one section, including the line numbers of its parents part, if any */
static void
lDumpSectionParentsLineNumber(FILE* stream, int name)
{   lDumpMemberLineNumber(stream, name, 
			  idRels.sectionParents2FirstLine,
			  idRels.sectionParents2LastLine);
}

/* print one class of a section, including the line numbers of its
   inheritance etc. parts */
static void
lDumpClassInheritanceLineNumber(FILE* stream, int name, int section)
{   lDumpSelectorMemberLineNumber(stream, name, 
				  idRels.sectionClassInherit2FirstLine,
				  idRels.sectionClassInherit2LastLine,
				  section);
}

/* print one member of a set, including its line numbers for 
   the given line number relations, and including a flag which is
   given by two other sets, and for which a dump function is provided. */
static void
lDumpSelectorsMemberLineNumberFlag(FILE* stream, int name, 
             relRelationH toFirstLine, relRelationH toLastLine,
	     relSetH setFlag1, relSetH setFlag2,
	     void (*dumpFlag)(FILE *, int, int))
{   relSetH tmpSet;
    int retCode, retCode2;

    tmpSet = relNewSet();
    lDumpMember(stream, name);
    fprintf(stream, "[");
    retCode = relApply(tmpSet, toFirstLine, name);
    if(retCode == RELOK)
	relDumpSet(stream, tmpSet, lDumpLineNumber, relDumpMemberSpace);
    fprintf(stream, "-");
    retCode = relApply(tmpSet, toLastLine, name);
    if(retCode == RELOK)
	relDumpSet(stream, tmpSet, lDumpLineNumber, relDumpMemberSpace);
    fprintf(stream, "]");

    retCode = relIsMember(setFlag1, name);
    retCode2 = relIsMember(setFlag2, name);
    (*dumpFlag)(stream, (retCode == RELOK), (retCode2 == RELOK));

    relFreeSet(tmpSet);
}

/* print one section, including its line numbers
   and including the changeable/essential flag */
static void
lDumpSection(FILE* stream, int section, int selector)
{   lDumpSelectorsMemberLineNumberFlag(stream, section, 
				       idRels.section2FirstLine,
				       idRels.section2LastLine,
				       idRels.defaultSections,
				       idRels.collectSections,
				       lDumpSectionFlag);
}

/*-------------------------------------------------------------------------
The functions for external use
-------------------------------------------------------------------------*/

/* dump the raw relations */
void
dbgDumpRawRelations()
{   printf("The features of the family\n");
    printf("==========================\n");
    relDumpSet(stdout, idRels.features, lDumpFeature, relDumpMemberSpace);
    printf("\n\n");
    printf("The features of the family with a feature_end construct\n");
    printf("=======================================================\n");
    relDumpSet(stdout,
	       idRels.featureEnds, lDumpFeature, relDumpMemberSpace);
    printf("\n\n");
    printf("The features which are anonymous\n");
    printf("================================\n");
    relDumpSet(stdout,
	       idRels.anonymousFeatures, lDumpFeature, relDumpMemberSpace);
    printf("\n\n");
    printf("The family members of the family\n");
    printf("================================\n");
    relDumpSet(stdout, idRels.familyMembers, 
	       lDumpFamMem,
	       relDumpMemberSpace);
    printf("\n\n");
    printf("The family members of the family with a familymember_end construct\n");
    printf("==================================================================\n");
    relDumpSet(stdout, idRels.familyMemberEnds, 
	       lDumpFamMem,
	       relDumpMemberSpace);
    printf("\n\n");
    printf("The non-pure family members\n");
    printf("===========================\n");
    relDumpSet(stdout, idRels.nonPureFamilyMembers, 
	       lDumpFamMem,
	       relDumpMemberSpace);
    printf("\n\n");
    printf("The sections of the family\n");
    printf("==========================\n");
    relDumpSet(stdout, idRels.sections, 
	       lDumpSectionLineNumber,
	       relDumpMemberSpace);
    printf("\n\n");
    printf("The feature lists of the family members\n");
    printf("=======================================\n");
    relDumpRelation(stdout, idRels.famMem2FeatureList, 
		    lDumpDomainMemberS, lDumpRangeMemberS, 0,
		    relDumpMemberNewline, relDumpMemberSpace);
    printf("\n\n");
    printf("The sections of the features, and their default_section/collect_section flag\n");
    printf("============================================================================\n");
    relDumpRelation(stdout, idRels.feature2Section,
		    lDumpDomainMemberS,
		    lDumpSection,
		    0,
		    relDumpMemberNewline,
		    relDumpMemberSpace);
    printf("\n\n");
    printf("The sections with a section_end construct\n");
    printf("=========================================\n");
    relDumpSet(stdout, idRels.sectionEnds, 
	       lDumpSectionLineNumber,
	       relDumpMemberSpace);
    printf("\n\n");
    printf("the parents lists of the sections\n");
    printf("=================================\n");
    relDumpRelation(stdout, idRels.sectionParents,
		     lDumpDomainMemberS,
		     lDumpRangeMemberS,
		     0,
		     relDumpMemberNewline,
		     relDumpMemberSpace);
    printf("\n\n");
    printf("the inherited classes of the classes of the sections\n");
    printf("====================================================\n");
    relDumpRelation3(stdout, idRels.sectionClass2Inherits,
		     lDumpFirstPosMember,
		     lDumpSecondPosMemberS,
		     lDumpThirdPosMemberS,
		     relDumpMemberNewline,
		     relDumpMemberNewline,
		     relDumpMemberSpace);
    printf("\n\n");
    printf("those inherited classes that appear in \"default_properties\"\n");
    printf("===========================================================\n");
    relDumpRelation3(stdout, idRels.sectionClass2DefaultProperties,
		     lDumpFirstPosMember,
		     lDumpSecondPosMemberS,
		     lDumpThirdPosMemberS,
		     relDumpMemberNewline,
		     relDumpMemberNewline,
		     relDumpMemberSpace);
    printf("\n\n");
    printf("The removed sections of the features\n");
    printf("====================================\n");
    relDumpRelation(stdout, idRels.remove,
		    lDumpDomainMemberS, lDumpRangeMemberS, 0,
		    relDumpMemberNewline, relDumpMemberSpace);
    printf("\n\n");
    printf("The classes of the sections of the family and their line numbers\n");
    printf("================================================================\n");
    relDumpRelation(stdout, idRels.section2Class, 
		    lDumpDomainMemberS, lDumpSectionClassLineNumber, 0,
		    relDumpMemberNewline, relDumpMemberSpace);
    printf("\n\n");
    printf("The classes of the sections, and their inheritance etc. part line numbers\n");
    printf("=========================================================================\n");
    relDumpRelation(stdout, idRels.section2Class, 
		    lDumpDomainMemberS, lDumpClassInheritanceLineNumber, 0,
		    relDumpMemberNewline, relDumpMemberSpace);
    printf("\n\n");
    printf("The sections of the family, and their parents part line numbers\n");
    printf("===============================================================\n");
    relDumpSet(stdout, idRels.sections, 
	       lDumpSectionParentsLineNumber,
	       relDumpMemberSpace);
    printf("\n");
}

/* dump the reduced relations */
void
dbgDumpReducedRelations()
{   printf("\n");
    printf("The features selected by the selected family member\n");
    printf("===================================================\n");
    relDumpSet(stdout, idRels.selectedFeatures,
	       lDumpMember,
	       relDumpMemberSpace);
    printf("\n\n");
    printf("The sections removed by the selected family member\n");
    printf("==================================================\n");
    relDumpSet(stdout, idRels.selectedRemove,
	       lDumpMember,
	       relDumpMemberSpace);
    printf("\n\n");
    printf("The sections selected by the selected family member\n");
    printf("===================================================\n");
    relDumpSet(stdout, idRels.selectedSections,
	       lDumpMember,
	       relDumpMemberSpace);
    printf("\n\n");
    printf("the (selected) parents lists of the selected sections\n");
    printf("=====================================================\n");
    relDumpRelation(stdout, idRels.selectedSectionParents,
		     lDumpDomainMemberS,
		     lDumpRangeMemberS,
		     0,
		     relDumpMemberNewline,
		     relDumpMemberSpace);
    printf("\n\n");
    printf("the deleted parents of the selected sections\n");
    printf("============================================\n");
    relDumpRelation(stdout, idRels.selectedSectionDeletedParents,
		     lDumpDomainMemberS,
		     lDumpRangeMemberS,
		     0,
		     relDumpMemberNewline,
		     relDumpMemberSpace);
    printf("\n\n");
    printf("for each class of each selected section, the classes deleted in its \"inherit\" area\n");
    printf("==================================================================================\n");
    relDumpRelation3(stdout,
    		     idRels.selectedSectionClassDeletedDefaultProperties,
		     lDumpFirstPosMember,
		     lDumpSecondPosMemberS,
		     lDumpThirdPosMemberS,
		     relDumpMemberNewline,
		     relDumpMemberNewline,
		     relDumpMemberSpace);
    printf("\n\n");
    printf("the features deleted because not selected\n");
    printf("=========================================\n");
    relDumpSet(stdout, idRels.deletedFeatures,
	       lDumpMember,
	       relDumpMemberSpace);
    printf("\n\n");
    printf("the family members deleted because not selected\n");
    printf("===============================================\n");
    relDumpSet(stdout, idRels.deletedFamilyMembers,
	       lDumpMember,
	       relDumpMemberSpace);
    printf("\n\n");
    printf("the line ranges deleted because not selected\n");
    printf("============================================\n");
    linDumpLineRange(stdout, idRels.deletedLineRanges);
    printf("\n\n");
    printf("the line ranges of \"parents\" areas\n");
    printf("==================================\n");
    linDumpLineRange(stdout, idRels.parentsLineRanges);
    printf("\n\n");
    printf("the line ranges of \"inherit\" areas\n");
    printf("==================================\n");
    linDumpLineRange(stdout, idRels.inheritLineRanges);
    printf("\n\n");
    			
}

/* dump the reduced relations */
void
dbgDumpUsesHierarchy()
{   printf("\n");
    printf("The sections selected\n");
    printf("=====================\n");
    relDumpSet(stdout, idRels.selectedSections,
	       lDumpMember,
	       relDumpMemberSpace);
    printf("\n\n");
    printf("the uses-hierarchy\n");
    printf("==================\n");
    relDumpRelation(stdout, idRels.usesHierarchy,
		     lDumpDomainMemberS,
		     lDumpRangeMemberS,
		     0,
		     relDumpMemberNewline,
		     relDumpMemberSpace);
    printf("\n\n");
    printf("the uses-hierarchy root nodes\n");
    printf("=============================\n");
    relDumpSet(stdout, idRels.usesHierarchyRoots,
	       lDumpMember,
	       relDumpMemberSpace);
    printf("\n\n");
    printf("the uses-hierarchy levels\n");
    printf("=========================\n");
    relDumpRelation(stdout, idRels.usesHierarchyLevel,
		     lDumpDomainMemberS,
		     lDumpNumberS,
		     0,
		     relDumpMemberNewline,
		     relDumpMemberSpace);
    printf("\n\n");
    printf("the max. uses-hierarchy level\n");
    printf("=============================\n");
    printf("%i\n", idRels.usesHierarchyDepth);
    printf("\n\n");
    			
}
