/*-------------------------------------------------------------------------

project: genFamMem - maintaining families of requirements

author: Jan Bredereke
copyright: Jan Bredereke, 2000
$Id: filterspec.c,v 1.29 2001/02/15 08:59:52 brederek Exp $

file: filterspec.c - filter a specification, deleting and
		     editing lines as appropriate

If NDEBUG is defined, the internal assertions are not checked anymore.

-------------------------------------------------------------------------*/

#include <stdio.h>
#include <string.h>
#include <regex.h>
#include <assert.h>
#include "global.h"
#include "filterspec.h"
#include "identifier.h"
#include "linerange.h"

extern FILE *sptyyin; /* input stream for second pass */
extern FILE *sptyyout; /* output stream for second pass */

/* The state of the filtering, as far as it needs to be accessed globally 
   by the lexer */
struct filStateType filState;

/* Further local state about the filtering */
struct lFilStateType
{   int nextChangePrintingLine; /* next line number where "printing" will
    				   change */
    linLineRangeIt deleteLinIt; /* iterator over the set of line ranges 
    				   to delete */
    int nextChangeInheritLine; /* next line number where a "inherit" area will
    				   start/stop */
    linLineRangeIt inheritLinIt; /* iterator over the set of line ranges 
    				    of "inherit" areas */
    int nextChangeParentsLine; /* next line number where a "parents" area will
    				   start/stop */
    linLineRangeIt parentsLinIt; /* iterator over the set of line ranges 
    				    of "parents" areas */
};
static struct lFilStateType lFilState;

/* the buffer where "inherit" or "parents" areas are buffered */
  /* length of a line, without terminating \n\0 */
#define BUFLINELENGTH 300
#define BUFLINENUM 300
static char lTextBuffer[BUFLINENUM][BUFLINELENGTH+2];
/* the max. buffer line in use (-1 == none) */
static int lBufferMaxLine = -1;
/* the first free position in the max. buffer line */
static int lBufferMaxChar = 0;


/*-------------------------------------------------------------------------
The functions for local use
-------------------------------------------------------------------------*/

/* convert an ID number of a name into an ID number of a completely
   escaped regex */
/* returns: the ID number of the completely escaped regex on success */
/* aborts otherwise */
static int
lEscapeRegex(int name)
{   char *resultStr = NULL; /* temporary buffer to build the escaped regex */
    char *nameStr = NULL; /* pointer to the string of the identifier */
    char* resultPtr = NULL; /* pointer into result string */
    int result = IDERR; /* ID number for result string */

    nameStr = idId2Identifier(name);
    assert(nameStr);
    resultStr = malloc((strlen(nameStr) * 3 * sizeof(char)) + 1);
    resultPtr = resultStr;
    while(*nameStr)
    {   *resultPtr++ = '[';
        *resultPtr++ = *nameStr++;
        *resultPtr++ = ']';
    }
    *resultPtr = '\0';
    result = idIdentifier2Id(resultStr);

    /* free dynamically allocated memory */
    free(resultStr);

    return(result);
}

/* set up memory to buffer the text in the "inherit" or "parents" area */
static void
lInitSaveBuffer()
{
    lBufferMaxLine = 0;
    lBufferMaxChar = 0;
    return;
}

/* edit the "parents or "inherit" area, and print it */
static void
lEditPrintArea(relSetH deleteNames, int isInheritArea)
{   int retCode; /* diverse return codes to check */
    relSetIt setIt; /* an iterator over a set */
    int currLine = LINERR; /* the line we are currently in */
    int currLine2 = LINERR; /* the line we are currently in while searching
    			       for an \endinherit */
    regex_t *patBufs; /* array of regex pattern buffers */
    int patBufLen = 0; /* number of entries in above array */
    regex_t *patBufPtr = NULL; /* pointer into above array */
    int regexPattern; /* ID of string of regex pattern */
    char wordStart[] = "^(|[^%]*[^[:alnum:]_%])"; /* regex for a start of a
    					   Z word with no LaTeX comment
					   character before it */
    char wordEnd[] =   "($|[^[:alnum:]_])"; /* regex for an end of a Z word */
    char commaRegex[] = "^[^%]*,"; /* regex for a comma with no LaTeX 
    			              comment character before it */
    regex_t patBufComma; /* regex pattern buffer for the above comma pattern */
    int noCommaFound = 0; /* flag whether no comma was found in previous line */
    int delCommaFound = 0; /* flag whether a comma to delete was found */
    char delCommaRegex[] = "^[^%]*(,)[^,%]*($|%)"; /* regex for deleting the
					last comma with no LaTeX
					comment character before it */
    regex_t patBufDelComma; /* regex pattern buffer for the above comma
    			       pattern */
    regmatch_t delCommaPos[2]; /* address of the found comma */
    char beginInheritRegex[] =
    "(\\\\inherit(interface)?[^[:alnum:]_]|\\\\begin\\{inherit(interface)?\\})";
    			/* regex for the begin of an "inherit" sub-area */
    regex_t patBufBeginInherit; /* regex pattern buffer for the above begin
    			           inherit pattern */
    char endInheritRegex[] =
  "(\\\\endinherit(interface)?[^[:alnum:]_]|\\\\end\\{inherit(interface)?\\})";
    			/* regex for the end of an "inherit" sub-area */
    regex_t patBufEndInherit; /* regex pattern buffer for the above end
				 inherit pattern */
    int cnt; /* loop index */

    /* this should never happen, but check anyway for non-terminated
       last line */
    if(lBufferMaxChar)
        filNextBufferLine();
    /* let the index point to the last line used */
    lBufferMaxLine--;

    /* pre-compile the comma and "inherit" regexs once */
    retCode = regcomp(&patBufComma, commaRegex, REG_EXTENDED | REG_NOSUB);
    assert(!retCode);
    retCode = regcomp(&patBufDelComma, delCommaRegex, REG_EXTENDED);
    assert(!retCode);
    retCode = regcomp(&patBufBeginInherit,
    		      beginInheritRegex,
		      REG_EXTENDED | REG_NOSUB);
    assert(!retCode);
    retCode = regcomp(&patBufEndInherit,
    		      endInheritRegex,
		      REG_EXTENDED | REG_NOSUB);
    assert(!retCode);

    /* pre-compile all variable regexs once */
    patBufLen = relCardSet(deleteNames);
    patBufs = (regex_t *) malloc(patBufLen * sizeof(regex_t));
    patBufPtr = &patBufs[0];
    for(relInitSetIt(&setIt, deleteNames);
	relMoreSetIt(&setIt);
	relNextSetIt(&setIt)
	)
    {   /* The identifier may also comprise a WORDSTROKE ',!,? .
	   At least the last must be escaped in the regex. */
	/* We can escape *any* character x in a regex by enclosing it in
	   brackets [x] . */
	/* [[:alnum:]] matches a letter or a digit. */
	/* [[:<:]] marks the start of a word, and [[:>:]] marks the end
	   of a word. (This does not appear to work, though.)
	   A word is an alnum or an underscore.
	   So, watch out for WORDSTROKEs! */
	regexPattern = idConcatIds2Id(
			  idIdentifier2Id(wordStart),
			  lEscapeRegex(relCurrSetIt(&setIt)));
	regexPattern = idConcatIds2Id(
			  regexPattern,
			  idIdentifier2Id(wordEnd));
	assert(regexPattern != IDERR);
	retCode = regcomp(patBufPtr++,
	                  idId2Identifier(regexPattern),
			  REG_EXTENDED | REG_NOSUB);
	assert(!retCode);
    }
    assert(patBufPtr == &patBufs[patBufLen]);

    /* check each line in the buffer whether it contains a name to delete */
    noCommaFound = 0;
    for(currLine = lBufferMaxLine; currLine >= 0; currLine--)
    {
	/* check for all name to delete whether they are in the current
	   line of the buffer */
	patBufPtr = &patBufs[0];
	for(cnt = 0; cnt < patBufLen; cnt++)
	{   /* does the name appear literally as a word somewhere
	       on the line? */
	    /* (This is kind of dirty, but there is no other way to
	       get back from a Z name to its LaTeX representation(s).) */

	    /* try to match regex on current line */
	    retCode = regexec(patBufPtr++,
			      lTextBuffer[currLine],
			      (size_t) 0,
			      (regmatch_t *) 0,
			      0);
	    if(!retCode)
	    {   /* found a match somewhere in the line */

		/* if not already pending, */
		/* find out whether there is a comma in this line */
		if(!noCommaFound)
		{   noCommaFound = regexec(&patBufComma,
					   lTextBuffer[currLine],
					   (size_t) 0,
					   (regmatch_t *) 0,
					   0);
		}

		/* delete this line, with all the formatting it contains */
		/* (We assume that there is at most one name per line!) */
		strcpy(lTextBuffer[currLine], "%\n");
	    }
	    else
	    {   /* no match, this line will not be deleted because of 
	           a name contained in it */

	        /* if there was no comma in deleted line(s), then delete
		   (last) comma in this line */
		if(noCommaFound)
		{   delCommaFound = ! regexec(&patBufDelComma,
					      lTextBuffer[currLine],
					      (size_t) 2,
					      delCommaPos,
					      0);

		    /* if this line indeed has a comma */
		    if(delCommaFound)
		    {
			/* check that there is indeed a start position of
			   the comma */
			assert(delCommaPos[1].rm_so != -1);
			/* check that the length of the comma is indeed 1 */
			assert(delCommaPos[1].rm_so + 1
			       == delCommaPos[1].rm_eo);
			/* delete this comma character */
			memmove(
			    &lTextBuffer[currLine][delCommaPos[1].rm_so],
			    &lTextBuffer[currLine][delCommaPos[1].rm_so + 1],
			    strlen(&lTextBuffer[currLine][delCommaPos[1].rm_so])
			        * sizeof(char));
			/* reset flag for pending comma deletion */
			noCommaFound = 0;
		    }
		    /* else if we just have hit an "\inherit", in our
		       search for a comma to delete */
		    else
		    if(isInheritArea
		       && 
		       ! regexec(&patBufBeginInherit,
				 lTextBuffer[currLine],
				 (size_t) 0,
				 (regmatch_t *) 0,
				 0))
		    {   /* then the entire sub-area is empty now, and must be
			   deleted up to the next following "\endinherit" */
			for(currLine2 = currLine;
			    currLine2 <= lBufferMaxLine;
			    currLine2++)
			{   if(! regexec(&patBufEndInherit,
					 lTextBuffer[currLine2],
					 (size_t) 0,
					 (regmatch_t *) 0,
					 0))
			    {   break;
			    }
			}
			/* check that \endinherit indeed was found */
			assert(currLine2 <= lBufferMaxLine);

			/* delete the area */
			for(cnt = currLine; cnt <= currLine2; cnt++)
			{   strcpy(lTextBuffer[cnt], "%\n");
			}

			/* reset flag for pending comma deletion */
			noCommaFound = 0;
		    }
		}
	    }
	}
	assert(patBufPtr == &patBufs[patBufLen]);
    }

    /* print the edited "parents" or "inherit" area */
    for(currLine = 0; currLine <= lBufferMaxLine; currLine++)
        fputs(lTextBuffer[currLine], sptyyout);

    /* free allocated dynamic memory */

    /* free the memory for the regexs */
    for(cnt = 0; cnt < patBufLen; cnt++)
	regfree(&patBufs[cnt]);
    free((void *)patBufs);

    /* free the memory for the comma regex */
    regfree(&patBufComma);

    return;
}

/* edit the "inherit" area, and print it */
static void
lEditPrintInherit()
{   int retCode; /* diverse return codes to check */
    relSetH tmpSet = NULL; /* an auxiliary set */
    relRelationH tmpRel = NULL; /* an auxiliary relation */
    int currFirstLine = LINERR; /* the first line of a section/class */
    int currSection = IDERR; /* the section we are currently in */
    int currClass = IDERR; /* the class we are currently in */
    relRelation3It rel3It; /* an iterator over a ternary relation */

    /* initialize the sets and relations */

    tmpSet = relNewSet();
    tmpRel = relNewRelation();

    /* get the name of the current section/class */
    for(relInitRelation3It(&rel3It, idRels.sectionClassInherit2LastLine);
	relMoreRelation3It(&rel3It);
	relNextRelation3It(&rel3It)
	)
    {   if(   relThirdElem3(relCurrRelation3It(&rel3It)) + 1
           == filState.currLineNo)
        {   currSection = relFirstElem3(relCurrRelation3It(&rel3It));
	    currClass = relSecondElem3(relCurrRelation3It(&rel3It));
	    break;
	}
    }
    /* check that something indeed was found */
    assert(currSection != IDERR);

    /* get the set of inherited classes to delete for the current class */
    relApply3(tmpRel,
	      idRels.selectedSectionClassDeletedDefaultProperties,
	      currSection);
    relApply(tmpSet, tmpRel, currClass);

    /* do the editing and printing */
    lEditPrintArea(tmpSet, 1);

    /* free allocated dynamic memory */
    relFreeSet(tmpSet);
    relFreeRelation(tmpRel);

    return;
}

/* edit the "parents" area, and print it */
static void
lEditPrintParents()
{   int retCode; /* diverse return codes to check */
    relSetH tmpSet = NULL; /* an auxiliary set */
    relRelationIt relIt; /* an iterator over a relation */
    int currFirstLine = LINERR; /* the first line of a section */
    int lastLine = LINERR; /* the corresponding last line */
    int currSection = IDERR; /* the section we are currently in */

    /* initialize the sets */

    tmpSet = relNewSet();

    /* get the name of the current section */
    /* We assume that there is a section around the "parents" area,
       such that its line range can be found. */
    /* REMARK: We now get the section name from the relation 
       section2FirstLine. We could have used, as well, the relation
       sectionParents2LastLine to get the section name of the
       "parents" area. That might have been a little simpler. */
    currFirstLine = LINERR;
    currSection = IDERR;
    /* look for the largest section start line number smaller than or equal
       to the current line */
    for(relInitRelationIt(&relIt, idRels.section2FirstLine);
	relMoreRelationIt(&relIt);
	relNextRelationIt(&relIt)
	)
    {   if((   relSecondElem(relCurrRelationIt(&relIt)) 
	    <= filState.currLineNo)
	   && (  (currFirstLine == LINERR) 
	      || (currFirstLine < relSecondElem(relCurrRelationIt(&relIt)))))
	{   currFirstLine = relSecondElem(relCurrRelationIt(&relIt));
	    currSection = relFirstElem(relCurrRelationIt(&relIt));
	}
    }
    /* check that something indeed was found */
    assert(currFirstLine != LINERR);
    /* check that the last line is indeed larger than or equal to the
       current line */
    lastLine = relApplyFunction(idRels.section2LastLine, currSection);
    /* we cannot do: assert(lastLine != RELERR);
       since lastLine may legally be -1, which is == RELERR */
    assert(linCmp(currFirstLine, lastLine) <= 0);

    /* get the set of parents to delete for the current section */
    retCode = relApply(tmpSet,
    		       idRels.selectedSectionDeletedParents,
		       currSection);

    /* do the editing and printing */
    lEditPrintArea(tmpSet, 0);

    /* free allocated dynamic memory */
    relFreeSet(tmpSet);

    return;
}

/* print the argument to the output stream, and count the number of
   newlines printed, for which we must make up later */
static void
lGenerateBody(char *text)
{   
    assert(text);
    fputs(text, sptyyout);
    while(*text)
        if(*text++ == '\n')
    	    filState.pendingEmptyLinesToDelete++;
    return;
}

/* check whether we have to append a generated feature chapter body
   for a family member */
/* This function should be called when the current line contains an
   "\endfeatures". */
static void
lCheckGeneratedFamMemBody()
{   int retCode; /* diverse return codes to check */
    relRelationIt relIt; /* an iterator over a relation */
    int currFirstLine = LINERR; /* the first line of a family member */
    int lastLine = LINERR; /* the corresponding last line */
    int currFamMem = IDERR; /* the family member we are currently in */
    relSetIt setIt; /* an iterator over a set */
    int firstItem = 0; /* flag, whether we just print the first item */
    
    /* find out in which family member we currently are */
    /* We assume that there is a family member around the "\endfeatures" line,
       such that it can be found. */
    currFirstLine = LINERR;
    currFamMem = IDERR;
    /* look for the largest family member start line number smaller than or
       equal to the current line */
    for(relInitRelationIt(&relIt, idRels.famMem2FirstLine);
	relMoreRelationIt(&relIt);
	relNextRelationIt(&relIt)
	)
    {   if((   relSecondElem(relCurrRelationIt(&relIt)) 
	    <= filState.currLineNo)
	   && (  (currFirstLine == LINERR) 
	      || (currFirstLine < relSecondElem(relCurrRelationIt(&relIt)))))
	{   currFirstLine = relSecondElem(relCurrRelationIt(&relIt));
	    currFamMem = relFirstElem(relCurrRelationIt(&relIt));
	}
    }
    /* check that something indeed was found */
    assert(currFirstLine != LINERR);
    /* check that the last line is indeed larger than or equal to the
       current line */
    lastLine = relApplyFunction(idRels.famMem2LastLine, currFamMem);
    /* we cannot do: assert(lastLine != RELERR);
       since lastLine may legally be -1, which is == RELERR */
    assert(linCmp(currFirstLine, lastLine) <= 0);

    /* if this family member is not pure, then we have to do nothing */
    if(relIsMember(idRels.nonPureFamilyMembers, currFamMem) == RELOK)
	return;

    /* otherwise, generate a feature chapter body */

    /* The following is somewhat dirty, since we make some implicit
       assumptions about the LaTeX representation of the Z names. */

    /* print sub-chapter heading (memorizing the number of lines generated) */
    lGenerateBody("\\zcollectsection{\\protect\\fvnextisdef\\");
    lGenerateBody(idId2Identifier(currFamMem));
    lGenerateBody("}%\n");
    lGenerateBody("{Collection of all the Requirements}\n");
    lGenerateBody("% \\startcsp\n");
    lGenerateBody("\\begin{zed} \\parents");
    /* print the list of parents, generated from the list of selected 
       non-anonymous features */
       /* TODO!!! */
    firstItem = 1;
    for(relInitSetIt(&setIt,idRels.selectedFeatures);
        relMoreSetIt(&setIt);
	relNextSetIt(&setIt))
    {   if(relIsMember(idRels.anonymousFeatures, relCurrSetIt(&setIt)) == RELOK)
    	    continue;
	if(!firstItem)
	    lGenerateBody(",");
	firstItem = 0;
        lGenerateBody(" \\\\\n    \\t1 \\");
	lGenerateBody(idId2Identifier(relCurrSetIt(&setIt)));
	lGenerateBody("Feature");
    }
    lGenerateBody("\n\\end{zed}\n");
    lGenerateBody("% \\stopcsp\n");
    /* print the collecting class for the system requirements */
    lGenerateBody("\\begin{class}{\\");
    lGenerateBody(idId2Identifier(currFamMem));
    lGenerateBody("Sys}\n");
    if(relCardSet(idRels.selectedFeatures) > 0)
    {   lGenerateBody("    \\inherit");
	firstItem = 1;
	for(relInitSetIt(&setIt,idRels.selectedFeatures);
	    relMoreSetIt(&setIt);
	    relNextSetIt(&setIt))
	{   if(   relIsMember(idRels.anonymousFeatures, relCurrSetIt(&setIt))
	       == RELOK)
		continue;
	    if(!firstItem)
		lGenerateBody(",");
	    firstItem = 0;
	    lGenerateBody(" \\\\\n    \\t1 \\");
	    lGenerateBody(idId2Identifier(relCurrSetIt(&setIt)));
	    lGenerateBody("FeatureSys");
	}
	lGenerateBody("\n    \\endinherit\n");
    }
    lGenerateBody("\\end{class}\n");
    /* print the collecting class for the environment requirements */
    lGenerateBody("\\begin{class}{\\");
    lGenerateBody(idId2Identifier(currFamMem));
    lGenerateBody("Env}\n");
    if(relCardSet(idRels.selectedFeatures) > 0)
    {   lGenerateBody("    \\inherit");
	firstItem = 1;
	for(relInitSetIt(&setIt,idRels.selectedFeatures);
	    relMoreSetIt(&setIt);
	    relNextSetIt(&setIt))
	{   if(   relIsMember(idRels.anonymousFeatures, relCurrSetIt(&setIt))
	       == RELOK)
		continue;
	    if(!firstItem)
		lGenerateBody(",");
	    firstItem = 0;
	    lGenerateBody(" \\\\\n    \\t1 \\");
	    lGenerateBody(idId2Identifier(relCurrSetIt(&setIt)));
	    lGenerateBody("FeatureEnv");
	}
	lGenerateBody("\n    \\endinherit\n");
    }
    lGenerateBody("\\end{class}\n");
    /* print the composition of system and environment */
    lGenerateBody("% \\startcsp\n");
    lGenerateBody("\\begin{class}{\\");
    lGenerateBody(idId2Identifier(currFamMem));
    lGenerateBody("}\n");
    lGenerateBody("    \\inheritinterface \\");
    lGenerateBody(idId2Identifier(currFamMem));
    lGenerateBody("Env \\endinheritinterface \\\\\n");
    lGenerateBody("    \\begin{csp}\n");
    lGenerateBody("        main & = & \\");
    lGenerateBody(idId2Identifier(currFamMem));
    lGenerateBody("Env \\parallel \\");
    lGenerateBody(idId2Identifier(currFamMem));
    lGenerateBody("Sys\n");
    lGenerateBody("    \\end{csp}\n");
    lGenerateBody("\\end{class}\n");
    lGenerateBody("% \\stopcsp\n");

    return;
}

/*-------------------------------------------------------------------------
The functions for external use
-------------------------------------------------------------------------*/

/* call-back function from the lexer to store text to memory from an
  "inherit" or "parents" area */
void
filSaveBuffer(char* text)
{   int len;

    len = strlen(text);
    if(lBufferMaxChar + len > BUFLINELENGTH)
    {   fprintf(stderr,
                "error: input line longer than %i characters. Aborting...\n",
		BUFLINELENGTH);
        exit(EXITRESOURCES);
    }
    strcpy(&lTextBuffer[lBufferMaxLine][lBufferMaxChar], text);
    lBufferMaxChar += len;
    return;
}

/* call-back function from the lexer to switch to next line in an
  "inherit" or "parents" area */
void
filNextBufferLine()
{
    lTextBuffer[lBufferMaxLine][lBufferMaxChar++] = '\n';
    lTextBuffer[lBufferMaxLine++][lBufferMaxChar] = '\0';
    lBufferMaxChar = 0;
    if(lBufferMaxLine >= BUFLINENUM)
    {   fprintf(stderr,
                "error: more input lines in \"inherit\" construct area than %i lines.\nAborting...\n",
		BUFLINENUM);
        exit(EXITRESOURCES);
    }
    return;
}


/* copy the file literally to stdout, except for deleted lines, and
   except for editing of "inherit" areas */
void
filFilterSpec(char* inFileName, char language)
{   
    /* initializations */

    /* open input file for the second pass */
    sptyyin = fopen(inFileName, "r");
    if(!sptyyin)
    {   fprintf(stderr, "%s: failed to open input file '%s' for second pass: ", 
                PROGNAME, inFileName);
        perror("");
	exit(EXITRESOURCES);
    }
    /* initialize current line number */
    filState.currLineNo = 1;
    /* initialize flag about "\endfeatures" */
    filState.endfeaturesFound = 0;
    /* no pending empty lines to delete yet */
    filState.pendingEmptyLinesToDelete = 0;

    /* initialize iterator over the set of line ranges to delete */
    linInitLRangeIt(&lFilState.deleteLinIt, idRels.deletedLineRanges);
    /* determine next line at which we have to change the deletion mode */
    /* (line -1 means end-of-input) */
    if(!linMoreLRangeIt(&lFilState.deleteLinIt))
    {   filState.printing = 1;
        lFilState.nextChangePrintingLine = -1;
    }
    else
    {   lFilState.nextChangePrintingLine =
            linCurrFirstLRangeIt(&lFilState.deleteLinIt);
        /* is the first line untouched? */
        if(lFilState.nextChangePrintingLine > 1)
	{   filState.printing = 1;
	}
	else
	{   filState.printing = 0;
	    lFilState.nextChangePrintingLine = 
	        linCurrLastLRangeIt(&lFilState.deleteLinIt);
	    if(lFilState.nextChangePrintingLine != -1)
	        lFilState.nextChangePrintingLine++;
	}
    }

    /* determine next line at which we have to change the "parents" area mode */
    /* (line -1 means end-of-input) */
    /* initialize iterator over the set of line ranges of "parents" areas */
    linInitLRangeIt(&lFilState.parentsLinIt, idRels.parentsLineRanges);
    /* determine next line at which we have to change the editing mode */
    /* (line -1 means end-of-input) */
    if(!linMoreLRangeIt(&lFilState.parentsLinIt))
        lFilState.nextChangeParentsLine = -1;
    else
    {   lFilState.nextChangeParentsLine =
            linCurrFirstLRangeIt(&lFilState.parentsLinIt);
        /* is the first line untouched? */
        if(lFilState.nextChangeParentsLine > 1)
	{   filState.parentsArea = 0;
	}
	else
	{   filState.parentsArea = 1;
	    lFilState.nextChangeParentsLine = 
	        linCurrLastLRangeIt(&lFilState.parentsLinIt);
	    if(lFilState.nextChangeParentsLine != -1)
	        lFilState.nextChangeParentsLine++;
	}
    }

    /* determine next line at which we have to change the "inherit" area mode */
    /* (line -1 means end-of-input) */
    /* initialize iterator over the set of line ranges of "inherit" areas */
    linInitLRangeIt(&lFilState.inheritLinIt, idRels.inheritLineRanges);
    /* determine next line at which we have to change the editing mode */
    /* (line -1 means end-of-input) */
    if(!linMoreLRangeIt(&lFilState.inheritLinIt))
        lFilState.nextChangeInheritLine = -1;
    else
    {   lFilState.nextChangeInheritLine =
            linCurrFirstLRangeIt(&lFilState.inheritLinIt);
        /* is the first line untouched? */
        if(lFilState.nextChangeInheritLine > 1)
	{   filState.inheritArea = 0;
	}
	else
	{   filState.inheritArea = 1;
	    lFilState.nextChangeInheritLine = 
	        linCurrLastLRangeIt(&lFilState.inheritLinIt);
	    if(lFilState.nextChangeInheritLine != -1)
	        lFilState.nextChangeInheritLine++;
	}
    }


    /* loop over the input stream */

    /* one call to sptyylex() for each input line */
    while(sptyylex())
    {   /* check whether we have to append a generated feature chapter body
           for a family member */
        if(filState.endfeaturesFound)
        {   lCheckGeneratedFamMemBody();
	    filState.endfeaturesFound = 0;
	}
        /* increment current line number */
	if(++filState.currLineNo == lFilState.nextChangeInheritLine)
	{   /* change to/from "inherit" area mode detected */
	    if(!filState.inheritArea)
	    {   /* if we found the begin of the next "inherit" area,
		   take the line after its end as the next goal */
		lFilState.nextChangeInheritLine = 
		    linCurrLastLRangeIt(&lFilState.inheritLinIt);
		if(lFilState.nextChangeInheritLine != -1)
		    lFilState.nextChangeInheritLine++;
		if(filState.parentsArea)
		{   fprintf(stderr,
			    "internal error: overlapping \"inherit\" and \"parents\" areas.\nAborting...\n");
		    exit(EXITRESOURCES);
		}
		/* set up memory to buffer the text in the "inherit" area */
		lInitSaveBuffer();
	    }
	    else
	    {   /* if we found the end of an "inherit" area, 
		   step to next line range, if any */
		linNextLRangeIt(&lFilState.inheritLinIt);
		if(!linMoreLRangeIt(&lFilState.inheritLinIt))
		    lFilState.nextChangeInheritLine = -1;
		else
		    lFilState.nextChangeInheritLine =
		        linCurrFirstLRangeIt(&lFilState.inheritLinIt);
		/* do editing of the inherit area, and print it */
		lEditPrintInherit();
	    }
	    /* finally, invert the "inherit" area flag */
	    filState.inheritArea = !filState.inheritArea;
	}

	if(filState.currLineNo == lFilState.nextChangeParentsLine)
	{   /* change to/from "parents" area mode detected */
	    if(!filState.parentsArea)
	    {   /* if we found the begin of the next "parents" area,
		   take the line after its end as the next goal */
		lFilState.nextChangeParentsLine = 
		    linCurrLastLRangeIt(&lFilState.parentsLinIt);
		if(lFilState.nextChangeParentsLine != -1)
		    lFilState.nextChangeParentsLine++;
		if(filState.inheritArea)
		{   fprintf(stderr,
			    "internal error: overlapping \"inherit\" and \"parents\" areas.\nAborting...\n");
		    exit(EXITRESOURCES);
		}
		/* set up memory to buffer the text in the "parents" area */
		lInitSaveBuffer();
	    }
	    else
	    {   /* if we found the end of a "parents" area, 
		   step to next line range, if any */
		linNextLRangeIt(&lFilState.parentsLinIt);
		if(!linMoreLRangeIt(&lFilState.parentsLinIt))
		    lFilState.nextChangeParentsLine = -1;
		else
		    lFilState.nextChangeParentsLine =
		        linCurrFirstLRangeIt(&lFilState.parentsLinIt);
		/* do editing of the "parents" area, and print it */
		lEditPrintParents();
	    }
	    /* finally, invert the "parents" area flag */
	    filState.parentsArea = !filState.parentsArea;
	}

	if(filState.currLineNo == lFilState.nextChangePrintingLine)
	{   /* change in printing/deleting mode detected */
	    if(filState.printing)
	    {   /* if we found the begin of the next deletion range,
		   take its end as the next goal */
		lFilState.nextChangePrintingLine = 
		    linCurrLastLRangeIt(&lFilState.deleteLinIt);
		if(lFilState.nextChangePrintingLine != -1)
		    lFilState.nextChangePrintingLine++;
		/* check that there is no unfinished "inherit" or "parents"
		   area */
		if(filState.inheritArea || filState.parentsArea)
		{   fprintf(stderr,
			    "internal error: \"inherit\" or \"parents\" area unfinished at start of\ndeletion range. Aborting...\n");
		    exit(EXITRESOURCES);
		}
	    }
	    else
	    {   /* if we found the end of a deletion range, 
		   step to next line range, if any */
		linNextLRangeIt(&lFilState.deleteLinIt);
		if(!linMoreLRangeIt(&lFilState.deleteLinIt))
		    lFilState.nextChangePrintingLine = -1;
		else
		    lFilState.nextChangePrintingLine =
		        linCurrFirstLRangeIt(&lFilState.deleteLinIt);
	    }
	    /* finally, invert the "printing" flag */
	    filState.printing = !filState.printing;
	}
    }
    if(filState.inheritArea)
    {   /* if still buffered text pending for editing, */
	/* do editing of the inherit area, and print it */
	lEditPrintInherit();
    }
}

