/*-------------------------------------------------------------------------

project: genFamMem - maintaining families of requirements

author: Jan Bredereke
copyright: Jan Bredereke, 2000
$Id: firstpass.y,v 1.51 2001/02/15 08:59:52 brederek Exp $

file: firstpass.y - bison grammar for the first pass

-------------------------------------------------------------------------*/

%{
#include <stdio.h>
#include <assert.h>
#include "global.h"
#include "genidentifier.h"

/* we want to have more specific error messages than "parse error" */
#define YYERROR_VERBOSE

/* use the count of line numbers from lexical analysis */
extern int fplyylineno;

/* check whether the input language is not Z */
#define checkNotZ() \
{ \
    if(progOpts.language == LANGUAGE_Z) \
    {   fplyyerror("syntax error: language construct of CSP_Z or CSP-OZ not allowed in pure Z"); \
        YYERROR; \
	yyerrok; \
    } \
}

/* check whether the input language is CSP-OZ */
#define checkCspoz() \
{ \
    if(progOpts.language != LANGUAGE_CSPOZ) \
    {   fplyyerror("syntax error: language construct of CSP-OZ not allowed in CSP_Z or pure Z"); \
        YYERROR; \
	yyerrok; \
    } \
}

/* make the line number of the currently lexed token available (for error
   messages) */
extern int mylineno;

/* define some state variables handling context-sensitivity */
static int currFeature = IDERR;		/* current feature, if any */
static int currFamilyMember = IDERR;	/* current family member, if any */
	/* Please note that (currFeature != IDERR) may hold when
	   (currFamilyMember != IDERR). This is the case when a family
	   member has a feature chapter body. These are anonymous features. */
static char inDefaultProperties = 0;	/* 1: in "default_properties"
					   0: in "inherit" */
static char inDefaultSection = 0;	/* 1: in "default_section"
					   0: otherwise */
static char inCollectSection = 0;	/* 1: in "collect_section"
					   0: otherwise */
static int currSection = IDERR;		/* current section or default section 
					   or collect section, if any */
static int currSectionFirstLine = 0;	/* first line of current 
                                           (default/collect) section, if any */
static int implicitSectionClass = 0;	/* 1: the current class also starts
					      implicitly a section 
					   0: otherwise */
static int currClass = IDERR;		/* current class, if any */

/* auxiliary variables which cannot be declared locally */
static int retCode; /* return code */
#define ERRSTRINGLEN 1000
static char errStringBuf[ERRSTRINGLEN+1]; /* buffer for error string */

%}

	/* expect 4 shift/reduce conflicts. They are due to the ambiguity
	   whether a class in a section could be interpreted as an end
	   of that section and the start of a new (implicit) section.
	   Bison's default behaviour of choosing shift over reduce prevents 
	   this. */
%expect 4

	/* the tokens */

	/* Z tokens */
	/* **************************************************************
	   IF YOU ADD HERE, YOU PROBABLY HAVE TO ADD AT inParagraphStuff 
	   BELOW, TOO! 
	   **************************************************************
	*/
%nonassoc FEATURE FAMILYMEMBER FEATURES REMOVE DEFAULT_PROPERTIES
%right FEATURE_END FAMILYMEMBER_END
%nonassoc SECTION SECTION_END DEFAULT_SECTION COLLECT_SECTION PARENTS
%nonassoc CLASS
	/* CHAN should have low precedence for "interfaces" */
%nonassoc CHAN 
	/* RENAME should have low precedence for "renamingOpt" */
%nonassoc RENAME
%nonassoc AX GENAX SCH GENSCH END ENDCLASS
%left INHERIT INHERITINTERFACE
%nonassoc VISIBILITY
%nonassoc CDEF
%nonassoc NL
%nonassoc LPAREN RPAREN LBRACKET RBRACKET LBRACE RBRACE
%left COLON COMMA SEMI
%nonassoc NUMBER WORD WORDSTROKE
%nonassoc STROKE
%nonassoc STUFF

	/* the start symbol */

	/* %start specification */
%start family

%%

/* the grammar rules */

	/* specification:	  /* empty * /  */
	/* 		| sections */
	/* 		| paragraphs */
	/* 		; */

family:		  /* empty */
		| family chapter
		; 

chapter:	  featureChapter
		| familymemberChapter
		;

featureChapter:	  FEATURE name
                  { currFeature = $2; 
		    assert(currFamilyMember == IDERR);
		    retCode = relAddMember(idRels.features, currFeature);
		    if(retCode == RELCANTDO)
		    {   snprintf(errStringBuf, ERRSTRINGLEN,
		                "syntax error: redefinition of feature %s",
		                idId2Identifier(currFeature));
		        fplyyerror(errStringBuf);
			YYERROR;
			yyerrok;
		    }
		    else
		    {   retCode = relAddRelship(idRels.feature2FirstLine, 
						currFeature, 
						@1.first_line);
			assert(retCode == RELOK);
		    }
		  } featureChapter2
		| FEATURE error
		;

featureChapter2:  featureChapter3
                  { relAddRelship(idRels.feature2LastLine, 
		                  currFeature,
				  @1.last_line);
		    currFeature = IDERR;
		  }
		;

featureChapter3:  /* empty */
		| featureChapterBody
		| FeatureChapterEnd
		| featureChapterBody FeatureChapterEnd
		;

featureChapterBody: 
		  sectionsOrRemoves
		;

FeatureChapterEnd:
		  FEATURE_END name
		  {
		    assert(currFeature != IDERR);
		    if($2 != currFeature)
		    {   retCode = relIsMember(idRels.features, $2);
		    	if(retCode == RELOK)
			{   snprintf(errStringBuf, ERRSTRINGLEN,
			   "syntax error: terminating feature %s instead of %s",
				   idId2Identifier($2),
				   idId2Identifier(currFeature));
			    fplyyerror(errStringBuf);
			    YYERROR;
			    yyerrok;
			}
			else
			{   snprintf(errStringBuf, ERRSTRINGLEN,
			"syntax error: unknown feature %s; expected feature %s",
				   idId2Identifier($2),
				   idId2Identifier(currFeature));
			    fplyyerror(errStringBuf);
			    YYERROR;
			    yyerrok;
			}
		    }
		    retCode = relAddMember(idRels.featureEnds, currFeature);
		    /* The retCode might be RELCANTDO in case this is a
		       doubly defined feature. Since we have printed
		       an error message on this before, we do nothing 
		       here. */
		  }
		;

familymemberChapter:
		  FAMILYMEMBER name
                  { currFamilyMember = $2; 
		    assert(currFeature == IDERR);
		    retCode = relAddMember(idRels.familyMembers,
		    		           currFamilyMember);
		    if(retCode == RELCANTDO)
		    {   snprintf(errStringBuf, ERRSTRINGLEN,
		               "syntax error: redefinition of family member %s",
			       idId2Identifier(currFamilyMember));
		        fplyyerror(errStringBuf);
			YYERROR;
			yyerrok;
		    }
		    else
		    {   retCode = relAddRelship(idRels.famMem2FirstLine, 
						currFamilyMember, 
						@1.first_line);
			assert(retCode == RELOK);
		    }
		  } featuresList END
		  { assert(currFamilyMember != IDERR);
		    assert(currFeature == IDERR);
		    /* generate a name for this anonymous feature, even
		       though we might not use the name because the
		       feature could be empty: */
		    currFeature = gidGenUniqueIdentifier(
		    	idId2Identifier(currFamilyMember));
		  }
		    familymemberChapter2
                  { currFeature = IDERR; 
		    relAddRelship(idRels.famMem2LastLine, 
		                  currFamilyMember,
				  @7.last_line);
                    currFamilyMember = IDERR; 
		  }
		| FAMILYMEMBER error
		;

familymemberChapter2: 
		  /* empty */
		| familymemberChapterEnd
		| featureChapterBodyInFamilymemberChapter
		  { /* mark this family member as non-pure */
		    assert(currFamilyMember != IDERR);
		    retCode = relAddMember(idRels.nonPureFamilyMembers,
		    		           currFamilyMember);
		    /* add the generated feature name to the list of features */
		    retCode = relAddMember(idRels.features, currFeature);
		    assert(retCode == RELOK);
		    /* mark it as an anonymous feature */
		    retCode = relAddMember(idRels.anonymousFeatures,
		                           currFeature);
		    assert(retCode == RELOK);
		    /* add the name to the list of features of this family
		       member */
		    retCode = relAddRelship(idRels.famMem2FeatureList, 
					    currFamilyMember,
					    currFeature);
		    assert(retCode == RELOK);
		    /* This anonymous feature does not have any meaningful
		       first line, so set it to zero: */
		    retCode = relAddRelship(idRels.feature2FirstLine, 
					    currFeature, 
					    0);
		    assert(retCode == RELOK);
		    /* The last line of an anonymous *is* meaningful, though,
		       so we set it appropriately: */
                    relAddRelship(idRels.feature2LastLine, 
		                  currFeature,
				  @1.last_line);
		  }
		;

featureChapterBodyInFamilymemberChapter:
		  featureChapterBody
		| featureChapterBody familymemberChapterAndFeatChEnd
		;

familymemberChapterEnd:
		  FAMILYMEMBER_END
		  addFamilyMemberEnd
		;

familymemberChapterAndFeatChEnd:
		  FAMILYMEMBER_END
		  addFamilyMemberEnd
		  { assert(currFeature != IDERR);
		    retCode = relAddMember(idRels.featureEnds, currFeature);
		    /* The retCode might be RELCANTDO in case this is a
		       doubly defined family member. Since we have printed
		       an error message on this before, we do nothing 
		       here. */
		  }
		;

addFamilyMemberEnd: name
		  {
		    assert(currFamilyMember != IDERR);
		    if($1 != currFamilyMember)
		    {   retCode = relIsMember(idRels.familyMembers, $1);
		    	if(retCode == RELOK)
			{   snprintf(errStringBuf, ERRSTRINGLEN,
		     "syntax error: terminating family member %s instead of %s",
				   idId2Identifier($1),
				   idId2Identifier(currFamilyMember));
			    fplyyerror(errStringBuf);
			    YYERROR;
			    yyerrok;
			}
			else
			{   snprintf(errStringBuf, ERRSTRINGLEN,
	   "syntax error: unknown family member %s; expected family member %s",
				   idId2Identifier($1),
				   idId2Identifier(currFamilyMember));
			    fplyyerror(errStringBuf);
			    YYERROR;
			    yyerrok;
			}
		    }
		    retCode = relAddMember(idRels.familyMemberEnds, 
		                           currFamilyMember);
		    /* The retCode might be RELCANTDO in case this is a
		       doubly defined family member. Since we have printed
		       an error message on this before, we do nothing 
		       here. */
		  }
		;

featuresList:	  FEATURES
		| FEATURES featuresListName featuresList2
		| FEATURES error
		;

featuresList2:	  /* empty */
		| featuresList2 COMMA featuresListName
		;

featuresListName: name
		  {   assert(currFamilyMember != IDERR);
		      retCode = relAddRelship(idRels.famMem2FeatureList, 
					      currFamilyMember,
					      $1);
		      if(retCode == RELCANTDO)
		      {   snprintf(errStringBuf, ERRSTRINGLEN,
		                  "syntax error: feature %s appears twice in feature list of family member %s",
				  idId2Identifier($1),
				  idId2Identifier(currFamilyMember));
		          fplyyerror(errStringBuf);
			  YYERROR;
			  yyerrok;
		      }
		  }
		;

paragraphs:	  paragraphs paragraph
		| paragraph
		;

section:	  sectionKeyword name
		  { assert(currSection == IDERR);
		    currSection = $2;
		    implicitSectionClass = 0;
		  } addSectionName 
		  parentsOrClass
		| sectionKeyword error
		| sectionClass
		;

parentsOrClass:   parentsList END section2
		| CLASS { checkCspoz(); } WORD
		  { assert(currSection != IDERR);
		    assert(currClass == IDERR);
		    currClass = $3;
		    if(currSection != currClass)
		    {   snprintf(errStringBuf, ERRSTRINGLEN,
		            "syntax error: no parents specified for section %s, \n\tand no following class with the section's name",
			    idId2Identifier(currSection));
			fplyyerror(errStringBuf);
			YYERROR;
			yyerrok;
		    }
		    relAddRelship3(idRels.sectionClass2FirstLine, 
				   currSection, 
				   currClass,
				   @1.first_line);
		    implicitSectionClass = 1;
		  }
		  class2 
		  { implicitSectionClass = 0;
		  }
		  section2
		;

sectionClass:	  CLASS { checkCspoz(); } WORD
		  { assert(currSection == IDERR);
		    assert(currClass == IDERR);
		    currClass = $3;
		    currSection = currClass;
		    relAddRelship3(idRels.sectionClass2FirstLine, 
				   currSection, 
				   currClass,
				   @1.first_line);
		    inDefaultSection = 0;
		    inCollectSection = 0;
		    currSectionFirstLine = @1.first_line;
		    implicitSectionClass = 1;
		  } addSectionName 
		  class2 
		  { implicitSectionClass = 0;
		  }
		  section2
		| CLASS error
		;

sectionKeyword:   SECTION 
		  { inDefaultSection = 0;
		    inCollectSection = 0;
		    currSectionFirstLine = @1.first_line;
		  }
		| DEFAULT_SECTION
		  { inDefaultSection = 1;
		    inCollectSection = 0;
		    currSectionFirstLine = @1.first_line;
		  }
		| COLLECT_SECTION
		  { inDefaultSection = 0;
		    inCollectSection = 1;
		    currSectionFirstLine = @1.first_line;
		  }
		;

addSectionName:   /* empty */
		  { assert(currFeature != IDERR);
		    assert(currSection != IDERR);
		    assert(currSectionFirstLine);
		    retCode = relAddMember(idRels.sections, currSection);
		    if(retCode == RELCANTDO)
		    {   snprintf(errStringBuf, ERRSTRINGLEN,
		            "syntax error: section %s redefined",
			    idId2Identifier(currSection));
			fplyyerror(errStringBuf);
			YYERROR;
			yyerrok;
		    }
		    else
		    {   retCode = relAddRelship(idRels.feature2Section,
						currFeature, 
						currSection);
			assert(retCode == RELOK);
		        if(inDefaultSection)
		        {   assert(!inCollectSection);
			    retCode = relAddMember(idRels.defaultSections,
						   currSection);
			    assert(retCode == RELOK);
			}
		        else if(inCollectSection)
		        {   retCode = relAddMember(idRels.collectSections,
						   currSection);
			    assert(retCode == RELOK);
			}
		        retCode = relAddRelship(idRels.section2FirstLine, 
						currSection, 
						currSectionFirstLine);
			assert(retCode == RELOK);
		    }
		  }
		;

section2:	  section3
                  { relAddRelship(idRels.section2LastLine, 
		                  currSection,
				  @1.last_line);
		    currSection = IDERR;
		  }
		;

section3:	  /* empty */
		| sectionEnd
		| paragraphs
		| paragraphs sectionEnd
		;

sectionEnd:	  SECTION_END name
		  {
		    assert(currSection != IDERR);
		    assert(currFeature != IDERR);
		    if($2 != currSection)
		    {   retCode = relIsMember(idRels.sections, $2);
		    	if(retCode == RELOK)
			{   snprintf(errStringBuf, ERRSTRINGLEN,
			   "syntax error: terminating section %s instead of %s",
				   idId2Identifier($2),
				   idId2Identifier(currSection));
			    fplyyerror(errStringBuf);
			    YYERROR;
			    yyerrok;
			}
			else
			{   snprintf(errStringBuf, ERRSTRINGLEN,
		        "syntax error: unknown section %s; expected section %s",
				   idId2Identifier($2),
				   idId2Identifier(currSection));
			    fplyyerror(errStringBuf);
			    YYERROR;
			    yyerrok;
			}
		    }
		    retCode = relAddMember(idRels.sectionEnds, 
		    			   currSection);
		    /* The retCode might be RELCANTDO in case this is a
		       doubly defined section. Since we have printed
		       an error message on this before, we do nothing 
		       here. */
		  }
		;

parentsList:	  parentsList2
		  { assert(currSection != IDERR);
		    if(@1.first_line <= @1.last_line)
		    {
		      retCode = relAddRelship(idRels.sectionParents2FirstLine,
					      currSection, 
					      @1.first_line);
		      assert(retCode == RELOK);
		      retCode = relAddRelship(idRels.sectionParents2LastLine,
					      currSection, 
					      @1.last_line);
		      assert(retCode == RELOK);
		    }
		  }
		;

parentsList2:	  PARENTS
		| PARENTS parentsListName parentsList3
		| PARENTS error
		;

parentsList3:	  /* empty */
		| parentsList3 COMMA parentsListName
		;

parentsListName:  name
		  { assert(currFeature != IDERR);
		    assert(currSection != IDERR);
		    retCode = relAddRelship(
			idRels.sectionParents,
			currSection, 
			$1);
		    if(retCode == RELCANTDO)
		    {   snprintf(errStringBuf, ERRSTRINGLEN,
			  "syntax error: parent %s appears twice in section %s",
			    idId2Identifier($1),
			    idId2Identifier(currSection));
			fplyyerror(errStringBuf);
			YYERROR;
			yyerrok;
		    }
		  }
		;

sectionsOrRemoves:
		  sectionOrRemove
		| sectionsOrRemoves sectionOrRemove
		;

sectionOrRemove:
		  section
		| remove
		;

paragraph:	  paragraph_o
		;

paragraph_o:	  paragraph_c
		| class
		;

paragraph_c:	  paragraph_z
		| channel
		| process
		;

remove:		  REMOVE END
		| REMOVE removeName remove2 END
		| REMOVE error END
		;

remove2:	  /* empty */
		| remove2 COMMA removeName
		;

removeName:	  name
		  { assert(currFeature != IDERR);
		    assert(currFamilyMember == IDERR);
		    retCode = relAddRelship(idRels.remove, currFeature, $1);
		    if(retCode == RELCANTDO)
		    {   snprintf(errStringBuf, ERRSTRINGLEN,
			    "syntax error: '%s' in feature %s is removed twice",
				$1,
				idId2Identifier(currFeature));
		        fplyyerror(errStringBuf);
			YYERROR;
			yyerrok;
		    }
		  }
		;

name:		  WORD { $$ = $1; }
		| WORDSTROKE { $$ = $1; }
		;

paragraph_z:	  inParagraphStuffs END
		;

inParagraphStuffs:
		  inParagraphStuffs inParagraphStuff
		| inParagraphStuff
		;

		/* this matches *all* tokens in paragraphs except END, 
		   SECTION, PARENTS, CDEF, INHERIT, and INHERITINTERFACE: */
inParagraphStuff:  
		  noBracketStuff
		| LBRACKET
		| RBRACKET
		;

	/* channel:	  CHAN declNameList END */
	/* 		  CHAN declNameList COLON inParagraphStuffs END */
	/* 		; */

channel:	  CHAN { checkNotZ(); } inParagraphStuffs END
		| CHAN error END
		;

process:	  name CDEF { checkNotZ(); } inParagraphStuffs END  ;

		  /* "class" is both "class_c" and "class_o" */
class:		  CLASS { checkCspoz(); } WORD
		  { assert(currClass == IDERR);
		    currClass = $3;
		    assert(currSection != IDERR);
		    relAddRelship3(idRels.sectionClass2FirstLine, 
				   currSection, 
				   currClass,
				   @1.first_line);
		  }
		  class2
		| CLASS error
		;

class2:		  { assert(currSection != IDERR);
		    assert(currClass != IDERR);
		    retCode = relAddRelship(
			idRels.section2Class,
			currSection, 
			currClass);
		    if(retCode == RELCANTDO)
		    {   snprintf(errStringBuf, ERRSTRINGLEN,
			  "syntax error: class %s appears twice in section %s",
			    idId2Identifier(currClass),
			    idId2Identifier(currSection));
			fplyyerror(errStringBuf);
			YYERROR;
			yyerrok;
		    }
		  }
		  bracketParamOpt parenParamOpt 
		  visibilityListOrInterfaces inheritedClasses inClassStuffs
		  ENDCLASS
		  { assert(currSection != IDERR);
		    assert(currClass != IDERR);
		    relAddRelship3(idRels.sectionClass2LastLine, 
				   currSection, 
				   currClass,
				   @7.last_line);
		    currClass = IDERR;
		  }
		;

inClassStuffs:	  /* empty */
		| inClassStuffs inClassStuff
		;

inClassStuff:	  inParagraphStuff
		| END
		| CDEF
		;

bracketParamOpt:  /* empty */ %prec RENAME
		| LBRACKET noBracketStuffs bracketParamOpt noBracketStuffs 
		  RBRACKET
		| LBRACKET error RBRACKET
		;

parenParamOpt:	  /* empty */ %prec RENAME
		| parenParam
		;

parenParam:	  LPAREN noParenStuffs parenParamOpt noParenStuffs 
		  RPAREN
		| LPAREN error RPAREN
		;

noBracketStuffs:  /* empty */
		| noBracketStuffs noBracketStuff
		;

noParenStuffs:	  /* empty */
		| noParenStuffs noParenStuff
		;

noBracketStuff:   noBracketNoParenStuff
		| LPAREN
		| RPAREN
		;

noParenStuff:   noBracketNoParenStuff
		| LBRACKET
		| RBRACKET
		;

noBracketNoParenStuff:
		  AX
		| GENAX
		| SCH
		| GENSCH
		| NL
		| NUMBER
		| LBRACE
		| RBRACE
		| STROKE
		| WORD
		| WORDSTROKE
		| COLON
		| COMMA
		| SEMI
		| RENAME
		| STUFF
		;

visibilityListOrInterfaces:
		  /* empty */
		| visibilityList
		| interfaces
		;

visibilityList:	  VISIBILITY parenParam 
		| VISIBILITY error
		;

interfaces:	  interface
		| interfaces interface
		;

		  /* CHAN denotes all kinds of chan*, method*, local_chan* */
interface:	  CHAN inParagraphStuffs
		| CHAN error
		;

inheritedClasses: inheritedClasses2
		  { assert(currSection != IDERR);
		    assert(currClass != IDERR);
		    if(@1.first_line <= @1.last_line)
		    {
		      retCode = relAddRelship3(
		      			idRels.sectionClassInherit2FirstLine,
					currSection, 
					currClass, 
					@1.first_line);
		      assert(retCode == RELOK);
		      retCode = relAddRelship3(
		      			idRels.sectionClassInherit2LastLine,
					currSection, 
					currClass, 
					@1.last_line);
		      assert(retCode == RELOK);
		    }
		  }
		;

inheritedClasses2:   /* empty */
		| inheritedClasses2 inheritedClass
		;

inheritedClass:
		  inheritedClass_o
		| DEFAULT_PROPERTIES { inDefaultProperties = 1; }
		  inheritedClassNames
		| DEFAULT_PROPERTIES error
		;

		  /* Note: the keyword is mandatory, contrary to Fischer's
		     definition, but at least he recommends it anyway . */
		  /* Extension to Fischer: more than one item is allowed, 
		     separated by commas. */
inheritedClass_o:
		  INHERIT { inDefaultProperties = 0; } 
		  inheritedClassNames
		| INHERITINTERFACE { inDefaultProperties = 0; } 
		  inheritedClassNames
		| INHERIT error
		| INHERITINTERFACE error
		;

inheritedClassNames:
		  inheritedClassNames2 END
		;

inheritedClassNames2:
		  inheritedClassName inheritedClassNames3
		;

inheritedClassNames3:
		  /* empty */
		| COMMA inheritedClassNames2
		| error
		;

inheritedClassName:
		  WORD
		  { assert(currFeature != IDERR);
		    assert(currSection != IDERR);
		    assert(currClass != IDERR);
		    retCode = relAddRelship3(
			idRels.sectionClass2Inherits,
			currSection, 
			currClass, 
			$1);
		    if(retCode == RELCANTDO)
		    {   snprintf(errStringBuf, ERRSTRINGLEN,
			     "syntax error: %s class %s is inherited twice in class %s of section %s of %s %s",
				(inDefaultProperties ? 
				    "default_properties" : 
				    "inherited"),
				idId2Identifier($1),
				idId2Identifier(currClass),
				idId2Identifier(currSection),
				((currFamilyMember != IDERR) ? 
				    "family member" :
				    "feature"),
				idId2Identifier(
				    ((currFamilyMember != IDERR) ? 
				    currFamilyMember :
				    currFeature)));
			fplyyerror(errStringBuf);
			YYERROR;
			yyerrok;
		    }
		    if(inDefaultProperties)
		    {   retCode = relAddRelship3(
			    idRels.sectionClass2DefaultProperties,
			    currSection, 
			    currClass, 
			    $1);
		    }
		    if(implicitSectionClass)
		    { relAddRelship(
			  idRels.sectionParents,
			  currSection, 
			  $1);
		    }
		  } bracketParamOpt parenParamOpt renamingOpt
		;

renamingOpt:	  /* empty */ %prec CHAN
		| RENAME inParagraphStuffs
		| RENAME error
		;

%%

/* error printing function of bison: */
int
fplyyerror(char *str)
{   /* make sure that error string is zero-terminated, even if snprintf()
       truncated it: */
    errStringBuf[ERRSTRINGLEN] = '\0';
    /* print error message ... */
    fprintf(stderr, "%s:%i: ", progOpts.inFileName, mylineno);
    /* ... and error position, if available */
    if((yychar != YYEMPTY)&&(yychar>=258))
	fprintf(stderr, "at token '%s': ", yytname[yychar-255]);
    fprintf(stderr, "%s\n", str);
}

