/*-------------------------------------------------------------------------

project: genFamMem - maintaining families of requirements

author: Jan Bredereke
copyright: Jan Bredereke, 2000
$Id: firstpasslex.l,v 1.21 2001/01/29 20:16:06 brederek Exp $

file: firstpasslex.l - flex rules for the lexical analysis of the first pass

-------------------------------------------------------------------------*/

%{

#include <assert.h>
#include "global.h"
#include "identifier.h"
#include "firstpass.tab.h"

/* macro that copies the string in yytext to bison's fplyylval */ 
#define yytext2yylval() { fplyylval = idIdentifier2Id(yytext); }

/* initialize the line numbers of the current token */
extern YYLTYPE yylloc;
#define YY_USER_INIT { yylloc.first_line = 1; yylloc.last_line = 1; }
/* do my own book-keeping of the currently lexed line number, since it
   works much more precise than flex's built-in yylineno. */
int mylineno = 1;

%}

	/* there is one input file only */
%option noyywrap
	/* make yytext a pointer, such that we are allowed to modify it */
%pointer
	/* if no rule matches, print error instead of printing on stdout */
%option nodefault
	/* do not use lex's fread(), but read character-wise, until we
	   find a newline. The input stream must be terminated by a
	   newline, too, since the usual end-of-file mechanisms appear to 
	   not work on pipes. */
%option always-interactive

	/* definition of Z characters */

	/* Please note that this definition is incomplete, since our 
	   mark-up recognizes only those Z characters that we actually
	   use. All other stuff remains un-interpreted ASCII text. */

	/* The mark-up phase translates multi-character sequences into
	   Z characters. The Z standard defines Unicode codes for these.
	   In order to be able to process Unicode characters with
	   the 8-bit program flex, we represent them by an escape
	   character (\200) followed by the four hex digits of their
	   code. As an exception, we represent Unicode codes that
	   fit into 7 bits by their original code with no escape. */
UC		\200

AXCHAR		({UC}"2577")
SCHCHAR		({UC}"250C")
GENCHAR		({UC}"2550")
ENDCHAR		({UC}"2029")
NLCHAR		({UC}"2028")
CAPITALDELTA	({UC}"0394")
CAPITALXI	({UC}"039E")
THETA		({UC}"03B8")
LAMBDA		({UC}"03BB")
MU		({UC}"03BC")

DOWN		({UC}"2198")
ENDDOWN		({UC}"2196")

	/* additional characters for CSP-Z,... */
CLASSCHAR	({UC}"class")
CDEFCHAR	({UC}"cdef")
VISIBILITYCHAR	({UC}"visib")
	/* this really should be ENDCHAR, but we want to find the end of
	   a class despite that we don't want to parse the entire class: */
ENDCLASSCHAR	({UC}"endclass")

	/* end-of-input marker */
EOI		({UC}"EOI\n")

DIGIT		[0-9]
LATIN		[A-Za-z]
GREEK		({CAPITALDELTA}|{CAPITALXI}|{THETA}|{LAMBDA}|{MU})
LETTER		({LATIN}|{GREEK})
  /* LETTER	({LATIN}|{GREEK}|{OTHERLETTER}) */
STROKECHAR	("'"|"!"|"?")
WORDGLUE	"_"
  /* WORDGLUE	("_"|...) */
BRACKET		("("|")"|"["|"]"|"{"|"}")
  /* BRACKET	("("|")"|"["|"]"|"{"|"}"|...) */
BOXCHAR		({AXCHAR}|{SCHCHAR}|{GENCHAR}|{ENDCHAR}|{ENDCLASSCHAR}|{NLCHAR})
SPACE		" "
SPECIAL		({STROKECHAR}|{WORDGLUE}|{BRACKET}|{BOXCHAR}|{SPACE})
/* SYMBOL ... */

ALPHASTR	(({LETTER}|{DIGIT})*)
WORDPART	({WORDGLUE}{ALPHASTR})
  /* WORDPART	({WORDGLUE}({ALPHASTR}|({SYMBOL}*))) */


	/* our handling of word boundaries, space, etc. */
NONWORDCHAR	([^A-Za-z0-9_])
NWC		{NONWORDCHAR}
NONDIGITCHAR	([^0-9])
NDC		{NONDIGITCHAR}

%%

	/* the rules */

	/* alphabetical keywords */

"feature"/{NWC}		return(FEATURE);
"feature_end"/{NWC}	return(FEATURE_END);
"familymember"/{NWC}	return(FAMILYMEMBER);
"familymember_end"/{NWC} return(FAMILYMEMBER_END);
"section_end"/{NWC}	return(SECTION_END);
"features"/{NWC}	return(FEATURES);
"remove"/{NWC}		return(REMOVE);
"default_properties"/{NWC}	return(DEFAULT_PROPERTIES);
"inherit-interface"/{NWC}	return(INHERITINTERFACE);
"inherit"/{NWC}		return(INHERIT);
"parents"/{NWC}		return(PARENTS);
"section"/{NWC}		return(SECTION);
"default_section"/{NWC}	return(DEFAULT_SECTION);
"collect_section"/{NWC}	return(COLLECT_SECTION);
"chan"/{NWC}		return(CHAN);
  /* ... */

  /*
  INHERIT
  INHERITINTERFACE
  */

	/* symbolic keywords */

":"			return(COLON);
","			return(COMMA);
"/"			return(RENAME);
";"			return(SEMI);
  /* ... */

	/* further tokens from lexis */

{AXCHAR}{GENCHAR}	return(GENAX);
{AXCHAR}		return(AX);
{SCHCHAR}		return(SCH);
{SCHCHAR}{GENCHAR}	return(GENSCH);
{ENDCHAR}		return(END);
{ENDCLASSCHAR}		return(ENDCLASS);
  /* ... */

	/* Newlines are either soft or hard, see Z standard Sect. 7.5.
	   As far as we are concerned, hard newlines are not interesting.
	   Therefore we don't implement the algorithm that decides on
	   the context whether a newline is hard or soft. We just make
	   *all* newlines soft. This means that they serve as a token
	   separator only, but are not passed on to Bison. */
{NLCHAR}		{ /* empty */
			/* DON'T return() here! Bison doesn't want to know. */
			/* For a hard newline, we would do: return(NL); */
			}

	/* additional character for CSP-Z,... */
{CDEFCHAR}		return(CDEF);
{CLASSCHAR}		return(CLASS);
{VISIBILITYCHAR}	return(VISIBILITY);

{DIGIT}+/{NDC}		return(NUMBER);

"("			return(LPAREN);
")"			return(RPAREN);
"["			return(LBRACKET);
"]"			return(RBRACKET);
"{"			return(LBRACE);
"}"			return(RBRACE);

{STROKECHAR}		return(STROKE);
{DOWN}{DIGIT}{ENDDOWN}	return(STROKE);

	/* WORDSTROKE needs to be split into two parts in order to have
	   no variable first part */
{WORDPART}+{STROKECHAR}+/{NWC}	{
			yytext2yylval();
			return(WORDSTROKE);
			}

{LETTER}{ALPHASTR}{WORDPART}*{STROKECHAR}+/{NWC}	{
			yytext2yylval();
			return(WORDSTROKE);
			}

	/* WORD needs to be split into two parts in order to have
	   no variable first part */
{WORDPART}+/{NWC}	{
			yytext2yylval();
			return(WORD);
			}

  /* (({LETTER}{ALPHASTR})|({SYMBOL}+))({WORDPART}*)/{NWC} ... */
{LETTER}{ALPHASTR}{WORDPART}*/{NWC}	{
			yytext2yylval();
			return(WORD);
			}

		/* keep current line number count in yylloc and mylineno */
"\n"			{
			++yylloc.first_line;
			++yylloc.last_line; 
			++mylineno;
			/* DON'T return() here! Bison doesn't want to know. */
			}

" "			{ /* empty */
			/* DON'T return() here! Bison doesn't want to know. */
			}

		/* This marks our end-of-input, since the usual
		   mechanisms for end-of-file appear to not work */
{EOI}			return(0);

		/* make sure that all escape sequences have been recognized */
{UC}			{ assert(0); }

		/* do not lex the rest of the text */

.			return(STUFF);

%%

