/*-------------------------------------------------------------------------

project: genFamMem - maintaining families of requirements

author: Jan Bredereke
copyright: Jan Bredereke, 2000
$Id: firstpasstex.l,v 1.24 2001/01/29 20:16:06 brederek Exp $

file: firstpasstex.l - flex rules for the LaTeX mark-up of the first pass

-------------------------------------------------------------------------*/

%{

#include "global.h"

	/* The mark-up phase translates multi-character sequences into
	   Z characters. The Z standard defines Unicode codes for these.
	   In order to be able to process Unicode characters with
	   the 8-bit program flex, we represent them by an escape
	   character (\200) followed by the four hex digits of their
	   code. As an exception, we represent Unicode codes that
	   fit into 7 bits by their original code with no escape. 
	   Furthermore, we preserve any newline characters, such that
	   the lexical analysis is able to keep track of the input
	   line count. They need to be treated like ' ' otherwise. */

#define UC '\200'

#define myputc(chr)	fputc(chr, yyout)
#define myputs(str)	fputs(str, yyout)

static int stackEntries = 0; /* current number of entries on start condition
				stack */
#define my_push_state(state)	{ yy_push_state(state); stackEntries++; }
#define my_pop_state()		{ if(stackEntries-->0){ yy_pop_state();} \
  else{ fprintf(stderr,"%s: mark-up error: too many closing Z environments.\n",\
  PROGNAME); exit(EXITSYNERR);} }
#define my_top_state()		yy_top_state()
#define my_stack_empty()	(stackEntries<=0)
#define my_check_stack_empty()	{ if(!my_stack_empty()){ \
  fprintf(stderr, "%s: mark-up error: closing Z environment missing.\n", \
  PROGNAME);} }

static int missingRBraces = -1; 
			   /* In order to mark-up the names of features and
			      of family members, we must skip briefly into
			      the Z text mode for them. A value of >= 0
			      indicates this mode. The actual number 
			      indicates the right braces still missing to
			      end the mode. The value -1 indicates that
			      this mode is not active. */

static int commentLBraces = 0;
		/* In order to delete braced comments in the Z mode,
		   we briefly skip into IN_BRACEDCOMMENT mode and
		   start counting the braces. */

%}

	/* there is one input file only */
%option noyywrap
	/* make yytext a pointer, such that we are allowed to modify it */
%pointer
	/* if no rule matches, print error instead of printing on stdout */
%option nodefault
	/* use stack of start conditions */
%option stack

	/* start conditions for the rules */
%s IN_FORMAL OUTOF_FORMAL IN_BRACEDCOMMENT IN_CLASSCOMMENT

LATEXTOKEN	("\\"[A-Za-z]+)
NONALPHA	([^A-Za-z])
NA		{NONALPHA}
WHITESPACE	([ \t])
WS		{WHITESPACE}

%%

	/* the rules */

	/* start outside of Z text */
	BEGIN(OUTOF_FORMAL);

	/* rules when outside Z text */

<OUTOF_FORMAL>{

"\\feat"/{NA}		{ 
			  myputs(" feature ");
			  /* go briefly into Z text mode to get name */
			  missingRBraces = 0;
			  my_push_state(IN_FORMAL);
			}
"\\familymember"/{NA}	{ 
			  myputs(" familymember ");
			  /* go briefly into Z text mode to get name */
			  missingRBraces = 0;
			  my_push_state(IN_FORMAL);
			}

"\\begin{zed}"		{ myputc(' '); my_push_state(IN_FORMAL); }

"\\begin{syntax}"	{ 
			  myputc(' ');
			  my_push_state(IN_FORMAL);
			}

"\\zsection"/{NA}	{ 
			  myputs(" section ");
			  /* go briefly into Z text mode to get name */
			  missingRBraces = 0;
			  my_push_state(IN_FORMAL);
			}

"\\zdefaultsection"/{NA} {
			  myputs(" default_section ");
			  /* go briefly into Z text mode to get name */
			  missingRBraces = 0;
			  my_push_state(IN_FORMAL);
			}

"\\zcollectsection"/{NA} {
			  myputs(" collect_section ");
			  /* go briefly into Z text mode to get name */
			  missingRBraces = 0;
			  my_push_state(IN_FORMAL);
			}

	/* additional character for CSP-Z,... */

	/* CLASS */
"\\begin{class}"	{ 
			  myputc(UC); myputs("class"); 
			  my_push_state(IN_FORMAL);
			}

"\\features"/{NA}	{
			  myputs(" features ");
			  my_push_state(IN_FORMAL);
			}
"\\begin{features}"	{
			  myputs(" features ");
			  my_push_state(IN_FORMAL);
			}

"\\zremove"/{NA}	{
			  myputs(" remove ");
			  my_push_state(IN_FORMAL);
			}
"\\begin{zremove}"	{
			  myputs(" remove ");
			  my_push_state(IN_FORMAL);
			}

.			{ /* delete informal text */ }

}

	/* rules when either inside or outside of Z text */

<OUTOF_FORMAL,IN_FORMAL>{

	/* AXCHAR */
"\\begin{axdef}"	{ 
			  myputc(UC); myputs("2577"); 
			  my_push_state(IN_FORMAL);
			}

	/* AXCHAR, GENCHAR */
"\\begin{gendef}"	{ 
			  myputc(UC); myputs("2577");
			  myputc(UC); myputs("2550");
			  my_push_state(IN_FORMAL);
			}

	/* SCHCHAR, GENCHAR */
"\\begin{schema}"/"{"[^}]*"}["	{ 
			  myputc(UC); myputs("250C");
			  myputc(UC); myputs("2550");
			  my_push_state(IN_FORMAL);
			}

	/* SCHCHAR */
"\\begin{schema}"	{ 
			  myputc(UC); myputs("250C"); 
			  my_push_state(IN_FORMAL);
			}

"\\zsectionend"/{NA}	{ myputs(" section_end ");
			  /* go briefly into Z text mode to get name */
			  missingRBraces = 0;
			  my_push_state(IN_FORMAL);
			}
"\\featend"/{NA}	{ myputs(" feature_end ");
			  /* go briefly into Z text mode to get name */
			  missingRBraces = 0;
			  my_push_state(IN_FORMAL);
			}
"\\familymemberend"/{NA} { myputs(" familymember_end ");
			  /* go briefly into Z text mode to get name */
			  missingRBraces = 0;
			  my_push_state(IN_FORMAL);
			}

}

	/* rules when either inside or outside of Z text, or of Z comments */

<OUTOF_FORMAL,IN_FORMAL,IN_BRACEDCOMMENT,IN_CLASSCOMMENT>{

	/* delete LaTeX comments */
"%".*$			{ /* empty */ }

	/* pass on newlines to the lexical analysis, such that it can
	   keep track of the line count */
"\n"			{ ECHO; }

}

	/* rules when in Z text */

<IN_FORMAL>{

"\\end{axdef}"		|
"\\end{gendef}"		|
"\\end{schema}"		|
"\\end{zed}"		| 
"\\end{syntax}"		{ 
			/* ENDCHAR */
			myputc(UC); myputs("2029");
			my_pop_state();
			}

	/* additional characters for CSP-Z,... */
"\\endfeatures"/{NA}	{ 
			/* ENDCHAR */
			myputc(UC); myputs("2029");
			my_pop_state();
			}
"\\end{features}"	{ 
			/* ENDCHAR */
			myputc(UC); myputs("2029");
			my_pop_state();
			}
"\\endzremove"/{NA}	{ 
			/* ENDCHAR */
			myputc(UC); myputs("2029");
			my_pop_state();
			}
"\\end{zremove}"	{ 
			/* ENDCHAR */
			myputc(UC); myputs("2029");
			my_pop_state();
			}
"\\enddefaultproperties"/{NA} { 
			/* ENDCHAR */
			myputc(UC); myputs("2029");
			}
"\\end{defaultproperties}" { 
			/* ENDCHAR */
			myputc(UC); myputs("2029");
			}
"\\inheritinterface"/{NA} {
			  myputs(" inherit-interface ");
			}
"\\begin{inheritinterface}" {
			  myputs(" inherit-interface ");
			}
"\\endinheritinterface"/{NA} {
			  /* ENDCHAR */
			  myputc(UC); myputs("2029");
			}
"\\end{inheritinterface}" {
			  /* ENDCHAR */
			  myputc(UC); myputs("2029");
			}
"\\inherit"/{NA}	{
			  myputs(" inherit ");
			}
"\\begin{inherit}"	{
			  myputs(" inherit ");
			}
"\\endinherit"/{NA}	{
			  /* ENDCHAR */
			  myputc(UC); myputs("2029");
			}
"\\end{inherit}"	{
			  /* ENDCHAR */
			  myputc(UC); myputs("2029");
			}
"\\defaultproperties"/{NA} {
			  myputs(" default_properties ");
			}
"\\begin{defaultproperties}" {
			  myputs(" default_properties ");
			}
"\\end{class}"		{ 
			  /* ENDCHAR */
			  /* myputc(UC); myputs("2029"); */
	/* this really should be ENDCHAR, but we want to find the end of
	   a class despite that we don't want to parse the entire class: */
			  myputc(UC); myputs("endclass");
			  my_pop_state();
			}
"\\cdef"/{NA}		{ myputc(UC); myputs("cdef"); }
"\\visibility"/{NA}	{ myputc(UC); myputs("visib"); }
	  /* the token CHAN denotes all kinds of chan*, method*, local_chan* */
"\\chan"/{NA}		{ myputs(" chan "); }
"\\chan{"[12]"}"	{ myputs(" chan "); }
"\\lchan"/{NA}		{ myputs(" chan "); }
"\\lchan{"[12]"}"	{ myputs(" chan "); }
"\\method"/{NA}		{ myputs(" chan "); }
"\\method{"[12]"}"	{ myputs(" chan "); }

"\\rcomment"/{NA}	{ 
			  myputc(' ');
			  my_push_state(IN_BRACEDCOMMENT);
			}
"\\begin{classcom}"	{ 
			  myputc(' ');
			  my_push_state(IN_CLASSCOMMENT);
			}

			/* NLCHAR */
"\\\\"			{ myputc(UC); myputs("2028"); }
"\\zbreak"/{NA}		{ myputc(UC); myputs("2028"); }
"\\Zbreak"/{NA}		{ myputc(UC); myputs("2028"); }
"\\ZBREAK"/{NA}		{ myputc(UC); myputs("2028"); }
"\\also"/{NA}		{ myputc(UC); myputs("2028"); }
"\\Also"/{NA}		{ myputc(UC); myputs("2028"); }
"\\ALSO"/{NA}		{ myputc(UC); myputs("2028"); }
"\\znewpage"/{NA}	{ myputc(UC); myputs("2028"); }
"\\eol"/{NA}		{ myputc(UC); myputs("2028"); }

	/* CAPITALDELTA */
"\\Delta"{WS}*		{ myputc(UC); myputs("0394"); }
"\\Delta"/{NA}		{ myputc(UC); myputs("0394"); }
	/* CAPITALXI */
"\\Xi"{WS}*		{ myputc(UC); myputs("039E"); }
"\\Xi"/{NA}		{ myputc(UC); myputs("039E"); }
	/* THETA */
"\\theta"{WS}*		{ myputc(UC); myputs("03B8"); }
"\\theta"/{NA}		{ myputc(UC); myputs("03B8"); }
	/* LAMBDA */
"\\lambda"{WS}*		{ myputc(UC); myputs("03BB"); }
"\\lambda"/{NA}		{ myputc(UC); myputs("03BB"); }
	/* MU */
"\\mu"{WS}*		{ myputc(UC); myputs("03BC"); }
"\\mu"/{NA}		{ myputc(UC); myputs("03BC"); }

"\\_"			{ myputc('_'); }
"\\{"			{ myputc('{'); }
"\\}"			{ myputc('}'); }
  /* ... */

	/* DOWN */
	/* "_{"/{LATEXTOKEN}"}"	{ myputc(UC); myputs("2198"); BEGIN(INDOWN); }*/

"{"			{ if(missingRBraces >= 0) missingRBraces++; }
"}"			{ 
			  if(missingRBraces >= 0)
			  { if(--missingRBraces == 0)
			    { missingRBraces = -1;
			      my_pop_state();
			    }
			  }
			}

"\t"			|
"~"			|
"\\,"			|
"\\!"			|
"\\ "			|
"\\;"			|
"\\:"			|
"\\t"[1-9]		{ myputc(' '); }
"\\Tone"/{NA}		{ myputc(' '); }
"\\protect"/{NA}	{ myputc(' '); }

"\\fvnextisdef"/{NA}	{ myputc(' '); }

"{"{LATEXTOKEN}"}"	{
			  yytext++; yytext++; /* delete '{' and '\\' */
			  yytext[strlen(yytext)-1] = '\0'; /* delete '}' */
			  myputs(yytext);
			  if(missingRBraces == 0)
			  { missingRBraces = -1;
			    my_pop_state();
			  }
			}

{LATEXTOKEN}/{NA}	{
			  yytext++; /* delete '\\' */
			  myputc(' ');
			  myputs(yytext);
			  myputc(' ');
			}

" "+			{ myputc(' '); }

		/* do not mark up the rest of the text */
.			ECHO;


}

	/* rules when in a braced comment in Z */

<IN_BRACEDCOMMENT>{

"\\\\"			{ /* delete comment */ }
"\\{"			{ /* delete comment */ }
"\\}"			{ /* delete comment */ }

"{"			{ commentLBraces++; }
"}"			{ 
			  if(--commentLBraces == 0)
			  { my_pop_state();
			  }
			}

.			{ /* delete comment */ }

}

	/* rules when in a classcom comment in CSP-OZ */

<IN_CLASSCOMMENT>{

"\\\\"			{ /* delete comment */ }

"\\end{classcom}"	{ my_pop_state(); }

.			{ /* delete comment */ }

}


	/* ENDDOWN */
	/* <INDOWN>"}"		{ myputc(UC); myputs("2196"); BEGIN(IN_FORMAL); } */

<<EOF>>			{
			  /* Terminate the output with an end-of-input 
			     marker. It is crucial to append a newline,
			     too, since the end-of-file tests appear to
			     not work on pipes. */
			  myputc(UC); myputs("EOI\n");
			  return(0);
			}

<INITIAL>(.|\n)		{
			  fprintf(stderr, "%s: mark-up: internal error: illegal start condition reached.\n", PROGNAME);
	    		  exit(EXITRESOURCES);
			  }


%%

