/*-------------------------------------------------------------------------

project: genFamMem - maintaining families of requirements

author: Jan Bredereke
copyright: Jan Bredereke, 2000
$Id: identifier.c,v 1.5 2001/01/29 09:23:13 brederek Exp $

file: identifier.c - conversion of identifiers to integer
		     ID numbers

The ID numbers are generated as we go on and should not be interpreted.

We internally use a hash table to speed up the look-up of strings.

-------------------------------------------------------------------------*/

#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <stdio.h>
/* hash table */
#include <search.h>
#include "identifier.h"

/*-------------------------------------------------------------------------
Local data
-------------------------------------------------------------------------*/

/* The table of identifiers, it associates id numbers (= indices) to
   identifiers */
#define INITIALNAMETABLELEN 10000
static char **lNameTable;
/* the max. factor of usage of the entries in the hash table */
#define MAXHASHTABLEUSAGE 0.8
/* Currently memory-allocated number of entries in the table */
static int lNameSize = 0;
/* Currently used number of entries in the table */
static int lNameNum = 0;
/* Did we not yet initialize the hash table? */
static int lHashTableUninitialized = 1;

/*-------------------------------------------------------------------------
The functions for internal use
-------------------------------------------------------------------------*/

/* assertMem() */
/* check that an expression is true; if not, complain about insufficient
   memory and exit with code 2 */
#define assertMem(expr) (expr ? 0 : assertMemFail())
static int
assertMemFail()
{   fprintf(stderr, "\n\nCannot allocate further memory for the table of identifiers.\nExiting...\n");
    exit(2);
}

/*-------------------------------------------------------------------------
The functions for external use
-------------------------------------------------------------------------*/

/* convert an identifier into an id number */
/* returns: a non-negative id number on success */
/* exits with an error message and code 2 if there is no more space in the */
/*          internal table */
/* Same identifiers always get the same id number. */
/* Identifiers may be arbitrary long. */
int
idIdentifier2Id(char *str)
{   ENTRY hashTableEntry;
    ENTRY *hashResult;
    int idx;

    assert(str);

    if(lHashTableUninitialized)
    {   /* if not yet done, initialize hash table */
        lNameSize = INITIALNAMETABLELEN;
        hcreate( (unsigned) (lNameSize / MAXHASHTABLEUSAGE) );
        lHashTableUninitialized = 0;
	/* also allocate a table of identifiers */
	lNameTable = (char **) malloc(lNameSize * sizeof(char *));
	assertMem(lNameTable);
    }

    /* do a hash table lookup whether we already have got this identifier */
    hashTableEntry.key = str;
    hashResult = hsearch(hashTableEntry, FIND);
    if(hashResult)
    {   /* found: return its id number */
        return( (int) hashResult->data);
    }
    /* not found */

    /* check whether still space in name table, and make room if necessary */
    if(lNameNum >= lNameSize)
    {   lNameSize *= 2;
        /* 
	fprintf(stderr, "\nExpanding the identifier tables to size %i.\n",
	        lNameSize);
	*/
        lNameTable = (char **) realloc((void *) lNameTable,
    				       lNameSize * sizeof(char *));
	assertMem(lNameTable);
	assert(lNameNum < lNameSize);
	/* create a larger hash table, too, and re-enter the values */
	hdestroy();
        hcreate( (unsigned) (lNameSize / MAXHASHTABLEUSAGE) );
	for(idx = 0; idx < lNameNum; idx++)
	{   hashTableEntry.key = lNameTable[idx];
	    hashTableEntry.data = (char *) idx;
	    hashResult = hsearch(hashTableEntry, ENTER);
	    assertMem(hashResult);
	}
    }
    /* copy identifier into some freshly allocated memory */
    lNameTable[lNameNum] = (char *) malloc((strlen(str)+1) * sizeof(char));
    assertMem(lNameTable[lNameNum]);
    strcpy(lNameTable[lNameNum], str);
    /* enter identifier/id pair into hash table, too */
    hashTableEntry.key = lNameTable[lNameNum];
    hashTableEntry.data = (char *) lNameNum;
    hashResult = hsearch(hashTableEntry, ENTER);
    assertMem(hashResult);

    /* finally, return id number */
    return(lNameNum++);
}

/* convert an id number into an identifier */
/* returns: a pointer to the identifier on success */
/*          NULL if the id number has not yet been allocated */
char *
idId2Identifier(int id)
{   if((id < 0) || (lNameNum <= id))
	return((char *) NULL);
    return(lNameTable[id]);
}

/* take two identifiers, concatenate their string values, and return the
   id number of the result */
/* returns: a non-negative id number on success */
/*          IDERR if some of the id numbers have not yet been allocated */
/* exits with an error message and code 2 if there is no more space in the */
/*          internal table */
int
idConcatIds2Id(int id1, int id2)
{   char *tmpStr;
    int tmpId;

    if((id1 < 0) || (lNameNum <= id1) || (id2 < 0) || (lNameNum <= id2))
        return(IDERR);
    tmpStr = (char *) malloc(( strlen(idId2Identifier(id1))
                              +strlen(idId2Identifier(id2))
			      +1)
			     * sizeof(char));
    strcpy(tmpStr, idId2Identifier(id1));
    strcat(tmpStr, idId2Identifier(id2));
    tmpId = idIdentifier2Id(tmpStr);
    free(tmpStr);
    return(tmpId);
}

