/*-------------------------------------------------------------------------

project: genFamMem - maintaining families of requirements

author: Jan Bredereke
copyright: Jan Bredereke, 2000
$Id: linerange.c,v 1.10 2001/02/09 09:00:39 brederek Exp $

file: linerange.c - ranges of lines and operations on them

If NDEBUG is defined, the internal assertions are not checked anymore.

-------------------------------------------------------------------------*/

#include <assert.h>
#include <stdio.h>
#include <string.h>
#include "linerange.h"

/*-------------------------------------------------------------------------
The local functions for internal use
-------------------------------------------------------------------------*/


/*-------------------------------------------------------------------------
The functions for external use
-------------------------------------------------------------------------*/

/* alloc and initialize a new set of line ranges */
/* The set will be empty. */
/* returns: the set or NULL on error, the error code being in errno */
linLineRangeH 
linNewLineRange()
{   linLineRangeH range;

    range = (linLineRangeH) malloc(sizeof(struct lin_LineRange));
    if(!range)
	return((linLineRangeH) NULL);
    range->pairNum = 0;
    return(range);
}

/* free the memory of a set of line ranges */
void
relFreeLineRange(linLineRangeH range)
{   assert(range);
    free((void *) range);
}

/* add a line pair to a set of line ranges (with merging, if necessary) */
/* returns: LINERR on error (out of memory) */
/*          LINOK on success */ 
/* a failed assertion is rised of the line numbers are not legal. */
int
linAddRange(linLineRangeH range, int first, int last)
{   int idx;

    assert(range);
    assert((last == -1) || (first <= last));
    assert((last == -1) || (last >= 1));
    assert(first >= 1);
    /* make sure there is space for at least one line pair in the list */
    if(range->pairNum >= LINMAXRANGES)
	return(LINERR);
    /* search for a first line that is not smaller than the new first line */
    for(idx = 0; idx < range->pairNum; idx++)
    {   if(first <= range->linePairs[idx].first)
    	    break;
    }
    /* (if not found, idx points to the first free entry in the list) */

    /* move the rest of the list one position to the back */
    memmove((void *) &range->linePairs[idx + 1],
	    (void *) &range->linePairs[idx],
	    (range->pairNum - idx) * sizeof(struct lin_LinePair));

    /* add the line pair to the list; no merging yet */
    range->linePairs[idx].first = first;
    range->linePairs[idx].last = last;
    range->pairNum++;

    /* if possible, merge with the previous line pair */
    if(idx && linCmp(range->linePairs[idx - 1].last + 1, first) >= 0)
    {   range->linePairs[idx - 1].last = 
	    linMax(range->linePairs[idx - 1].last, last);
	range->pairNum--;
	idx--;
        memmove((void *) &range->linePairs[idx + 1],
	        (void *) &range->linePairs[idx + 2],
		(range->pairNum - (idx + 1)) * sizeof(struct lin_LinePair));
    }

    /* while possible, merge with the following line pairs */
    while( (idx < range->pairNum - 1)
          && (linCmp(range->linePairs[idx].last + 1,
	             range->linePairs[idx + 1].first) >= 0)
	 )
    {   range->linePairs[idx].last = 
	    linMax(range->linePairs[idx].last, range->linePairs[idx + 1].last);
	range->pairNum--;
	memmove((void *) &range->linePairs[idx + 1],
		(void *) &range->linePairs[idx + 2],
		(range->pairNum - (idx + 1)) * sizeof(struct lin_LinePair));
    }

    return(LINOK);
}

/* dump a set of line ranges to a stream in ASCII format */
extern void
linDumpLineRange(FILE *stream, linLineRangeH range)
{   int idx;

    assert(stream);
    assert(range);
    for(idx = 0; idx < range->pairNum; idx++)
    {   if(idx)
	    fprintf(stream, ",");
	fprintf(stream, "%d-%d",
	        range->linePairs[idx].first,
		range->linePairs[idx].last); 
    }
}

/* for a line range (which may be non-contiguous), find out what is the
   smallest line number for which there is a contiguous sub-range up
   to a given line number */
/* returns: the smallest line on success */
/*          LINERR if the line is not in the range */
int
linLocalMinimum(linLineRangeH range, int line)
{   int idx;

    assert(range);
    if(line == LINERR)
        return(LINERR);
    /* search for a last line that is not smaller than the given line */
    for(idx = 0; idx < range->pairNum; idx++)
    {   if(linCmp(line, range->linePairs[idx].last) <= 0)
    	    break;
    }
    if(idx >= range->pairNum)
        return(LINERR);
    if(line < range->linePairs[idx].first)
        return(LINERR);
    return(range->linePairs[idx].first);
}

/* iterators */

/* Please note that the iterated-over sets must not be modified while
   iterating. */

/* initialize an iterator over a set of line ranges */
/* This might be implemented as a macro. */
void
linInitLRangeIt(linLineRangeIt *it, linLineRangeH range)
{   assert(it);
    assert(range);
    it->range = range;
    it->currIdx = 0;
}

/* test an iterator over a set of line ranges whether an element is available */
/* returns: 1 on yes */
/*          0 on no */
/* This might be implemented as a macro. */
char
linMoreLRangeIt(linLineRangeIt *it)
{   assert(it);
    return(it->currIdx < it->range->pairNum);
}

/* return the current first line of the set of the iterator */
/* relMoreLRangeIt() must be 1. */
/* This might be implemented as a macro. */
int
linCurrFirstLRangeIt(linLineRangeIt *it)
{   assert(it);
    assert(it->range);
    assert(0 <= it->currIdx);
    assert(it->currIdx < it->range->pairNum);
    return(it->range->linePairs[it->currIdx].first);
}

/* return the current last line of the set of the iterator */
/* relMoreLRangeIt() must be 1. */
/* This might be implemented as a macro. */
int
linCurrLastLRangeIt(linLineRangeIt *it)
{   assert(it);
    assert(it->range);
    assert(0 <= it->currIdx);
    assert(it->currIdx < it->range->pairNum);
    return(it->range->linePairs[it->currIdx].last);
}

/* step to the next element of the set of the iterator */
/* relMoreLRangeIt() must be 1. */
/* This might be implemented as a macro. */
void
linNextLRangeIt(linLineRangeIt *it)
{   assert(it);
    assert(it->range);
    assert(0 <= it->currIdx);
    assert(it->currIdx < it->range->pairNum);
    it->currIdx++;
}

/* calculate the maximum of two line numbers, considering -1 as +infinite */
/* returns: the maximum */
int
linMax(int line1, int line2)
{
    if(line1 == -1)
        return(line2);
    if(line2 == -1)
        return(line1);
    assert((line1 >= 1) && (line2 >= 1));
    return(line1 > line2 ? line1 : line2);
}

/* calculate the minimum of two line numbers, considering -1 as +infinite */
/* returns: the minimum */
int
linMin(int line1, int line2)
{
    if(line1 == -1)
        return(line1);
    if(line2 == -1)
        return(line2);
    assert((line1 >= 1) && (line2 >= 1));
    return(line1 < line2 ? line1 : line2);
}

/* compare two line numbers, considering -1 as +infinite */
/* returns: -1 if line1 < line2 */
/*          0  if line1 = line2 */
/*          1  if line1 > line2 */
int
linCmp(int line1, int line2)
{
    if(line1 == line2)
        return(0);
    if(line1 == -1)
        return(1);
    if(line2 == -1)
        return(-1);
    return(line1 < line2 ? -1 : 1);
}

