/*-------------------------------------------------------------------------

project: genFamMem - maintaining families of requirements

author: Jan Bredereke
copyright: Jan Bredereke, 2000
$Id: linerange.h,v 1.6 2001/02/09 09:00:39 brederek Exp $

file: linerange.h - header file: ranges of lines and operations on them

If NDEBUG is defined, the internal assertions are not checked anymore.

-------------------------------------------------------------------------*/

#ifndef LINERANGES_H
#define LINERANGES_H

#include <stdlib.h>
#include <stdio.h>


/*-------------------------------------------------------------------------
The constants for external use
-------------------------------------------------------------------------*/

/* return codes */
/* OK */
#define LINOK	  0
/* ERROR */
#define LINERR	  -1
/* could not do something because it already existed / did not exist */
#define LINCANTDO 1

/* Please note that the below constants may change when we switch
   to dynamic memory management. */
/* max. number of ranges in a range set */
#define LINMAXRANGES	500

/*-------------------------------------------------------------------------
The types for external use
-------------------------------------------------------------------------*/

/* handle for a set of line ranges */
typedef struct lin_LineRange *linLineRangeH;
/* iterator for a set of line ranges */
typedef struct lin_LineRangeIt linLineRangeIt;

/* Line numbers are integers, starting with 1; -1 stands for end-of-input. */

/*-------------------------------------------------------------------------
Some type definitions for internal use only, and for memory allocation
-------------------------------------------------------------------------*/
/* Please note that we might switch to dynamic memory allocation and
   re-allocation. Do not rely on the static structure below. */
/* pair of lines */
struct lin_LinePair
{   int first;
    int last;
};
/* set of line ranges */
/* Please note that the line pairs are in ascending order, and that there
   must be a gap between two adjacent pairs; otherwise they are merged. */
struct lin_LineRange
{   struct lin_LinePair linePairs[LINMAXRANGES];
    int pairNum;
};
/* iterator for a set of line ranges */
struct lin_LineRangeIt
{   struct lin_LineRange *range;
    int currIdx;
};

/*-------------------------------------------------------------------------
The functions for external use
-------------------------------------------------------------------------*/

/* alloc and initialize a new set of line ranges */
/* The set will be empty. */
/* returns: the set or NULL on error, the error code being in errno */
extern linLineRangeH 
linNewLineRange();

/* free the memory of a set of line ranges */
extern void
relFreeLineRange(linLineRangeH range);

/* add a line pair to a set of line ranges (with merging, if necessary) */
/* returns: LINERR on error */
/*          LINOK on success */ 
extern int
linAddRange(linLineRangeH range, int first, int last);

/* dump a set of line ranges to a stream in ASCII format */
extern void
linDumpLineRange(FILE *stream, linLineRangeH range);

/* for a line range (which may be non-contiguous), find out what is the
   smallest line number for which there is a contiguous sub-range up
   to a given line number */
/* returns: the smallest line on success */
/*          LINERR if the line is not in the range */
extern int
linLocalMinimum(linLineRangeH range, int line);

/* iterators */

/* Please note that the iterated-over sets must not be modified while
   iterating. */

/* initialize an iterator over a set of line ranges */
/* This might be implemented as a macro. */
extern void
linInitLRangeIt(linLineRangeIt *it, linLineRangeH range);

/* test an iterator over a set of line ranges whether an element is available */
/* returns: 1 on yes */
/*          0 on no */
/* This might be implemented as a macro. */
extern char
linMoreLRangeIt(linLineRangeIt *it);

/* return the current first line of the set of the iterator */
/* relMoreLRangeIt() must be 1. */
/* This might be implemented as a macro. */
extern int
linCurrFirstLRangeIt(linLineRangeIt *it);

/* return the current last line of the set of the iterator */
/* relMoreLRangeIt() must be 1. */
/* This might be implemented as a macro. */
extern int
linCurrLastLRangeIt(linLineRangeIt *it);

/* step to the next element of the set of the iterator */
/* relMoreLRangeIt() must be 1. */
/* This might be implemented as a macro. */
extern void
linNextLRangeIt(linLineRangeIt *it);

/* calculate the maximum of two line numbers, considering -1 as +infinite */
/* returns: the maximum */
extern int
linMax(int line1, int line2);

/* calculate the minimum of two line numbers, considering -1 as +infinite */
/* returns: the minimum */
extern int
linMin(int line1, int line2);

/* compare two line numbers, considering -1 as +infinite */
/* returns: -1 if line1 < line2 */
/*          0  if line1 = line2 */
/*          1  if line1 > line2 */
extern int
linCmp(int line1, int line2);

#endif /* LINERANGES_H */
