/*-------------------------------------------------------------------------

project: genFamMem - maintaining families of requirements

author: Jan Bredereke
copyright: Jan Bredereke, 2000
$Id: main.c,v 1.77 2001/02/20 08:01:42 brederek Exp $

file: main.c - main file of the generator program

-------------------------------------------------------------------------*/

#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <assert.h>

/* command line option processing */
#define _GNU_SOURCE
#include <getopt.h>
extern char *optarg; /* argument of current option, if present */
extern int optind;   /* after option processing, is the index in argv
		        to the first parameter. */

#include "global.h"
#include "relation.h"
#include "identifier.h"
#include "calc.h"
#include "filterspec.h"
#include "debug.h"


/* global variables */

/* debugging for bison */
#ifdef YYDEBUG
extern int fplyydebug;
#endif

/* error reporting for bison */
extern int fplyynerrs;

extern FILE *fpmtyyin; /* input stream for mark-up of first pass */
extern FILE *fpmtyyout; /* output stream for mark-up of first pass */
extern FILE *fplyyin; /* input stream for lexical analysis of first pass */

/* the command line options of the program, as far as they need
   to be accessed globally. */
struct progOptsType progOpts = 
{   /* inFileName= */ NULL,
    /* debug= */ 0,
    /* markup= */ 0,
    /* language= */ LANGUAGE_DEFAULT,
    /* selectedFamilyMember= */ IDERR,
    /* selectedFeature= */ IDERR,
    /* usesHier= */ UHIERNO,
    /* ascii= */ 0,
    /* onlyFeature= */ 0
};

/* the diverse relations on identifiers, globally accessed by the
   different lex/yacc parser stages */
struct idRelsType idRels;

/* functions */

/* print usage */
static void
printUsage()
{   fprintf(stderr, "Usage:\n");
    fprintf(stderr, "\t%s <options> familymember infile\n", PROGNAME);
    fprintf(stderr, "\t%s <options> --uhier-all infile\n", PROGNAME);
    fprintf(stderr, "\t%s <options> --uhier-familymember familymember infile\n",
                    PROGNAME);
    fprintf(stderr, "\t%s <options> --uhier-feature feature infile\n",
                    PROGNAME);
    fprintf(stderr, "Options are:\n");
    fprintf(stderr, "\t--uhier-all, -l:\n");
    fprintf(stderr, "\t\tprint uses-hierarchy of all sections, only\n");
    fprintf(stderr, "\t--uhier-familymember familymember, -a familymember:\n");
    fprintf(stderr, "\t\tsame, for one family member\n");
    fprintf(stderr, "\t--uhier-feature feature, -e feature:\n");
    fprintf(stderr, "\t\tsame, for one feature\n");
    fprintf(stderr, "\t--ascii, -x:\n");
    fprintf(stderr, "\t\twhen --uhier-*, use ASCII output format\n");
    fprintf(stderr, "\t--daVinci, -X:\n");
   fprintf(stderr, "\t\twhen --uhier-*, use daVinci output format (default)\n");
    fprintf(stderr, "\t--only-features, -f:\n");
    fprintf(stderr, "\t\twhen --uhier-*, abstract sections to features\n");
    fprintf(stderr, "\t--language-z, -z:\n");
    fprintf(stderr, "\t\tinput language is Z\n");
    fprintf(stderr, "\t--language-cspz, -c:\n");
    fprintf(stderr, "\t\tinput language is CSP_Z\n");
#if 0
    fprintf(stderr, "\t--language-cspm, -m:\n");
    fprintf(stderr, "\t\tinput language is CSP_M\n");
#endif
    fprintf(stderr, "\t--language-cspoz, -o:\n");
    fprintf(stderr, "\t\tinput language is CSP-OZ (default)\n");
    fprintf(stderr, "\t--help, -?, -h:\n");
    fprintf(stderr, "\t\tprint this help info\n");
    fprintf(stderr, "\t--version, -V:\n");
    fprintf(stderr, "\t\tprint version information\n");
    fprintf(stderr, "\t--debug, -d:\n");
    fprintf(stderr, "\t\tprint additional debug information\n");
#ifdef YYDEBUG
    fprintf(stderr, "\t--bison-debug, -D:\n");
    fprintf(stderr, "\t\tlet the bison parser print its debug information\n");
#endif
    fprintf(stderr, "\t--markup, -M:\n");
    fprintf(stderr,"\t\tdo mark-up only and print it (useful for debugging)\n");
    fprintf(stderr, "Output appears on stdout.\n");
    fprintf(stderr, "An option argument is specified as in \"--uhier-feature=foo\",\n");
    fprintf(stderr, "in \"--uhier-feature foo\", or in \"-efoo\".\n");

    return;
}

/* print version information */
static void
printVersion()
{   printf("This is %s, version %s.\n", PROGNAME, PROGVERSION);
    printf("Copyright: Jan Bredereke, 2000\n");
    return;
}

/* the main function */
int
main(int argc, char *argv[])
{   int retCode; /* diverse return codes to check */
    pid_t childPid; /* PID of child when forked */
    int childStatus; /* exit status of this child */
    int fpPipes[2]; /* the pipes to the mark-up process */
    int semanticErrorNum = 0; /* the numbers of semantic errors encountered */

    /* initialize the sets and relations */

    idRels.features = relNewSet();
    idRels.featureEnds = relNewSet();
    idRels.anonymousFeatures = relNewSet();
    idRels.familyMembers = relNewSet();
    idRels.familyMemberEnds = relNewSet();
    idRels.nonPureFamilyMembers = relNewSet();
    idRels.sections = relNewSet();
    idRels.famMem2FeatureList = relNewRelation();
    idRels.feature2Section = relNewRelation();
    idRels.defaultSections = relNewSet();
    idRels.collectSections = relNewSet();
    idRels.sectionEnds = relNewSet();
    idRels.sectionParents = relNewRelation();
    idRels.sectionClass2Inherits = relNewRelation3();
    idRels.sectionClass2DefaultProperties = relNewRelation3();
    idRels.remove = relNewRelation();
    idRels.selectedFeatures = relNewSet();
    idRels.selectedRemove = relNewSet();
    idRels.selectedSections = relNewSet();
    idRels.selectedSectionParents = relNewRelation();
    idRels.selectedSectionDeletedParents = relNewRelation();
    idRels.section2Class = relNewRelation();
    idRels.feature2FirstLine = relNewRelation();
    idRels.feature2LastLine = relNewRelation();
    idRels.famMem2FirstLine = relNewRelation();
    idRels.famMem2LastLine = relNewRelation();
    idRels.section2FirstLine = relNewRelation();
    idRels.section2LastLine = relNewRelation();
    idRels.sectionClass2FirstLine = relNewRelation3();
    idRels.sectionClass2LastLine = relNewRelation3();
    idRels.sectionClassInherit2FirstLine = relNewRelation3();
    idRels.sectionClassInherit2LastLine = relNewRelation3();
    idRels.selectedSectionClassDeletedDefaultProperties = relNewRelation3();
    idRels.sectionParents2FirstLine = relNewRelation();
    idRels.sectionParents2LastLine = relNewRelation();
    idRels.deletedFeatures = relNewSet();
    idRels.deletedFamilyMembers = relNewSet();
    idRels.deletedLineRanges = linNewLineRange();
    idRels.parentsLineRanges = linNewLineRange();
    idRels.inheritLineRanges = linNewLineRange();
    idRels.usesHierarchy = relNewRelation();
    idRels.usesHierarchyRoots = relNewSet();
    idRels.usesHierarchyLevel = relNewRelation();

    idRels.usesHierarchyDepth = 0;

    /* process command line options */

    /* loop over all options */
    while(1)
    {   int currOpt;	/* the current option */
        /* the set of long options */
        static struct option long_options[] =
        {   {"uhier-all",          no_argument, NULL, 'l'},
            {"uhier-familymember", required_argument, NULL, 'a'},
            {"uhier-feature",      required_argument, NULL, 'e'},
            {"ascii",              no_argument, NULL, 'x'},
            {"daVinci",            no_argument, NULL, 'X'},
            {"only-features",      no_argument, NULL, 'f'},
            {"debug",              no_argument, NULL, 'd'},
#ifdef YYDEBUG
            {"bison-debug",        no_argument, NULL, 'D'},
#endif
            {"markup",             no_argument, NULL, 'M'},
            {"version",            no_argument, NULL, 'V'},
            {"language-z",         no_argument, NULL, 'z'},
            {"language-cspz",      no_argument, NULL, 'c'},
            {"language-cspoz",     no_argument, NULL, 'o'},
            {"help",               no_argument, NULL, '?'},
	    {0,                    0, 0, 0}
	};
	/* the set of short options */
#ifdef YYDEBUG
	static char short_options[] ="la:e:xXfdDMVzco?";
#else
	static char short_options[] ="la:e:xXfdMVzco?";
#endif

	/* get next option */
	currOpt=getopt_long(argc, argv, short_options, long_options, NULL);
	/* done? */
	if(currOpt==-1)
	    break;
	/* process option */
	switch(currOpt)
	{
	case 'l': /* "uhier-all" */
	    progOpts.usesHier = UHIERALL;
	    break;

	case 'a': /* "uhier-familymember" */
	    progOpts.usesHier = UHIERFAMMEM;
	    assert(optarg);
	    progOpts.selectedFamilyMember = idIdentifier2Id(optarg);
	    break;

	case 'e': /* "uhier-feature" */
	    progOpts.usesHier = UHIERFEAT;
	    assert(optarg);
	    progOpts.selectedFeature = idIdentifier2Id(optarg);
	    break;

	case 'x': /* "ascii" */
	    progOpts.ascii = 1;
	    break;

	case 'X': /* "daVinci" */
	    progOpts.ascii = 0;
	    break;

	case 'f': /* "only-feature" */
	    progOpts.onlyFeature = 1;
	    break;

	case 'd': /* "debug" */
	    progOpts.debug=1;
	    break;

#ifdef YYDEBUG
	case 'D': /* "bison-debug" */
	    fplyydebug=1;
	    break;
#endif

	case 'M': /* "markup" */
	    progOpts.markup=1;
	    break;

	case 'z': /* "language-z" */
	    progOpts.language=LANGUAGE_Z;
	    break;

	case 'c': /* "language-cspz" */
	    progOpts.language=LANGUAGE_CSPZ;
	    break;

	case 'o': /* "language-cspoz" */
	    progOpts.language=LANGUAGE_CSPOZ;
	    break;

	case 'V': /* "version" */
	    printVersion();
	    exit(EXITOK);
	    break;

	case '?': /* "help" */
	case 'h':
	    printUsage();
	    exit(EXITOK);

	case ':': /* missing option parameter */
	default: /* unknown option */
	    printUsage();
	    exit(EXITPARAMS);
	}
    }
    /* exactly two parameters left, or one for --uhier-* ? */
    if(   (   (progOpts.usesHier == UHIERFAMMEM)
           || (progOpts.usesHier == UHIERFEAT)
           || (progOpts.usesHier == UHIERALL))
	? (optind+1 != argc)
	: (optind+2 != argc)
      )
    {   printUsage();
	exit(EXITPARAMS);
    }

    /* get remaining parameters */
    if(!(   (progOpts.usesHier == UHIERFAMMEM)
         || (progOpts.usesHier == UHIERFEAT)
	 || (progOpts.usesHier == UHIERALL)))
    {   progOpts.selectedFamilyMember = idIdentifier2Id(argv[optind]);
	optind++;
    }
    progOpts.inFileName = argv[optind];

    /* open input file for the first pass */
    fpmtyyin = fopen(progOpts.inFileName, "r");
    if(!fpmtyyin)
    {   fprintf(stderr, "%s: failed to open input file '%s': ", 
                PROGNAME, progOpts.inFileName);
        perror("");
	exit(EXITPARAMS);
    }

    /* if mark-up only, do it and terminate */
    if(progOpts.markup)
    {
	fpmtyyout = stdout;
        fpmtyylex();
    	exit(EXITOK);
    }

    /* spawn off the process that does the (LaTeX, ...) mark-up */

    /* create pipe */
    retCode = pipe(fpPipes);
    if(retCode)
    {   fprintf(stderr, "%s: failed to create pipe to mark-up process: ", 
                PROGNAME);
        perror("");
	exit(EXITRESOURCES);
    }
    /* Please note that end-of-file tests appear to not work on pipes.
       Therefore, we write a single newline as the end-of-input marker. */

    /* fork */
    childPid = fork();
    if(childPid==-1)
    {   fprintf(stderr, "%s: failed to fork mark-up process: ", 
                PROGNAME);
        perror("");
	exit(EXITRESOURCES);
    }
    if(!childPid)
    {   /* we are the child */

	/* create stream buffer for write pipe */
	fpmtyyout = fdopen(fpPipes[1], "w");
	if(!fpmtyyout)
	{   fprintf(stderr, "%s: failed to set up pipe to mark-up process for writing: ", 
		    PROGNAME);
	    perror("");
	    exit(EXITRESOURCES);
	}

        /* do mark-up */
	switch(progOpts.language)
	{
	case LANGUAGE_CSPOZ:
	case LANGUAGE_CSPZ:
	case LANGUAGE_Z:
            fpmtyylex();
	    break;
	/* case LANGUAGE_CSPM: */
	    /* fpmmyylex(); */
	default:
	    assert(0);
	}
	retCode = fclose(fpmtyyout);
	if(retCode)
	{   fprintf(stderr, "%s: failed to close pipe from mark-up process for writing: ", 
		    PROGNAME);
	    perror("");
	    exit(EXITRESOURCES);
	}
	exit(0);
    }

    /* create stream buffer for read pipe */
    fplyyin = fdopen(fpPipes[0], "r");
    if(!fplyyin)
    {   fprintf(stderr, "%s: failed to set up pipe to mark-up process for reading: ", 
                PROGNAME);
        perror("");
	exit(EXITRESOURCES);
    }

    /* first pass over input file: build tables of structure of specification */
    retCode = fplyyparse();
    if(fplyynerrs)
    {   fprintf(stderr, "%s: encountered %i syntax error(s)\n",
		PROGNAME, fplyynerrs);
    }
    if(fplyynerrs || retCode)
        exit(EXITSYNERR);

    childPid = waitpid(childPid, &childStatus, 0);
    if(childPid==-1)
    {   fprintf(stderr, "%s: failed to wait for terminated mark-up process: ", 
                PROGNAME);
        perror("");
	exit(EXITRESOURCES);
    }
    if(!WIFEXITED(childStatus))
    {   fprintf(stderr, "%s: mark-up process terminated with error code %i.\n", 
                PROGNAME, WEXITSTATUS(childStatus));
        exit(EXITSYNERR);
    }

    /* extend the line numbers towards the end of the document just before
       the next item begins, except for *_end. */
    calcExtendLineNumbers();

    /* add the implicitly generated sections and classes for pure
       family members */
    semanticErrorNum = calcAddGeneratedSections();

    /* if debug, dump the raw relations */
    if(progOpts.debug)
        dbgDumpRawRelations();

    /* branch on further behaviour: generate spec. / print uses-hierarchy */

    if(progOpts.usesHier == UHIERNO)
    {   /* generate the specification for the specific family member */

	/* check whether the selected family member is actually a family
	   member */
	if(relIsMember(idRels.familyMembers, progOpts.selectedFamilyMember) 
	   == RELCANTDO)
	{   fprintf(stderr, "%s: the selected family member '%s' was not found.\n",
		    PROGNAME, idId2Identifier(progOpts.selectedFamilyMember));
	    exit(EXITSYNERR);
	}

	/* calculate the relations for the specific family member */
	semanticErrorNum += calcRelsSpecificMember();
	/* don't continue if any semantic errors encountered */
	if(semanticErrorNum)
	{   fprintf(stderr, 
		    "\n%s: %i semantic error(s) encountered. Exiting ...\n\n",
		    PROGNAME,
		    semanticErrorNum); 
	    exit(EXITSYNERR);
	}

	/* if debug, dump the reduced relations */
	if(progOpts.debug)
	   dbgDumpReducedRelations();

	/* second pass over input file:
	   copy the file literally to stdout, except for deleted lines */
	filFilterSpec(progOpts.inFileName, progOpts.language);
    }
    else
    {   /* print the uses-hierarchy */

	/* calculate the uses-hierarchy for the parameters specified on the 
	   command line */
        calcUsesHierarchy();

	/* if debug, dump info on the uses-hierarchy */
	if(progOpts.debug)
	   dbgDumpUsesHierarchy();

        /* print the uses-hierarchy */
	uhPrintUsesHierarchy();
    }

    exit(0);
}
