/*-------------------------------------------------------------------------

project: genFamMem - maintaining families of requirements

author: Jan Bredereke
copyright: Jan Bredereke, 2000
$Id: relation.c,v 1.33 2001/02/20 08:01:42 brederek Exp $

file: relation.c - sets of ids and mathematical relations over them

If NDEBUG is defined, the internal assertions are not checked anymore.

This is a very simple implementation only. We should introduce dynamic
memory management, sorted arrays etc. if necessary for performance
reasons.

-------------------------------------------------------------------------*/

#include <assert.h>
#include <stdio.h>
#include <string.h>
#include "relation.h"

/*-------------------------------------------------------------------------
The local functions for internal use
-------------------------------------------------------------------------*/

/* assertMem() */
/* check that an expression is true; if not, complain about insufficient
   memory and exit with code 2 */
#define assertMem(expr) (expr ? 0 : assertMemFail())
static int
assertMemFail()
{   fprintf(stderr, "\n\nCannot allocate memory for tables.\nExiting...\n");
    exit(2);
}

/* assertTableSpace() */
/* check that an expression is true; if not, complain about insufficient
   space in table and exit with code 2 */
#define assertTableSpace(expr) (expr ? 0 : assertTableSpaceFail())
static int
assertTableSpaceFail()
{   fprintf(stderr, "\n\nNo more space in internal table.\nExiting...\n");
    exit(2);
}

/* check whether an id is member of a set, and return a pointer 
   to the internal copy */
/* returns: RELOK on yes */
/*          RELCANTDO on no */
static int
lRelIsMember(relSetH set, int name, int **member)
{   int idx;

    assert(set);
    assert(member);
    assert(set->memberNum <= RELMAXSETMEMBERS);
    for(idx = 0; idx < set->memberNum; idx++)
    {   if(name == set->members[idx])
        {   *member = &set->members[idx];
            return(RELOK);
        }
    }
    *member = (int *) NULL;
    return(RELCANTDO);
}

/* delete the pointed-to relationship from relation rel */
/* All relationships before the pointed-to relationship are not moved. */
void
lDelRelship(relRelationH rel, int relshipIdx)
{   int retCode;

    assert(rel);
    assert(relshipIdx < rel->relships.memberNum);

    rel->relships.memberNum--;
    memmove(&(rel->relships.members[relshipIdx]), 
	    &(rel->relships.members[relshipIdx+1]),
	    (rel->relships.memberNum - relshipIdx) * sizeof(relRelship));

    return;
}

/* add a relationship to relation rel in the pointed-to position */
/* All relationships before the pointed-to position are not moved. */
/* exits with an error message and code 2 if there is no more memory */
void
lAddRelship(relRelationH rel, int relshipIdx, relRelship relship)
{   int retCode;

    assert(rel);
    assert(relshipIdx <= rel->relships.memberNum);

    assertTableSpace(rel->relships.memberNum < RELMAXRELSHIPS);
    memmove(&(rel->relships.members[relshipIdx+1]), 
	    &(rel->relships.members[relshipIdx]),
	   (rel->relships.memberNum - relshipIdx) * sizeof(relRelship));
    rel->relships.memberNum++;
    rel->relships.members[relshipIdx] = relship;

    return;
}

/* check whether a path exists in relation "rel" from "from" to "to",
   but not using elements from set "visited" */
/* "rel" must not be modified while searching. */
/* returns: RELOK on yes */
/*          RELCANTDO on no */
static int
lExistsPathNotInSet(relRelationH rel, int from, int to, relSetH visited)
{   int retCode;
    int myRetCode; /* return code of this function */
    relSetH tmpSet = NULL; /* an auxiliary set */
    relSetIt setIt; /* an iterator over sets */

    tmpSet = relNewSet();

    /* Get all the successors of the "from" element */
    retCode = relApply(tmpSet, rel, from);

    /* Check for all successors */
    myRetCode = RELCANTDO;
    for(relInitSetIt(&setIt, tmpSet);
        relMoreSetIt(&setIt);
	relNextSetIt(&setIt))
    {   /* test/set: if the successor has not already been visited */
        if(relAddMember(visited, relCurrSetIt(&setIt)) == RELOK)
	{   /* is the successor the one we look for? */
	    if(relCurrSetIt(&setIt) == to)
	    {   /* then stop searching */
		myRetCode = RELOK;
		break;
	    }
	    else
	    {   /* otherwise, do recursive search */
		retCode = lExistsPathNotInSet(rel, relCurrSetIt(&setIt), to,
					      visited);

		/* was there an indirect path through this successor? */
		if(retCode == RELOK)
		{   /* then stop searching */
		    myRetCode = RELOK;
		    break;
		}
	    }
	}
    }

    /* free allocated dynamic memory */

    relFreeSet(tmpSet);

    return(myRetCode);
}

/* check whether a path exists in relation "rel" from "from" to "to" */
/* "rel" must not be modified while searching. */
/* returns: RELOK on yes */
/*          RELCANTDO on no */
static int
lExistsPath(relRelationH rel, int from, int to)
{   int retCode;
    relSetH tmpSet = NULL; /* an auxiliary set */

    assert(rel);

    tmpSet = relNewSet();

    /* check whether a path exists in relation "rel" from "from" to "to",
       but not using elements from set tmpSet */
    retCode = lExistsPathNotInSet(rel, from, to, tmpSet);

    /* free allocated dynamic memory */

    relFreeSet(tmpSet);

    return(retCode);
}

/*-------------------------------------------------------------------------
The functions for external use
-------------------------------------------------------------------------*/

/* alloc and initialize a new relation */
/* The relation will be empty. */
/* returns: the relation on success */
/* exits with an error message and code 2 if there is no more memory */
relRelationH 
relNewRelation()
{   relRelationH rel;

    rel = (relRelationH) malloc(sizeof(struct rel_Relation));
    assertMem(rel);
    rel->relships.memberNum = 0;
    return(rel);
}

/* alloc and initialize a new ternary relation */
/* The relation will be empty. */
/* returns: the relation on success */
/* exits with an error message and code 2 if there is no more memory */
relRelation3H 
relNewRelation3()
{   relRelation3H rel;

    rel = (relRelation3H) malloc(sizeof(struct rel_Relation3));
    assertMem(rel);
    rel->relships.memberNum = 0;
    return(rel);
}

/* alloc and initialize a new set */
/* The set will be empty. */
/* returns: the set on success */
/* exits with an error message and code 2 if there is no more memory */
relSetH 
relNewSet()
{   relSetH set;

    set = (relSetH ) malloc(sizeof(struct rel_Set));
    assertMem(set);
    set->memberNum = 0;
    return(set);
}

/* free the memory of a relation */
void
relFreeRelation(relRelationH rel)
{   assert(rel);
    free((void *) rel);
}

/* free the memory of a ternary relation */
extern void
relFreeRelation3(relRelation3H rel)
{   assert(rel);
    free((void *) rel);
}

/* free the memory of a set */
void
relFreeSet(relSetH set)
{   assert(set);
    free((void *) set);
}

/* clear an existing set, it becomes the empty set */
void
relClearSet(relSetH set)
{   assert(set);
    set->memberNum = 0;
    return;
}

/* clear an existing relation, it becomes the empty relation */
void
relClearRelation(relRelationH rel)
{   assert(rel);
    rel->relships.memberNum = 0;
    return;
}

/* clear an existing ternary relation, it becomes the empty ternary relation */
void
relClearRelation3(relRelation3H rel)
{   assert(rel);
    rel->relships.memberNum = 0;
    return;
}

/* copy a set */
/* The set for the result must already be allocated and is overwritten */
/* returns: RELOK on success */
/*          RELCANTDO if the set is empty */
/* Invariant: the iteration order of the result is the same as that of the
   source */
int
relCopySet(relSetH result, relSetH set)
{   assert(set);
    assert(result);
    assert(set->memberNum < RELMAXSETMEMBERS);
    result->memberNum = set->memberNum;
    if(!result->memberNum)
        return(RELCANTDO);
    memcpy((void *)result->members,
           (void *)set->members,
	   result->memberNum * sizeof(int));
    return(RELOK);
}

/* copy a relation */
/* The relation for the result must already be allocated and is overwritten */
/* returns: RELOK on success */
/*          RELCANTDO if the relation is empty */
/* Invariant: the iteration order of the result is the same as that of the
   source */
int
relCopyRelation(relRelationH result, relRelationH rel)
{   assert(rel);
    assert(result);
    assert(rel->relships.memberNum < RELMAXRELSHIPS);

    result->relships.memberNum = rel->relships.memberNum;
    if(!result->relships.memberNum)
        return(RELCANTDO);
    memcpy((void *)result->relships.members,
           (void *)rel->relships.members,
	   result->relships.memberNum * sizeof(relRelship));

    return(RELOK);
}

/* copy a ternary relation */
/* The relation for the result must already be allocated and is overwritten */
/* returns: RELOK on success */
/*          RELCANTDO if the ternary relation is empty */
/* Invariant: the iteration order of the result is the same as that of the
   source */
int
relCopyRelation3(relRelation3H result, relRelation3H rel)
{   assert(rel);
    assert(result);
    assert(rel->relships.memberNum < RELMAXRELSHIPS);
    result->relships.memberNum = rel->relships.memberNum;
    if(!result->relships.memberNum)
        return(RELCANTDO);
    memcpy((void *)result->relships.members,
           (void *)rel->relships.members,
	   result->relships.memberNum * sizeof(int));
    return(RELOK);
}

/* add the relationship between ids name1 and name2 to relation rel */
/* returns: RELOK when relationship did not exist before */ 
/*          RELCANTDO when relationship did exist before */ 
/* exits with an error message and code 2 if there is no more memory */
/* Invariant: the iteration order remains the same, except that the new relship
   is appended to the end */
int
relAddRelship(relRelationH rel, int name1, int name2)
{   int retCode;
    int idx;

    assert(rel);
    retCode = relIsRelship(rel, name1, name2);
    if(retCode == RELOK)
        return(RELCANTDO);
    assertTableSpace(rel->relships.memberNum < RELMAXRELSHIPS);
    rel->relships.members[rel->relships.memberNum].first = name1;
    rel->relships.members[rel->relships.memberNum++].second = name2;
    return(RELOK);
}

/* add the ternary relationship between ids name1, name2, and name3 to 
   relation rel */
/* returns: RELOK when relationship did not exist before */ 
/*          RELCANTDO when relationship did exist before */ 
/* exits with an error message and code 2 if there is no more memory */
/* Invariant: the iteration order remains the same, except that the new relship
   is appended to the end */
int
relAddRelship3(relRelation3H rel, int name1, int name2, int name3)
{   int retCode;
    int idx;

    assert(rel);
    retCode = relIsRelship3(rel, name1, name2, name3);
    if(retCode == RELOK)
        return(RELCANTDO);
    assertTableSpace(rel->relships.memberNum < RELMAXRELSHIPS);
    rel->relships.members[rel->relships.memberNum].first = name1;
    rel->relships.members[rel->relships.memberNum].second = name2;
    rel->relships.members[rel->relships.memberNum++].third = name3;
    return(RELOK);
}

/* delete the relationship between ids name1 and name2 from relation rel */
/* returns: RELOK when relationship did exist */ 
/*          RELCANTDO when relationship did not exist */ 
/* Invariant: the iteration order remains the same, except that the deleted
   relship is removed */
int
relDelRelship(relRelationH rel, int name1, int name2)
{   int retCode;
    int idx;
    int found = 0;

    assert(rel);
    for(idx = 0; idx < rel->relships.memberNum; idx++)
    {   if((rel->relships.members[idx].first == name1)
           && (rel->relships.members[idx].second == name2))
	{   found = 1;
	    lDelRelship(rel, idx);
	}
    }
    if(!found)
	return(RELCANTDO);
    return(RELOK);
}

/* delete the ternary relationship between ids name1, name2, and name3 from 
   relation rel */
/* returns: RELOK when relationship did exist */ 
/*          RELCANTDO when relationship did not exist */ 
/* Invariant: the iteration order remains the same, except that the deleted
   relship is removed */
extern int
relDelRelship3(relRelation3H rel, int name1, int name2, int name3)
{   int retCode;
    int idx;
    int found = 0;

    assert(rel);
    for(idx = 0; idx < rel->relships.memberNum; idx++)
    {   if((rel->relships.members[idx].first == name1)
           && (rel->relships.members[idx].second == name2)
           && (rel->relships.members[idx].second == name3))
	{   found = 1;
	    rel->relships.memberNum--;
	    memmove(&(rel->relships.members[idx]), 
	            &(rel->relships.members[idx+1]),
	          (rel->relships.memberNum - idx) * sizeof(relRelship3));
	}
    }
    if(!found)
	return(RELCANTDO);
    return(RELOK);
}

/* add an id to a set */
/* returns: RELOK when id was no member of the set */ 
/*          RELCANTDO when id already was member of the set */ 
/* exits with an error message and code 2 if there is no more memory */
/* Invariant: the iteration order remains the same, except that the new id
   is appended to the end */
int
relAddMember(relSetH set, int name)
{   int retCode;

    assert(set);
    retCode = relIsMember(set, name);
    if(retCode == RELOK)
        return(RELCANTDO);
    assertTableSpace(set->memberNum < RELMAXSETMEMBERS);
    set->members[set->memberNum++] = name;
    return(RELOK);
}

/* deletes an id from a set */
/* returns: RELOK when id was a member of the set */ 
/*          RELCANTDO when id already was no member of the set */ 
/* Invariant: the iteration order remains the same, except that the deleted
   id is removed */
int
relDelMember(relSetH set, int name)
{   int retCode;
    int *member;

    assert(set);
    retCode = lRelIsMember(set, name, &member);
    if(retCode != RELOK)
        return(retCode);
    set->memberNum--;
    memmove(member, 
	    &member[1],
	    (&set->members[set->memberNum] - member) * sizeof(int));
    return(RELOK);
}

/* dump a relation to a stream in ASCII format */
/* 3rd arg: a function that prints one domain member for a selector value */
/* 4rd arg: a function that prints one range member */
/* 5th arg: a selector value */
/* 6th arg: a function that prints the newline between two domain members */
/* 7th arg: a function that prints the space between two range members */
/* Dump is in the iteration order */
void
relDumpRelation(FILE *stream, 
                relRelationH rel, 
		void (*dumpDomainMember)(FILE *, int, int /* selector */),
		void (*dumpRangeMember)(FILE *, int, int /* domainMember */),
		int selector,
		void (*dumpMemberNewline)(FILE *),
		void (*dumpMemberSpace)(FILE *))
{   relSetH domain, tmpSet;
    int retCode;
    int domainIdx;

    assert(stream);
    assert(rel);
    domain = relNewSet();
    tmpSet = relNewSet();
    retCode = relDomain(domain, rel);
    for(domainIdx = 0; domainIdx < domain->memberNum; domainIdx++)
    {   (*dumpDomainMember)(stream, domain->members[domainIdx], selector);
        retCode = relApply(tmpSet, rel, domain->members[domainIdx]);
	assert(retCode == RELOK);
	relDumpSetS(stream, tmpSet,
	            dumpRangeMember,
		    dumpMemberSpace,
		    domain->members[domainIdx]);
	dumpMemberNewline(stream);
    }
}

/* dump a ternary relation to a stream in ASCII format */
/* 3rd arg: a function that prints a member at the first position of 
            a relationship */
/* 4rd arg: a function that prints a member at the second position of 
            a relationship, for a given selector value */
/* 5rd arg: a function that prints a member at the third position of 
            a relationship */
/* 6th arg: a function that prints the newline between two first pos. members */
/* 7th arg: a function that prints the newline between two 2nd pos. members */
/* 8th arg: a function that prints the space between two 3rd pos. members */
/* Dump is in the iteration order */
void
relDumpRelation3(FILE *stream, 
		 relRelation3H rel,
		 void (*dumpFirstPosMember)(FILE *, int),
		 void (*dumpSecondPosMember)(FILE *, int, int /* selector */),
		 void (*dumpThirdPosMember)(FILE *, int,int/*secondPosMember*/),
		 void (*dumpMember1Newline)(FILE *),
		 void (*dumpMember2Newline)(FILE *),
		 void (*dumpMemberSpace)(FILE *))
{   relSetH firstElements;
    relRelationH tmpRel;
    int retCode;
    int idx;

    assert(stream);
    assert(rel);
    firstElements = relNewSet();
    tmpRel = relNewRelation();
    retCode = relFirstElements3(firstElements, rel);
    for(idx = 0; idx < firstElements->memberNum; idx++)
    {   (*dumpFirstPosMember)(stream, firstElements->members[idx]);
        retCode = relApply3(tmpRel, rel, firstElements->members[idx]);
	assert(retCode == RELOK);
	relDumpRelation(stream,
			tmpRel,
			dumpSecondPosMember,
			dumpThirdPosMember,
			firstElements->members[idx],
			dumpMember2Newline,
			dumpMemberSpace);
	dumpMember1Newline(stream);
    }
    relFreeSet(firstElements);
}

/* dump a set to a stream in ASCII format */
/* 3rd arg: a function that prints one member */
/* 4th arg: a function that prints the space between two members */
/* Dump is in the iteration order */
void
relDumpSet(FILE *stream,
	   relSetH set,
	   void (*dumpMember)(FILE *, int),
	   void (*dumpMemberSpace)(FILE *))
{   int idx;

    assert(stream);
    assert(set);
    for(idx = 0; idx < set->memberNum; idx++)
    {   if(idx)
	    (*dumpMemberSpace)(stream);
        (*dumpMember)(stream, set->members[idx]);
    }
}

/* dump a set to a stream in ASCII format, passing a selector to the 
   print function for members */
/* 3rd arg: a function that prints one member */
/* 4th arg: a function that prints the space between two members */
/* 5th arg: the selector */
/* Dump is in the iteration order */
void
relDumpSetS(FILE *stream,
	   relSetH set,
	   void (*dumpMember)(FILE *, int, int),
	   void (*dumpMemberSpace)(FILE *),
	   int selector)
{   int idx;

    assert(stream);
    assert(set);
    for(idx = 0; idx < set->memberNum; idx++)
    {   if(idx)
	    (*dumpMemberSpace)(stream);
        (*dumpMember)(stream, set->members[idx], selector);
    }
}

/* some example functions for passing to the dump relations:

/* print the space between two members of a set */
void
relDumpMemberSpace(FILE* stream)
{   fprintf(stream, " ");
}

/* print the newline after a domain member of a relation */
void
relDumpMemberNewline(FILE* stream)
{   fprintf(stream, "\n");
}

/* calculate the domain of a relation */
/* The set for the result must already be allocated and is overwritten */
/* returns: RELOK on success */
/*          RELCANTDO if the domain is empty, i.e., if the relation is empty */
/* exits with an error message and code 2 if there is no more memory */
/* Invariant: the iteration order is the order in which the members appear
   first in the ordered relation */
int
relDomain(relSetH result, relRelationH rel)
{   int idx;

    assert(result);
    assert(rel);
    result->memberNum = 0;
    if(!rel->relships.memberNum)
        return(RELCANTDO);
    for(idx = 0; idx < rel->relships.memberNum; idx++)
        relAddMember(result, rel->relships.members[idx].first);
    return(RELOK);
}

/* calculate the range of a relation */
/* The set for the result must already be allocated and is overwritten */
/* returns: RELOK on success */
/*          RELCANTDO if the range is empty, i.e., if the relation is empty */
/* exits with an error message and code 2 if there is no more memory */
/* Invariant: the iteration order is the order in which the members appear
   first in the ordered relation */
int
relRange(relSetH result, relRelationH rel)
{   int idx;

    assert(result);
    assert(rel);
    result->memberNum = 0;
    if(!rel->relships.memberNum)
        return(RELCANTDO);
    for(idx = 0; idx < rel->relships.memberNum; idx++)
        relAddMember(result, rel->relships.members[idx].second);
    return(RELOK);
}

/* calculate the binary relation that contains the second and third elements
   of a ternary relation */
/* The relation for the result must already be allocated and is overwritten */
/* returns: RELOK on success */
/*          RELCANTDO if the the ternary relation is empty */
/* exits with an error message and code 2 if there is no more memory */
/* Invariant: the iteration order is the order in which the binary relships
   appear first in the ordered relation */
int
relRelationSecondThird3(relRelationH result, relRelation3H rel)
{   int idx;

    assert(result);
    assert(rel);
    result->relships.memberNum = 0;
    if(!rel->relships.memberNum)
        return(RELCANTDO);
    for(idx = 0; idx < rel->relships.memberNum; idx++)
        relAddRelship(result, 
		      rel->relships.members[idx].second,
		      rel->relships.members[idx].third);
    return(RELOK);
}

/* calculate the set of elements that appear at the first position of
   a relation */
/* The set for the result must already be allocated and is overwritten */
/* returns: RELOK on success */
/*          RELCANTDO if the set is empty, i.e., if the relation is empty */
/* exits with an error message and code 2 if there is no more memory */
/* Invariant: the iteration order is the order in which the members appear
   first in the ordered relation */
int
relFirstElements3(relSetH result, relRelation3H rel)
{   int idx;

    assert(result);
    assert(rel);
    result->memberNum = 0;
    if(!rel->relships.memberNum)
        return(RELCANTDO);
    for(idx = 0; idx < rel->relships.memberNum; idx++)
        relAddMember(result, rel->relships.members[idx].first);
    return(RELOK);
}

/* project a relation domain to a set,
   i.e., remove all relationships whose first elements are not in the set */
/* The relation for the result must already be allocated and is overwritten */
/* returns: RELOK on success */
/*          RELCANTDO if nothing could be removed */
/* exits with an error message and code 2 if there is no more memory */
/* Invariant: the iteration order of the relation remains the same, except for
   the removed relships */
int
relProjectDomain(relRelationH result, relRelationH rel, relSetH set)
{   int retCode;
    int allFound = 1;
    int idx;

    assert(result);
    assert(rel);
    assert(set);
    result->relships.memberNum = 0;
    for(idx = 0; idx < rel->relships.memberNum; idx++)
    {   if(relIsMember(set, rel->relships.members[idx].first) == RELOK)
	{   assertTableSpace(result->relships.memberNum < RELMAXRELSHIPS);
	    result->relships.members[result->relships.memberNum].first =
		rel->relships.members[idx].first;
	    result->relships.members[result->relships.memberNum++].second =
		rel->relships.members[idx].second;
	}
	else
	    allFound = 0;
    }
    if(allFound)
        return(RELCANTDO);
    return(RELOK);
}

/* project a relation range to a set,
   i.e., remove all relationships whose second elements are not in the set */
/* The relation for the result must already be allocated and is overwritten */
/* returns: RELOK on success */
/*          RELCANTDO if nothing could be removed */
/* exits with an error message and code 2 if there is no more memory */
/* Invariant: the iteration order of the relation remains the same, except for
   the removed relships */
int
relProjectRange(relRelationH result, relRelationH rel, relSetH set)
{   int retCode;
    int allFound = 1;
    int idx;

    assert(result);
    assert(rel);
    assert(set);
    result->relships.memberNum = 0;
    for(idx = 0; idx < rel->relships.memberNum; idx++)
    {   if(relIsMember(set, rel->relships.members[idx].second) == RELOK)
	{   assertTableSpace(result->relships.memberNum < RELMAXRELSHIPS);
	    result->relships.members[result->relships.memberNum].first =
		rel->relships.members[idx].first;
	    result->relships.members[result->relships.memberNum++].second =
		rel->relships.members[idx].second;
	}
	else
	    allFound = 0;
    }
    if(allFound)
        return(RELCANTDO);
    return(RELOK);
}

/* project a ternary relation's first element to a binary relation,
   i.e., remove all relationships whose first elements are not in the set */
/* The ternary relation for the result must already be allocated and 
   is overwritten */
/* returns: RELOK on success */
/*          RELCANTDO if nothing could be removed */
/* exits with an error message and code 2 if there is no more memory */
/* Invariant: the iteration order of the relation remains the same, except for
   the removed relships */
int
relProjectFirstElem3(relRelation3H result, relRelation3H rel, relSetH set)
{   int retCode;
    int allFound = 1;
    int idx;

    assert(result);
    assert(rel);
    assert(set);
    result->relships.memberNum = 0;
    for(idx = 0; idx < rel->relships.memberNum; idx++)
    {   if(relIsMember(set, rel->relships.members[idx].first) == RELOK)
	{   assertTableSpace(result->relships.memberNum < RELMAXRELSHIPS);
	    result->relships.members[result->relships.memberNum].first =
		rel->relships.members[idx].first;
	    result->relships.members[result->relships.memberNum].second =
		rel->relships.members[idx].second;
	    result->relships.members[result->relships.memberNum++].third =
		rel->relships.members[idx].third;
	}
	else
	    allFound = 0;
    }
    if(allFound)
        return(RELCANTDO);
    return(RELOK);
}

/* project a relation domain and range to a set,
   i.e., remove all relationships whose first or second elements are not 
   in the set */
/* The relation for the result must already be allocated and is overwritten */
/* returns: RELOK on success */
/*          RELCANTDO if nothing could be removed */
/* exits with an error message and code 2 if there is no more memory */
/* Invariant: the iteration order of the relation remains the same, except for
   the removed relships */
int
relProjectDomainRange(relRelationH result, relRelationH rel, relSetH set)
{   int retCode;
    int allFound = 1;
    int idx;

    assert(result);
    assert(rel);
    assert(set);
    result->relships.memberNum = 0;
    for(idx = 0; idx < rel->relships.memberNum; idx++)
    {   if((relIsMember(set, rel->relships.members[idx].first) == RELOK)
           && (relIsMember(set, rel->relships.members[idx].second) == RELOK))
	{   assertTableSpace(result->relships.memberNum < RELMAXRELSHIPS);
	    result->relships.members[result->relships.memberNum].first =
		rel->relships.members[idx].first;
	    result->relships.members[result->relships.memberNum++].second =
		rel->relships.members[idx].second;
	}
	else
	    allFound = 0;
    }
    if(allFound)
        return(RELCANTDO);
    return(RELOK);
}

/* relation difference */
/* Removes all relationships of rel2 from rel1. */
/* returns: RELOK if all relationships could be removed */
/*          RELCANTDO if some relationships did not exist in rel1 */
/* Invariant: the iteration order of rel1 remains the same, except for
   the removed relships */
int
relRelationDiff(relRelationH rel1, relRelationH rel2)
{   int retCode;
    int notFound = 0;
    int idx;

    assert(rel1);
    assert(rel2);
    for(idx = 0; idx < rel2->relships.memberNum; idx++)
    {   retCode = relDelRelship(rel1, 
                                rel2->relships.members[idx].first, 
				rel2->relships.members[idx].second );
	if(retCode == RELCANTDO)
	    notFound = 1;
    }
    if(notFound)
        return(RELCANTDO);
    return(RELOK);
}

/* relation union */
/* Merges all relationships of rel2 into rel1. */
/* returns: RELOK if rel1 became bigger */
/*          RELCANTDO if rel1 did not change */
/* exits with an error message and code 2 if there is no more memory */
/* Invariant: no specific iteration order of the result is guaranteed */
extern int
relRelationUnion(relRelationH rel1, relRelationH rel2)
{   int retCode;
    int somethingAdded = 0;
    int idx;

    assert(rel1);
    assert(rel2);
    for(idx = 0; idx < rel2->relships.memberNum; idx++)
    {   retCode = relAddRelship(rel1, 
                                rel2->relships.members[idx].first, 
				rel2->relships.members[idx].second );
	if(retCode == RELOK)
	    somethingAdded = 1;
    }
    if(!somethingAdded)
        return(RELCANTDO);
    return(RELOK);
}

/* set difference */
/* Removes all members of set2 from set1. */
/* returns: RELOK if all members could be removed */
/*          RELCANTDO if some members did not exist in set1 */
/* Invariant: the iteration order of set1 remains the same, except for
   the removed members */
int
relSetDiff(relSetH set1, relSetH set2)
{   int retCode;
    int notFound = 0;
    int idx;

    assert(set1);
    assert(set2);
    for(idx = 0; idx < set2->memberNum; idx++)
    {   retCode = relDelMember(set1, set2->members[idx]);
	if(retCode == RELCANTDO)
	    notFound = 1;
    }
    if(notFound)
        return(RELCANTDO);
    return(RELOK);
}

/* set union */
/* Merges all members of set2 into set1. */
/* returns: RELOK if set1 became bigger */
/*          RELCANTDO if set1 did not change */
/* exits with an error message and code 2 if there is no more memory */
/* Invariant: no specific iteration order of the result is guaranteed */
int
relSetUnion(relSetH set1, relSetH set2)
{   int retCode;
    int somethingAdded = 0;
    int idx;

    assert(set1);
    assert(set2);
    for(idx = 0; idx < set2->memberNum; idx++)
    {   retCode = relAddMember(set1, set2->members[idx]);
	if(retCode == RELOK)
	    somethingAdded = 1;
    }
    if(!somethingAdded)
        return(RELCANTDO);
    return(RELOK);
}

/* set intersection */
/* returns: RELOK if intersection is non-empty */
/*          RELCANTDO if intersection is empty */
/* exits with an error message and code 2 if there is no more memory */
/* The set for the result must already be allocated and is overwritten */
/* Invariant: no specific iteration order of the result is guaranteed */
int
relSetIntersect(relSetH result, relSetH set1, relSetH set2)
{   int somethingAdded = 0;
    int idx;

    assert(result);
    assert(set1);
    assert(set2);
    relClearSet(result);
    for(idx = 0; idx < set1->memberNum; idx++)
    {   if(relIsMember(set2, set1->members[idx]) == RELOK)
        {   somethingAdded = 1;
            relAddMember(result, set1->members[idx]);
	}
    }
    if(!somethingAdded)
        return(RELCANTDO);
    return(RELOK);
}

/* apply a relation */
/* Determines the set of ids with which an id is in a 
   relationship (standing on the left hand side). */
/* The set for the result must already be allocated and is overwritten */
/* returns: RELOK on success */
/*          RELCANTDO if the id is not in the domain */
/* exits with an error message and code 2 if there is no more memory */
/* Invariant: the iteration order is the order in which the relships appear
   in the ordered relation */
int
relApply(relSetH result, relRelationH rel, int name)
{   int idx;
    int found = 0;

    assert(result);
    assert(rel);
    result->memberNum = 0;
    for(idx = 0; idx < rel->relships.memberNum; idx++)
    {   if(rel->relships.members[idx].first == name)
        {   found = 1;
	    relAddMember(result, rel->relships.members[idx].second);
	}
    }
    if(!found)
        return(RELCANTDO);
    return(RELOK);
}

/* apply a function */
/* Determines the unique id with which an id is in a 
   relationship (standing on the left hand side). */
/* There must be such a result, and it must be unique. */
/* returns: the id on success */
/*          RELERR if there is no unique result */
/* exits with an error message and code 2 if there is no more memory */
int
relApplyFunction(relRelationH rel, int name)
{   int retCode;
    relSetH tmpSet;
    relSetIt setIt;
    int result;

    assert(rel);
    tmpSet = relNewSet();

    retCode = relApply(tmpSet, rel, name);
    if((retCode == RELCANTDO) || (relCardSet(tmpSet) != 1))
        result = RELERR;
    else
    {   relInitSetIt(&setIt, tmpSet);
        result = relCurrSetIt(&setIt);
    }

    relFreeSet(tmpSet);

    return(result);
}

/* apply a ternary relation */
/* Determines the binary relation with which an id is in a 
   ternary relationship (standing on the left hand side). */
/* The binary relation for the result must already be allocated and is 
   overwritten */
/* returns: RELOK on success */
/*          RELCANTDO if the id is not in the domain */
/* exits with an error message and code 2 if there is no more memory */
/* Invariant: the iteration order is the order in which the relships appear
   in the ordered relation */
extern int
relApply3(relRelationH result, relRelation3H rel, int name)
{   int idx;
    int found = 0;

    assert(result);
    assert(rel);
    result->relships.memberNum = 0;
    for(idx = 0; idx < rel->relships.memberNum; idx++)
    {   if(rel->relships.members[idx].first == name)
        {   found = 1;
	    relAddRelship(result, 
			  rel->relships.members[idx].second,
			  rel->relships.members[idx].third);
	}
    }
    if(!found)
        return(RELCANTDO);
    return(RELOK);
}

/* apply a function with two arguments */
/* Determines the unique id with which two ids are in a 
   ternary relationship (standing on the left hand and middle side). */
/* There must be such a result, and it must be unique. */
/* returns: the id on success */
/*          RELERR if there is no unique result */
/* exits with an error message and code 2 if there is no more memory */
int
relApplyFunction3(relRelation3H rel, int name1, int name2)
{   int retCode;
    relRelationH tmpRel;
    relRelationIt relIt;
    relSetH tmpSet;
    relSetIt setIt;
    int result;

    assert(rel);
    tmpRel = relNewRelation();
    tmpSet = relNewSet();

    retCode = relApply3(tmpRel, rel, name1);
    if(retCode == RELCANTDO)
        result = RELERR;
    else
    {   retCode = relApply(tmpSet, tmpRel, name2);
	if((retCode == RELCANTDO) || (relCardSet(tmpSet) != 1))
	    result = RELERR;
	else
	{   relInitSetIt(&setIt, tmpSet);
	    result = relCurrSetIt(&setIt);
	}
    }

    relFreeRelation(tmpRel);
    relFreeSet(tmpSet);

    return(result);
}

/* apply the inverse of a relation */
/* Determines the set of ids with which an id is in an 
   inverse relationship (i.e., standing on the right hand side). */
/* The set for the result must already be allocated and is overwritten */
/* returns: RELOK on success */
/*          RELCANTDO if the id is not in the range */
/* exits with an error message and code 2 if there is no more memory */
/* Invariant: the iteration order is the order in which the relships appear
   in the ordered relation */
int
relApplyInverse(relSetH result, relRelationH rel, int name)
{   int idx;
    int found = 0;

    assert(result);
    assert(rel);
    result->memberNum = 0;
    for(idx = 0; idx < rel->relships.memberNum; idx++)
    {   if(rel->relships.members[idx].second == name)
        {   relAddMember(result, rel->relships.members[idx].first);
	    found = 1;
	}
    }
    if(!found)
        return(RELCANTDO);
    return(RELOK);
}

/* apply a function (or relation) to a set */
/* Each element x is replaced by an element f(x). */
/* If f() is a relation, then multiple elements may be inserted. */
/* The set for the result must already be allocated and is overwritten */
/* returns: RELOK on success */
/*          RELCANTDO if a function argument is not in the domain */
/* exits with an error message and code 2 if there is no more memory */
/* Invariant: the iteration order is unchanged, except of the replacement
   of x by f(x) */
int
relApplyToSet(relSetH result, relRelationH func, relSetH set)
{   int idx, idx2;
    int foundInSet;
    int notFoundInDomain;

    assert(result);
    assert(func);
    assert(set);
    result->memberNum = 0;
    /* loop over all elements x */
    notFoundInDomain = 0;
    for(idx = 0; idx < set->memberNum; idx++)
    {   /* search for f(x) */
        foundInSet = 0;
	for(idx2 = 0; idx2 < func->relships.memberNum; idx2++)
	{   if(   func->relships.members[idx2].first
	       == set->members[idx])
	    {   foundInSet = 1;
		/* add element f(x) to the result */
		relAddMember(result, func->relships.members[idx2].second);
	    }
	}
	if(!foundInSet)
	    notFoundInDomain = 1;
    }
    if(notFoundInDomain)
	return(RELCANTDO);

    return(RELOK);
}

/* apply a function (or relation) to a relation */
/* Each relationship (a,b) is replaced by a relationship (f(a),f(b)). */
/* If f() is a relation, then multiple relationships may be inserted. */
/* The relation for the result must already be allocated and is overwritten */
/* returns: RELOK on success */
/*          RELCANTDO if a function argument is not in the domain */
/* exits with an error message and code 2 if there is no more memory */
/* Invariant: no specific iteration order of the result is guaranteed */
int
relApplyToRelation(relRelationH result, relRelationH func, relRelationH rel)
{   int idx, idx2, idx3;
    int found, found2;

    assert(result);
    assert(func);
    assert(rel);
    result->relships.memberNum = 0;
    /* loop over all relationships (a,b) */
    for(idx = 0; idx < rel->relships.memberNum; idx++)
    {   /* search for f(a) */
        found = 0;
	for(idx2 = 0; idx2 < func->relships.memberNum; idx2++)
	{   if(   func->relships.members[idx2].first
	       == rel->relships.members[idx].first)
	    {   found = 1;
	        /* search for f(b) */
		found2 = 0;
		for(idx3 = 0; idx3 < func->relships.memberNum; idx3++)
		{   if(   func->relships.members[idx3].first
		       == rel->relships.members[idx].second)
		    {   found2 = 1;
			/* add relationship (f(a),f(b)) to the result */
			relAddRelship(result,
				      func->relships.members[idx2].second,
				      func->relships.members[idx3].second);
		    }
		}
		if(!found2)
		    return(RELCANTDO);
	    }
	}
	if(!found)
	    return(RELCANTDO);
    }

    return(RELOK);
}

/* compute the inverse of a relation */
/* Inverts all relationships (a,b) to (b,a). */
/* The relation for the result must already be allocated and is overwritten */
/* returns: RELOK on success */
/*          RELCANTDO if the relation is empty */
/* Invariant: the iteration order of the relships is unchanged, except of
   the inversion */
int
relInverse(relRelationH result, relRelationH rel)
{   int retCode;
    int idx;
    int found = 0;

    assert(result);
    assert(rel);
    result->relships.memberNum = rel->relships.memberNum;
    for(idx = 0; idx < rel->relships.memberNum; idx++)
    {   found = 1;
        result->relships.members[idx].first = rel->relships.members[idx].second;
        result->relships.members[idx].second = rel->relships.members[idx].first;
    }

    if(!found)
	return(RELCANTDO);
    return(RELOK);
}

/* remove superfluous arrows from transitive closure */
/* If we have {(a,b), (b,c), (a,c)}, we get {(a,b), (b,c)}. */
/* Cycles are allowed, too. */
/* The relation is edited in-place */
/* returns: RELOK on success */
/*          RELCANTDO if the relation is empty */
/* Invariant: no specific iteration order of the result is guaranteed */
int
relDeTransitive(relRelationH rel)
{   int retCode;
    int idx;
    relRelship relship;

    assert(rel);
    if(!rel->relships.memberNum)
	return(RELCANTDO);
    /* loop over all relationships */
    for(idx = 0; idx < rel->relships.memberNum; idx++)
    {   /* save this relationship */
        relship = rel->relships.members[idx];

        /* remove this relationship, leaving all previous relationships
	   untouched */
        lDelRelship(rel, idx);

	/* check wether still a path exists */
        retCode = lExistsPath(rel, relship.first, relship.second);
	if(retCode == RELCANTDO)
	{   /* if no, re-insert the relationship in the same place */
	    lAddRelship(rel, idx, relship);
	}
	else
	{   /* if yes, adjust our loop index to the deletion */
	    idx--;
	}
    }

    return(RELOK);
}

/* remove superfluous reflexive arrows, i.e., back-arrows to the same node */
/* If we have {(a,b), (b,c), (a,a)}, we get {(a,b), (b,c)}. */
/* The relation is edited in-place */
/* returns: RELOK on success */
/*          RELCANTDO if the relation is empty */
/* Invariant: no specific iteration order of the result is guaranteed */
int
relDeReflexive(relRelationH rel)
{   int retCode;
    int idx;

    assert(rel);
    if(!rel->relships.memberNum)
	return(RELCANTDO);
    /* loop over all relationships */
    for(idx = 0; idx < rel->relships.memberNum; idx++)
    {   /* check whether this relationship is a back-arrow to the same node */
        if(   rel->relships.members[idx].first
	   == rel->relships.members[idx].second)
	{   /* if yes, delete it */
	    lDelRelship(rel, idx);
	    /* and adjust our loop index to the deletion */
	    idx--;
	}
    }

    return(RELOK);
}

/* check whether an id is member of a set */
/* returns: RELOK on yes */
/*          RELCANTDO on no */
int
relIsMember(relSetH set, int name)
{   int *dummy;

    assert(set);
    return(lRelIsMember(set, name, &dummy));
}

/* check whether a set is empty */
/* returns: RELOK on yes */
/*          RELCANTDO on no */
int
relIsEmptySet(relSetH set)
{   assert(set);
    if(set->memberNum)
	return(RELCANTDO);
    else
	return(RELOK);
}

/* get the cardinality of a set */
/* returns: the number of elements in the set */
int
relCardSet(relSetH set)
{   assert(set);
    return(set->memberNum);
}

/* check whether a relationship is member of a relation */
/* returns: RELOK on yes */
/*          RELCANTDO on no */
int
relIsRelship(relRelationH rel, int name1, int name2)
{   int idx;

    assert(rel);
    for(idx = 0; idx < rel->relships.memberNum; idx++)
    {   if((rel->relships.members[idx].first == name1)
           && (rel->relships.members[idx].second == name2))
        {   return(RELOK);
	}
    }
    return(RELCANTDO);
}

/* check whether a relationship is member of a ternary relation */
/* returns: RELOK on yes */
/*          RELCANTDO on no */
int
relIsRelship3(relRelation3H rel, int name1, int name2, int name3)
{   int idx;

    assert(rel);
    for(idx = 0; idx < rel->relships.memberNum; idx++)
    {   if((rel->relships.members[idx].first == name1)
           && (rel->relships.members[idx].second == name2)
           && (rel->relships.members[idx].third == name3))
        {   return(RELOK);
	}
    }
    return(RELCANTDO);
}

/* iterators */

/* Please note that the iterated-over sets and relations must not be
   modified while iterating. */

/* initialize an iterator over a set */
/* This might be implemented as a macro. */
void
relInitSetIt(relSetIt *it, relSetH set)
{   assert(it);
    assert(set);
    it->set = set;
    it->currIdx = 0;
}

/* test an iterator over a set whether an element is available */
/* returns: 1 on yes */
/*          0 on no */
/* This might be implemented as a macro. */
char
relMoreSetIt(relSetIt *it)
{   assert(it);
    return(it->currIdx < it->set->memberNum);
}

/* test an iterator over a set whether we are at the last element */
/* returns: 1 on yes */
/*          0 on no */
/* This might be implemented as a macro. */
char
relLastSetIt(relSetIt *it)
{   assert(it);
    return(it->currIdx == (it->set->memberNum - 1));
}

/* return the current id of the set of the iterator */
/* relMoreSetIt() must be 1. */
/* This might be implemented as a macro. */
int
relCurrSetIt(relSetIt *it)
{   assert(it);
    assert(it->set);
    assert(0 <= it->currIdx);
    assert(it->currIdx < it->set->memberNum);
    return(it->set->members[it->currIdx]);
}

/* step to the next element of the set of the iterator */
/* relMoreSetIt() must be 1. */
/* This might be implemented as a macro. */
/* The iteration order on sets is as specified with some of the operations
   on sets, and undefined if nothing was stated. */
void
relNextSetIt(relSetIt *it)
{   assert(it);
    assert(it->set);
    assert(0 <= it->currIdx);
    assert(it->currIdx < it->set->memberNum);
    it->currIdx++;
}

/* initialize an iterator over a relation */
/* This might be implemented as a macro. */
void
relInitRelationIt(relRelationIt *it, relRelationH rel)
{   assert(it);
    assert(rel);
    it->rel = rel;
    it->currIdx = 0;
}

/* test an iterator over a relation whether a relationship is available */
/* returns: 1 on yes */
/*          0 on no */
/* This might be implemented as a macro. */
char
relMoreRelationIt(relRelationIt *it)
{   assert(it);
    return(it->currIdx < it->rel->relships.memberNum);
}

/* return a pointer to the current relationship of the iterator */
/* relMoreRelationIt() must be 1. */
/* This might be implemented as a macro. */
relRelship *
relCurrRelationIt(relRelationIt *it)
{   assert(it);
    assert(it->rel);
    assert(0 <= it->currIdx);
    assert(it->currIdx < it->rel->relships.memberNum);
    return(&it->rel->relships.members[it->currIdx]);
}

/* step to the next relationship of the relation of the iterator */
/* relMoreRelationIt() must be 1. */
/* This might be implemented as a macro. */
void
relNextRelationIt(relRelationIt *it)
{   assert(it);
    assert(it->rel);
    assert(0 <= it->currIdx);
    assert(it->currIdx < it->rel->relships.memberNum);
    it->currIdx++;
}

/* access the first element of a relationship */
/* This might be implemented as a macro. */
int
relFirstElem(relRelship *relship)
{   assert(relship);
    return(relship->first);
}

/* access the second element of a relationship */
/* This might be implemented as a macro. */
int
relSecondElem(relRelship *relship)
{   assert(relship);
    return(relship->second);
}

/* initialize an iterator over a ternary relation */
/* This might be implemented as a macro. */
void
relInitRelation3It(relRelation3It *it, relRelation3H rel)
{   assert(it);
    assert(rel);
    it->rel = rel;
    it->currIdx = 0;
}

/* test an iterator over a ternary relation whether a relationship is
   available */
/* returns: 1 on yes */
/*          0 on no */
/* This might be implemented as a macro. */
char
relMoreRelation3It(relRelation3It *it)
{   assert(it);
    return(it->currIdx < it->rel->relships.memberNum);
}

/* return a pointer to the current ternary relationship of the iterator */
/* relMoreRelation3It() must be 1. */
/* This might be implemented as a macro. */
relRelship3 *
relCurrRelation3It(relRelation3It *it)
{   assert(it);
    assert(it->rel);
    assert(0 <= it->currIdx);
    assert(it->currIdx < it->rel->relships.memberNum);
    return(&it->rel->relships.members[it->currIdx]);
}

/* step to the next relationship of the ternary relation of the iterator */
/* relMoreRelation3It() must be 1. */
/* This might be implemented as a macro. */
/* The iteration order on ternary relations is as specified with some of the
   operations on ternary relations, and undefined if nothing was stated. */
void
relNextRelation3It(relRelation3It *it)
{   assert(it);
    assert(it->rel);
    assert(0 <= it->currIdx);
    assert(it->currIdx < it->rel->relships.memberNum);
    it->currIdx++;
}

/* access the first element of a ternary relationship */
/* This might be implemented as a macro. */
int
relFirstElem3(relRelship3 *relship)
{   assert(relship);
    return(relship->first);
}

/* access the second element of a ternary relationship */
/* This might be implemented as a macro. */
int
relSecondElem3(relRelship3 *relship)
{   assert(relship);
    return(relship->second);
}

/* access the third element of a ternary relationship */
/* This might be implemented as a macro. */
int
relThirdElem3(relRelship3 *relship)
{   assert(relship);
    return(relship->third);
}

