/*-------------------------------------------------------------------------

project: genFamMem - maintaining families of requirements

author: Jan Bredereke
copyright: Jan Bredereke, 2000
$Id: relation.h,v 1.28 2001/02/20 08:01:42 brederek Exp $

file: relation.h - header file: sets of ID numbers and 
		   mathematical relations over them

If NDEBUG is defined, the internal assertions are not checked anymore.

-------------------------------------------------------------------------*/

#ifndef RELATION_H
#define RELATION_H

#include <stdlib.h>
#include <stdio.h>


/*-------------------------------------------------------------------------
The constants for external use
-------------------------------------------------------------------------*/

/* return codes */
/* OK */
#define RELOK	  0
/* ERROR */
#define RELERR	  -1
/* could not do something because it already existed / did not exist */
#define RELCANTDO 1

/* Please note that the below two constants may change when we switch
   to dynamic memory management. */
/* max. number of members in a set */
#define RELMAXSETMEMBERS	500
/* max. number of relationships in a relation */
#define RELMAXRELSHIPS	2000

/*-------------------------------------------------------------------------
The types for external use
-------------------------------------------------------------------------*/

/* handle for a mathematical relation over ids */
typedef struct rel_Relation *relRelationH;
/* one relationship of such a relation */
typedef struct rel_Relship relRelship;
/* handle for a ternary mathematical relation over ids */
typedef struct rel_Relation3 *relRelation3H;
/* one relationship of such a relation */
typedef struct rel_Relship3 relRelship3;
/* handle for a mathematical set over ids */
typedef struct rel_Set *relSetH;
/* iterator for sets */
typedef struct rel_SetIt relSetIt;
/* iterator for relations */
typedef struct rel_RelationIt relRelationIt;
/* iterator for ternary relations */
typedef struct rel_Relation3It relRelation3It;

/* Ids are integers. */

/*-------------------------------------------------------------------------
Some type definitions for internal use only, and for memory allocation
-------------------------------------------------------------------------*/
/* Please note that we plan to switch to dynamic memory allocation and
   re-allocation. Do not rely on the static structure below. */
/* sets */
struct rel_Set
{   int members[RELMAXSETMEMBERS];
    int memberNum;
};
/* binary relations */
struct rel_Relship
{   int first;
    int second;
};
struct rel_Relships
{   relRelship members[RELMAXRELSHIPS];
    int memberNum;
};
struct rel_Relation
{   struct rel_Relships relships;
};
/* ternary relations */
struct rel_Relship3
{   int first;
    int second;
    int third;
};
struct rel_Relships3
{   relRelship3 members[RELMAXRELSHIPS];
    int memberNum;
};
struct rel_Relation3
{   struct rel_Relships3 relships;
};
/* iterator for sets */
struct rel_SetIt
{   struct rel_Set *set;
    int currIdx;
};
/* iterator for relations */
struct rel_RelationIt
{   struct rel_Relation *rel;
    int currIdx;
};
/* iterator for ternary relations */
struct rel_Relation3It
{   struct rel_Relation3 *rel;
    int currIdx;
};

/*-------------------------------------------------------------------------
The functions for external use
-------------------------------------------------------------------------*/

/* alloc and initialize a new relation */
/* The relation will be empty. */
/* returns: the relation on success */
/* exits with an error message and code 2 if there is no more memory */
extern relRelationH 
relNewRelation();

/* alloc and initialize a new ternary relation */
/* The relation will be empty. */
/* returns: the relation on success */
/* exits with an error message and code 2 if there is no more memory */
extern relRelation3H 
relNewRelation3();

/* alloc and initialize a new set */
/* The set will be empty. */
/* returns: the set on success */
/* exits with an error message and code 2 if there is no more memory */
extern relSetH 
relNewSet();

/* free the memory of a relation */
extern void
relFreeRelation(relRelationH rel);

/* free the memory of a ternary relation */
extern void
relFreeRelation3(relRelation3H rel);

/* free the memory of a set */
extern void
relFreeSet(relSetH set);

/* clear an existing set, it becomes the empty set */
extern void
relClearSet(relSetH set);

/* clear an existing relation, it becomes the empty relation */
extern void
relClearRelation(relRelationH rel);

/* clear an existing ternary relation, it becomes the empty ternary relation */
extern void
relClearRelation3(relRelation3H rel);

/* copy a set */
/* The set for the result must already be allocated and is overwritten */
/* returns: RELOK on success */
/*          RELCANTDO if the set is empty */
/* Invariant: the iteration order of the result is the same as that of the
   source */
extern int
relCopySet(relSetH result, relSetH set);

/* copy a relation */
/* The relation for the result must already be allocated and is overwritten */
/* returns: RELOK on success */
/*          RELCANTDO if the relation is empty */
/* Invariant: the iteration order of the result is the same as that of the
   source */
extern int
relCopyRelation(relRelationH result, relRelationH set);

/* copy a ternary relation */
/* The relation for the result must already be allocated and is overwritten */
/* returns: RELOK on success */
/*          RELCANTDO if the ternary relation is empty */
/* Invariant: the iteration order of the result is the same as that of the
   source */
extern int
relCopyRelation3(relRelation3H result, relRelation3H set);

/* add the relationship between ids name1 and name2 to relation rel */
/* returns: RELOK when relationship did not exist before */ 
/*          RELCANTDO when relationship did exist before */ 
/* exits with an error message and code 2 if there is no more memory */
/* Invariant: the iteration order remains the same, except that the new relship
   is appended to the end */
extern int
relAddRelship(relRelationH rel, int name1, int name2);

/* add the ternary relationship between ids name1, name2, and name3 to 
   relation rel */
/* returns: RELOK when relationship did not exist before */ 
/*          RELCANTDO when relationship did exist before */ 
/* exits with an error message and code 2 if there is no more memory */
/* Invariant: the iteration order remains the same, except that the new relship
   is appended to the end */
extern int
relAddRelship3(relRelation3H rel, int name1, int name2, int name3);

/* delete the relationship between ids name1 and name2 from relation rel */
/* returns: RELOK when relationship did exist */ 
/*          RELCANTDO when relationship did not exist */ 
/* Invariant: the iteration order remains the same, except that the deleted
   relship is removed */
extern int
relDelRelship(relRelationH rel, int name1, int name2);

/* delete the ternary relationship between ids name1, name2, and name3 from 
   relation rel */
/* returns: RELOK when relationship did exist */ 
/*          RELCANTDO when relationship did not exist */ 
/* Invariant: the iteration order remains the same, except that the deleted
   relship is removed */
extern int
relDelRelship3(relRelation3H rel, int name1, int name2, int name3);

/* add an id to a set */
/* returns: RELOK when id was no member of the set */ 
/*          RELCANTDO when id already was member of the set */ 
/* exits with an error message and code 2 if there is no more memory */
/* Invariant: the iteration order remains the same, except that the new id
   is appended to the end */
extern int
relAddMember(relSetH set, int name);

/* deletes an id from a set */
/* returns: RELOK when id was a member of the set */ 
/*          RELCANTDO when id already was no member of the set */ 
/* Invariant: the iteration order remains the same, except that the deleted
   id is removed */
extern int
relDelMember(relSetH set, int name);

/* dump a relation to a stream in ASCII format */
/* 3rd arg: a function that prints one domain member for a selector value */
/* 4rd arg: a function that prints one range member */
/* 5th arg: a selector value */
/* 6th arg: a function that prints the newline between two domain members */
/* 7th arg: a function that prints the space between two range members */
/* Dump is in the iteration order */
extern void
relDumpRelation(FILE *stream, 
                relRelationH rel, 
		void (*dumpDomainMember)(FILE *, int, int /* selector */),
		void (*dumpRangeMember)(FILE *, int, int /* domainMember */),
		int selector,
		void (*dumpMemberNewline)(FILE *),
		void (*dumpMemberSpace)(FILE *));

/* dump a ternary relation to a stream in ASCII format */
/* 3rd arg: a function that prints a member at the first position of 
            a relationship */
/* 4rd arg: a function that prints a member at the second position of 
            a relationship, for a given selector value */
/* 5rd arg: a function that prints a member at the third position of 
            a relationship */
/* 6th arg: a function that prints the newline between two first pos. members */
/* 7th arg: a function that prints the newline between two 2nd pos. members */
/* 8th arg: a function that prints the space between two 3rd pos. members */
/* Dump is in the iteration order */
extern void
relDumpRelation3(FILE *stream, 
		 relRelation3H rel,
		 void (*dumpFirstPosMember)(FILE *, int),
		 void (*dumpSecondPosMember)(FILE *, int, int /* selector */),
		 void (*dumpThirdPosMember)(FILE *, int,int/*secondPosMember*/),
		 void (*dumpMember1Newline)(FILE *),
		 void (*dumpMember2Newline)(FILE *),
		 void (*dumpMemberSpace)(FILE *));

/* dump a set to a stream in ASCII format */
/* 3rd arg: a function that prints one member */
/* 4th arg: a function that prints the space between two members */
/* Dump is in the iteration order */
extern void
relDumpSet(FILE *stream,
	   relSetH set,
	   void (*dumpMember)(FILE *, int),
	   void (*dumpMemberSpace)(FILE *));

/* dump a set to a stream in ASCII format, passing a selector to the 
   print function for members */
/* 3rd arg: a function that prints one member */
/* 4th arg: a function that prints the space between two members */
/* 5th arg: the selector */
/* Dump is in the iteration order */
extern void
relDumpSetS(FILE *stream,
	   relSetH set,
	   void (*dumpMember)(FILE *, int, int),
	   void (*dumpMemberSpace)(FILE *),
	   int selector);

/* some example functions for passing to the dump relations:

/* print the space between two members of a set */
extern void
relDumpMemberSpace(FILE* stream);

/* print the newline after a domain member of a relation */
extern void
relDumpMemberNewline(FILE* stream);

/* calculate the domain of a relation */
/* The set for the result must already be allocated and is overwritten */
/* returns: RELOK on success */
/*          RELCANTDO if the domain is empty, i.e., if the relation is empty */
/* exits with an error message and code 2 if there is no more memory */
/* Invariant: the iteration order is the order in which the members appear
   first in the ordered relation */
extern int
relDomain(relSetH result, relRelationH rel);

/* calculate the range of a relation */
/* The set for the result must already be allocated and is overwritten */
/* returns: RELOK on success */
/*          RELCANTDO if the range is empty, i.e., if the relation is empty */
/* exits with an error message and code 2 if there is no more memory */
/* Invariant: the iteration order is the order in which the members appear
   first in the ordered relation */
extern int
relRange(relSetH result, relRelationH rel);

/* calculate the binary relation that contains the second and third elements
   of a ternary relation */
/* The relation for the result must already be allocated and is overwritten */
/* returns: RELOK on success */
/*          RELCANTDO if the the ternary relation is empty */
/* exits with an error message and code 2 if there is no more memory */
/* Invariant: the iteration order is the order in which the binary relships
   appear first in the ordered relation */
extern int
relRelationSecondThird3(relRelationH result, relRelation3H rel);

/* calculate the set of elements that appear at the first position of
   a relation */
/* The set for the result must already be allocated and is overwritten */
/* returns: RELOK on success */
/*          RELCANTDO if the set is empty, i.e., if the relation is empty */
/* exits with an error message and code 2 if there is no more memory */
/* Invariant: the iteration order is the order in which the members appear
   first in the ordered relation */
extern int
relFirstElements3(relSetH result, relRelation3H rel);

/* project a relation domain to a set,
   i.e., remove all relationships whose first elements are not in the set */
/* The relation for the result must already be allocated and is overwritten */
/* returns: RELOK on success */
/*          RELCANTDO if nothing could be removed */
/* exits with an error message and code 2 if there is no more memory */
/* Invariant: the iteration order of the relation remains the same, except for
   the removed relships */
extern int
relProjectDomain(relRelationH result, relRelationH rel, relSetH set);

/* project a relation range to a set,
   i.e., remove all relationships whose second elements are not in the set */
/* The relation for the result must already be allocated and is overwritten */
/* returns: RELOK on success */
/*          RELCANTDO if nothing could be removed */
/* exits with an error message and code 2 if there is no more memory */
/* Invariant: the iteration order of the relation remains the same, except for
   the removed relships */
extern int
relProjectRange(relRelationH result, relRelationH rel, relSetH set);

/* project a ternary relation's first element to a binary relation,
   i.e., remove all relationships whose first elements are not in the set */
/* The ternary relation for the result must already be allocated and 
   is overwritten */
/* returns: RELOK on success */
/*          RELCANTDO if nothing could be removed */
/* exits with an error message and code 2 if there is no more memory */
/* Invariant: the iteration order of the relation remains the same, except for
   the removed relships */
extern int
relProjectFirstElem3(relRelation3H result, relRelation3H rel, relSetH set);

/* project a relation domain and range to a set,
   i.e., remove all relationships whose first or second elements are not 
   in the set */
/* The relation for the result must already be allocated and is overwritten */
/* returns: RELOK on success */
/*          RELCANTDO if nothing could be removed */
/* exits with an error message and code 2 if there is no more memory */
/* Invariant: the iteration order of the relation remains the same, except for
   the removed relships */
extern int
relProjectDomainRange(relRelationH result, relRelationH rel, relSetH set);

/* relation difference */
/* Removes all relationships of rel2 from rel1. */
/* returns: RELOK if all relationships could be removed */
/*          RELCANTDO if some relationships did not exist in rel1 */
/* Invariant: the iteration order of rel1 remains the same, except for
   the removed relships */
extern int
relRelationDiff(relRelationH rel1, relRelationH rel2);

/* relation union */
/* Merges all relationships of rel2 into rel1. */
/* returns: RELOK if rel1 became bigger */
/*          RELCANTDO if rel1 did not change */
/* exits with an error message and code 2 if there is no more memory */
/* Invariant: no specific iteration order of the result is guaranteed */
extern int
relRelationUnion(relRelationH rel1, relRelationH rel2);

/* set difference */
/* Removes all members of set2 from set1. */
/* returns: RELOK if all members could be removed */
/*          RELCANTDO if some members did not exist in set1 */
/* Invariant: the iteration order of set1 remains the same, except for
   the removed members */
extern int
relSetDiff(relSetH set1, relSetH set2);

/* set union */
/* Merges all members of set2 into set1. */
/* returns: RELOK if set1 became bigger */
/*          RELCANTDO if set1 did not change */
/* exits with an error message and code 2 if there is no more memory */
/* Invariant: no specific iteration order of the result is guaranteed */
extern int
relSetUnion(relSetH set1, relSetH set2);

/* set intersection */
/* returns: RELOK if intersection is non-empty */
/*          RELCANTDO if intersection is empty */
/* exits with an error message and code 2 if there is no more memory */
/* The set for the result must already be allocated and is overwritten */
/* Invariant: no specific iteration order of the result is guaranteed */
extern int
relSetIntersect(relSetH result, relSetH set1, relSetH set2);

/* apply a relation */
/* Determines the set of ids with which an id is in a 
   relationship (standing on the left hand side). */
/* The set for the result must already be allocated and is overwritten */
/* returns: RELOK on success */
/*          RELCANTDO if the id is not in the domain */
/* exits with an error message and code 2 if there is no more memory */
/* Invariant: the iteration order is the order in which the relships appear
   in the ordered relation */
extern int
relApply(relSetH result, relRelationH rel, int name);

/* apply a function */
/* Determines the unique id with which an id is in a 
   relationship (standing on the left hand side). */
/* There must be such a result, and it must be unique. */
/* returns: the id on success */
/*          RELERR if there is no unique result */
/* exits with an error message and code 2 if there is no more memory */
extern int
relApplyFunction(relRelationH rel, int name);

/* apply a ternary relation */
/* Determines the binary relation with which an id is in a 
   ternary relationship (standing on the left hand side). */
/* The binary relation for the result must already be allocated and is 
   overwritten */
/* returns: RELOK on success */
/*          RELCANTDO if the id is not in the domain */
/* exits with an error message and code 2 if there is no more memory */
/* Invariant: the iteration order is the order in which the relships appear
   in the ordered relation */
extern int
relApply3(relRelationH result, relRelation3H rel, int name);

/* apply a function with two arguments */
/* Determines the unique id with which two ids are in a 
   ternary relationship (standing on the left hand and middle side). */
/* There must be such a result, and it must be unique. */
/* returns: the id on success */
/*          RELERR if there is no unique result */
/* exits with an error message and code 2 if there is no more memory */
extern int
relApplyFunction3(relRelation3H rel, int name1, int name2);

/* apply the inverse of a relation */
/* Determines the set of ids with which an id is in an 
   inverse relationship (i.e., standing on the right hand side). */
/* The set for the result must already be allocated and is overwritten */
/* returns: RELOK on success */
/*          RELCANTDO if the id is not in the range */
/* exits with an error message and code 2 if there is no more memory */
/* Invariant: the iteration order is the order in which the relships appear
   in the ordered relation */
extern int
relApplyInverse(relSetH result, relRelationH rel, int name);

/* apply a function (or relation) to a set */
/* Each element x is replaced by an element f(x). */
/* If f() is a relation, then multiple elements may be inserted. */
/* The set for the result must already be allocated and is overwritten */
/* returns: RELOK on success */
/*          RELCANTDO if a function argument is not in the domain */
/* exits with an error message and code 2 if there is no more memory */
/* Invariant: the iteration order is unchanged, except of the replacement
   of x by f(x) */
extern int
relApplyToSet(relSetH result, relRelationH func, relSetH set);

/* apply a function (or relation) to a relation */
/* Each relationship (a,b) is replaced by a relationship (f(a),f(b)). */
/* If f() is a relation, then multiple relationships may be inserted. */
/* The relation for the result must already be allocated and is overwritten */
/* returns: RELOK on success */
/*          RELCANTDO if a function argument is not in the domain */
/* exits with an error message and code 2 if there is no more memory */
/* Invariant: no specific iteration order of the result is guaranteed */
extern int
relApplyToRelation(relRelationH result, relRelationH func, relRelationH rel);

/* compute the inverse of a relation */
/* Inverts all relationships (a,b) to (b,a). */
/* The relation for the result must already be allocated and is overwritten */
/* returns: RELOK on success */
/*          RELCANTDO if the relation is empty */
/* Invariant: the iteration order of the relships is unchanged, except of
   the inversion */
extern int
relInverse(relRelationH result, relRelationH rel);

/* remove superfluous arrows from transitive closure */
/* If we have {(a,b), (b,c), (a,c)}, we get {(a,b), (b,c)}. */
/* Cycles are allowed, too. */
/* The relation is edited in-place */
/* returns: RELOK on success */
/*          RELCANTDO if the relation is empty */
/* Invariant: no specific iteration order of the result is guaranteed */
extern int
relDeTransitive(relRelationH rel);

/* remove superfluous reflexive arrows, i.e., back-arrows to the same node */
/* If we have {(a,b), (b,c), (a,a)}, we get {(a,b), (b,c)}. */
/* The relation is edited in-place */
/* returns: RELOK on success */
/*          RELCANTDO if the relation is empty */
/* Invariant: no specific iteration order of the result is guaranteed */
extern int
relDeReflexive(relRelationH rel);

/* check whether an id is member of a set */
/* returns: RELOK on yes */
/*          RELCANTDO on no */
extern int
relIsMember(relSetH set, int name);

/* check whether a set is empty */
/* returns: RELOK on yes */
/*          RELCANTDO on no */
extern int
relIsEmptySet(relSetH set);

/* get the cardinality of a set */
/* returns: the number of elements in the set */
extern int
relCardSet(relSetH set);

/* check whether a relationship is member of a relation */
/* returns: RELOK on yes */
/*          RELCANTDO on no */
extern int
relIsRelship(relRelationH rel, int name1, int name2);

/* check whether a relationship is member of a ternary relation */
/* returns: RELOK on yes */
/*          RELCANTDO on no */
extern int
relIsRelship3(relRelation3H rel, int name1, int name2, int name3);

/* iterators */

/* Please note that the iterated-over sets and relations must not be
   modified while iterating. */

/* initialize an iterator over a set */
/* This might be implemented as a macro. */
extern void
relInitSetIt(relSetIt *it, relSetH set);

/* test an iterator over a set whether an element is available */
/* returns: 1 on yes */
/*          0 on no */
/* This might be implemented as a macro. */
extern char
relMoreSetIt(relSetIt *it);

/* test an iterator over a set whether we are at the last element */
/* returns: 1 on yes */
/*          0 on no */
/* This might be implemented as a macro. */
extern char
relLastSetIt(relSetIt *it);

/* return the current id of the set of the iterator */
/* relMoreSetIt() must be 1. */
/* This might be implemented as a macro. */
extern int
relCurrSetIt(relSetIt *it);

/* step to the next element of the set of the iterator */
/* relMoreSetIt() must be 1. */
/* This might be implemented as a macro. */
/* The iteration order on sets is as specified with some of the operations
   on sets, and undefined if nothing was stated. */
extern void
relNextSetIt(relSetIt *it);

/* initialize an iterator over a relation */
/* This might be implemented as a macro. */
extern void
relInitRelationIt(relRelationIt *it, relRelationH rel);

/* test an iterator over a relation whether a relationship is available */
/* returns: 1 on yes */
/*          0 on no */
/* This might be implemented as a macro. */
extern char
relMoreRelationIt(relRelationIt *it);

/* return a pointer to the current relationship of the iterator */
/* relMoreRelationIt() must be 1. */
/* This might be implemented as a macro. */
extern relRelship *
relCurrRelationIt(relRelationIt *it);

/* step to the next relationship of the relation of the iterator */
/* relMoreRelationIt() must be 1. */
/* This might be implemented as a macro. */
/* The iteration order on relations is as specified with some of the operations
   on relations, and undefined if nothing was stated. */
extern void
relNextRelationIt(relRelationIt *it);

/* access the first element of a relationship */
/* This might be implemented as a macro. */
extern int
relFirstElem(relRelship *relship);

/* access the second element of a relationship */
/* This might be implemented as a macro. */
extern int
relSecondElem(relRelship *relship);

/* initialize an iterator over a ternary relation */
/* This might be implemented as a macro. */
extern void
relInitRelation3It(relRelation3It *it, relRelation3H rel);

/* test an iterator over a ternary relation whether a relationship is
   available */
/* returns: 1 on yes */
/*          0 on no */
/* This might be implemented as a macro. */
extern char
relMoreRelation3It(relRelation3It *it);

/* return a pointer to the current ternary relationship of the iterator */
/* relMoreRelation3It() must be 1. */
/* This might be implemented as a macro. */
extern relRelship3 *
relCurrRelation3It(relRelation3It *it);

/* step to the next relationship of the ternary relation of the iterator */
/* relMoreRelation3It() must be 1. */
/* This might be implemented as a macro. */
/* The iteration order on ternary relations is as specified with some of the
   operations on ternary relations, and undefined if nothing was stated. */
extern void
relNextRelation3It(relRelation3It *it);

/* access the first element of a ternary relationship */
/* This might be implemented as a macro. */
extern int
relFirstElem3(relRelship3 *relship);

/* access the second element of a ternary relationship */
/* This might be implemented as a macro. */
extern int
relSecondElem3(relRelship3 *relship);

/* access the third element of a ternary relationship */
/* This might be implemented as a macro. */
extern int
relThirdElem3(relRelship3 *relship);

#endif /* RELATION_H */
