/*-------------------------------------------------------------------------

project: genFamMem - maintaining families of requirements

author: Jan Bredereke
copyright: Jan Bredereke, 2000
$Id: secondpasstex.l,v 1.13 2001/02/09 13:56:18 brederek Exp $

file: secondpasstex.l - flex rules for the specification filtering of 
			the second pass, for the LaTeX syntax

yylex returns on newline and on end-of-file only. When called, the 
current mode must be provided in filState.printing.
yylex returns either '\n' after a newline, or 0 at end-of-file.
yylex prints the filtered text on yyout.

-------------------------------------------------------------------------*/

%{

#include "global.h"
#include "filterspec.h"

#define myputc(chr)	fputc(chr, yyout)
#define myputs(str)	fputs(str, yyout)

%}

	/* there is one input file only */
%option noyywrap
	/* make yytext a pointer, such that we are allowed to modify it */
%pointer
	/* if no rule matches, print error instead of printing on stdout */
%option nodefault

	/* start conditions for the rules */
%s PRINTING INHERIT PARENTS DELETING STARTOFLINE LINEADJUST

NONALPHA	([^A-Za-z])
NA		{NONALPHA}

%%

	/* the rules */

	/* start in appropriate mode */
	BEGIN(STARTOFLINE);

	/* rules when in start-of-line mode */
	/* This mode does not read text, it just switches to the next
	   mode, according to the current global "printing",
	   "inherit", and "parents" mode flags. */

<STARTOFLINE>{

.|\n			{ /* push back token */
			  yyless(0);
			  /* switch into current mode */
			  if(filState.printing)
			  {   if(filState.inheritArea)
				  BEGIN(INHERIT);
			      else if(filState.parentsArea)
				  BEGIN(PARENTS);
			      else if(filState.pendingEmptyLinesToDelete)
				  BEGIN(LINEADJUST);
			      else
				  BEGIN(PRINTING);
			  }
			  else
			  {   myputc('%');
			      BEGIN(DELETING);
			  }
			  /* warn if we cannot delete empty lines as 
			     we want */
			  if(   filState.pendingEmptyLinesToDelete
			     && (YY_START != LINEADJUST))
			  {
			      fprintf(stderr,
				  "%s: warning: cannot make up for generated lines by\n",
				  PROGNAME);
			      fprintf(stderr,
				  "\tdeleting empty lines, in line %i. (Still %i lines too many.)\n",
				  filState.currLineNo,
				  filState.pendingEmptyLinesToDelete);
			      /* do not attempt further deletions */
			      filState.pendingEmptyLinesToDelete = 0;
			  }
			}

}

	/* rules when in "printing" mode */

<PRINTING>{

\n			{
			  ECHO;
			  BEGIN(STARTOFLINE);
			  return('\n');
			}

"%".*$			ECHO;

"\\endfeatures"/{NA}	{
			  ECHO;
			  /* save the information that there was an
			     \endfeatures in this line */
			  filState.endfeaturesFound = 1;
			}

"\\end{features}"	{
			  ECHO;
			  /* save the information that there was an
			     \endfeatures in this line */
			  filState.endfeaturesFound = 1;
			}

.			ECHO;

}

	/* rules when in "inherit" mode, i.e., when between \inherit,
	   \defaultproperties and \endinherit, \enddefaultproperties */
	/* We do not print anything, instead we save the input to
	   memory for further processing. */

<INHERIT>{

\n			{
			  filNextBufferLine();
			  BEGIN(STARTOFLINE);
			  return('\n');
			}

"\\defaultproperties"/{NA} {
			  filSaveBuffer("\\inherit");
			}

"\\begin{defaultproperties}" {
			  filSaveBuffer("\\begin{inherit}");
			}

"\\enddefaultproperties"/{NA} {
			  filSaveBuffer("\\endinherit");
			}

"\\end{defaultproperties}" {
			  filSaveBuffer("\\end{inherit}");
			}

.			filSaveBuffer(yytext);

}

	/* rules when in "parents" mode, i.e., when in a parents area */
	/* We do not print anything, instead we save the input to
	   memory for further processing. */

<PARENTS>{

\n			{
			  filNextBufferLine();
			  BEGIN(STARTOFLINE);
			  return('\n');
			}

.			filSaveBuffer(yytext);

}

	/* rules when in "LINEADJUST" mode, i.e., when we delete lines that
	   are essentially empty until  we are back at the desired line
	   count. (After we have inserted generated text.) */

<LINEADJUST>{

^[ \t]*(""|"%".*)\n	{ 
			  /* print nothing, but decrement the number of lines
			     still to do */
			  filState.pendingEmptyLinesToDelete--;
			  BEGIN(STARTOFLINE);
			  return('\n');
			}

.|\n			{
			  /* warn if we cannot do what we want */
			  fprintf(stderr,
			      "%s: warning: cannot make up for generated lines by\n",
			      PROGNAME);
			  fprintf(stderr,
			      "\tdeleting empty lines, in line %i. (Still %i lines too many.)\n",
			      filState.currLineNo + 1,
			      filState.pendingEmptyLinesToDelete);
			  /* do not attempt further deletions */
			  filState.pendingEmptyLinesToDelete = 0;
			  /* push back everything and go to normal printing */
			  yyless(0);
			  BEGIN(PRINTING);
			}

}

	/* rules when in "deleting" mode */

<DELETING>{

\n			{
			  /* newlines are printed even in deleted ranges,
			     in order to keep the line counts intact */
			  ECHO;
			  BEGIN(STARTOFLINE);
			  return('\n');
			}

.			{
			  /* empty */
			}

}


<<EOF>>			{
			  return(0);
			}

<INITIAL>(.|\n)		{
			  fprintf(stderr, "%s: filtering: internal error: illegal start condition reached.\n", PROGNAME);
	    		  exit(EXITRESOURCES);
			  }


%%

